@@ -1,7 +1,7 @@
 ;;; c-backend.scm - C-generating backend for the CHICKEN compiler
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -115,6 +115,13 @@
 		      (loop (cdr bs) (add1 i) (sub1 count)) ]
 		     [else (expr (car bs) i)] ) ) )
 
+	    ((##core#let_unboxed)
+	     (let ((name (first params)))
+	       (gen #t name #\=)
+	       (expr (first subs) i)
+	       (gen #\;)
+	       (expr (second subs) i)))
+
 	    ((##core#ref) 
 	     (gen "((C_word*)")
 	     (expr (car subs) i)
@@ -193,7 +200,7 @@
 	       (if block
 		   (gen "C_mutate(&lf[" index "]")
 		   (gen "C_mutate((C_word*)lf[" index "]+1") )
-	       (gen " /* (set! " (uncommentify (symbol->string var)) " ...) */,")
+	       (gen " /* (set! " (uncommentify (##sys#symbol->qualified-string var)) " ...) */,")
 	       (expr (car subs) i)
 	       (gen #\)) ) )
 
@@ -360,12 +367,12 @@
 	     (gen ");") )
 
 	    ((##core#inline)
-	     (gen "(C_word)" (first params) #\()
+	     (gen (first params) #\()
 	     (expr-args subs i)
 	     (gen #\)) )
 
 	    ((##core#inline_allocate)
-	     (gen "(C_word)" (first params) "(&a," (length subs))
+	     (gen (first params) "(&a," (length subs))
 	     (if (pair? subs)
 		 (begin
 		   (gen #\,)
@@ -395,6 +402,19 @@
 	       (expr (second subs) i) 
 	       (gen "),C_SCHEME_UNDEFINED)") ) )
 
+	    ((##core#unboxed_ref)
+	     (gen (first params)))
+
+	    ((##core#unboxed_set!)
+	     (gen "((" (first params) #\=)
+	     (expr (first subs) i) 
+	     (gen "),C_SCHEME_UNDEFINED)"))
+
+	    ((##core#inline_unboxed)	;XXX is this needed?
+	     (gen (first params) "(")
+	     (expr-args subs i)
+	     (gen #\)))
+
 	    ((##core#switch)
 	     (gen #t "switch(")
 	     (expr (first subs) i)
@@ -419,7 +439,7 @@
 	     (expr (third subs) i)
 	     (gen #\)) )
 
-	    (else (bomb "bad form")) ) ) )
+	    (else (bomb "bad form" (node-class n))) ) ) )
     
       (define (expr-args args i)
 	(pair-for-each
@@ -529,8 +549,6 @@
 		   [else
 		    (let ((uname (if unit-name (string-append unit-name "_toplevel") "toplevel")))
 		      (gen "C_noret_decl(C_" uname ")" #t)
-		      (when emit-unsafe-marker
-			(gen "C_externexport void C_dynamic_and_unsafe(void) {}" #t) )
 		      (gen "C_externexport void C_ccall ")
 		      (gen "C_" uname) ) ] )
 	     (gen #\()
@@ -615,9 +633,7 @@
 		      (gen ");}") ]
 		     [(or rest (> (lambda-literal-allocated ll) 0) (lambda-literal-external ll))
 		      (if (and rest (not (eq? rest-mode 'none)))
-			  (if (eq? rest-mode 'vector)
-			      (set! nsrv (lset-adjoin = nsrv argc))
-			      (set! nsr (lset-adjoin = nsr argc)) ) 
+			  (set! nsr (lset-adjoin = nsr argc)) 
 			  (set! ns (lset-adjoin = ns argc)) ) ] ) ) ) )
 	 lambdas)
 	(for-each
@@ -705,29 +721,39 @@
 	(##sys#copy-bytes s s2 start 0 len)
 	s2) )
 
+    (define (utype t)
+      (case t
+	((fixnum) "int")
+	((flonum) "double")
+	((char) "char")
+	((pointer) "void *")
+	((bool) "int")
+	(else (bomb "invalid unboxed type" t))))
+
     (define (procedures)
       (for-each
        (lambda (ll)
-	 (let* ([n (lambda-literal-argument-count ll)]
-		[id (lambda-literal-id ll)]
-		[rname (real-name id db)]
-		[demand (lambda-literal-allocated ll)]
-		[rest (lambda-literal-rest-argument ll)]
-		[customizable (lambda-literal-customizable ll)]
-		[empty-closure (and customizable (zero? (lambda-literal-closure-size ll)))]
-		[nec (- n (if empty-closure 1 0))]
-		[vlist0 (make-variable-list n "t")]
-		[alist0 (make-argument-list n "t")]
-		[varlist (intersperse (if empty-closure (cdr vlist0) vlist0) #\,)]
-		[arglist (intersperse (if empty-closure (cdr alist0) alist0) #\,)]
-		[external (lambda-literal-external ll)]
-		[looping (lambda-literal-looping ll)]
-		[direct (lambda-literal-direct ll)]
-		[rest-mode (lambda-literal-rest-argument-mode ll)]
-		[temps (lambda-literal-temporaries ll)]
-		[topname (if unit-name
+	 (let* ((n (lambda-literal-argument-count ll))
+		(id (lambda-literal-id ll))
+		(rname (real-name id db))
+		(demand (lambda-literal-allocated ll))
+		(rest (lambda-literal-rest-argument ll))
+		(customizable (lambda-literal-customizable ll))
+		(empty-closure (and customizable (zero? (lambda-literal-closure-size ll))))
+		(nec (- n (if empty-closure 1 0)))
+		(vlist0 (make-variable-list n "t"))
+		(alist0 (make-argument-list n "t"))
+		(varlist (intersperse (if empty-closure (cdr vlist0) vlist0) #\,))
+		(arglist (intersperse (if empty-closure (cdr alist0) alist0) #\,))
+		(external (lambda-literal-external ll))
+		(looping (lambda-literal-looping ll))
+		(direct (lambda-literal-direct ll))
+		(rest-mode (lambda-literal-rest-argument-mode ll))
+		(temps (lambda-literal-temporaries ll))
+		(ubtemps (lambda-literal-unboxed-temporaries ll))
+		(topname (if unit-name
 			     (string-append unit-name "_toplevel")
-			     "toplevel") ] )
+			     "toplevel") ) )
 	   (when empty-closure (debugging 'o "dropping unused closure argument" id))
 	   (gen #t #t)
 	   (gen "/* " (cleanup rname) " */" #t)
@@ -759,10 +785,15 @@
 	   (gen #t "C_word tmp;")
 	   (if rest
 	       (gen #t "C_word t" n #\;) ; To hold rest-list if demand is met
-	       (do ([i n (add1 i)]
-		    [j (+ temps (if looping (sub1 n) 0)) (sub1 j)] )
-		   ((zero? j))
-		 (gen #t "C_word t" i #\;) ) )
+	       (begin
+		 (do ([i n (add1 i)]
+		      [j (+ temps (if looping (sub1 n) 0)) (sub1 j)] )
+		     ((zero? j))
+		   (gen #t "C_word t" i #\;) )
+		 (for-each
+		  (lambda (ubt)
+		    (gen #t (utype (cdr ubt)) #\space (car ubt) #\;))
+		  ubtemps)))
 	   (cond [(eq? 'toplevel id) 
 		  (let ([ldemand (fold (lambda (lit n) (+ n (literal-size lit))) 0 literals)]
 			[llen (length literals)] )
@@ -820,6 +851,8 @@
 			 (unless direct (gen #t "C_word *a;"))
 			 (when looping (gen #t "loop:")) 
 			 (when (and direct (not unsafe) (not disable-stack-overflow-checking))
+			   ;;XXX this can be lifted out of the loop, if the procedure
+			   ;;    does not allocate (suggested by Benedikt Rosenau):
 			   (gen #t "C_stack_check;") ) ] )
 		  (when (and external (not unsafe) (not no-argc-checks) (not customizable))
 		    ;; (not customizable) implies empty-closure
@@ -842,7 +875,6 @@
 ;; 		   [(= nec 1) (gen #t "C_save(" (if empty-closure "t1" "t0") ");")] )
 	     (cond [rest
 		    (gen #t (if (> nec 0) "C_save_and_reclaim" "C_reclaim") "((void*)tr" n #\r)
-		    (when (eq? rest-mode 'vector) (gen #\v))
 		    (gen ",(void*)" id "r")
 		    (when (> nec 0)
 		      (gen #\, nec #\,)
@@ -850,9 +882,7 @@
 		    (gen ");}"
 			 #t "else{"
 			 #t "a=C_alloc((c-" n ")*3);")
-		    (case rest-mode
-		      [(list #f) (gen #t "t" n "=C_restore_rest(a,C_rest_count(0));")]
-		      [(vector) (gen #t "t" n "=C_restore_rest_vector(a,C_rest_count(0));")] )
+		    (gen #t "t" n "=C_restore_rest(a,C_rest_count(0));")
 		    (gen #t id "r(")
 		    (apply gen (intersperse (make-argument-list n "t") #\,))
 		    (gen ",t" n ");}}")
@@ -1277,7 +1307,7 @@
 	       ((const) (foreign-argument-conversion (cadr type)))
 	       ((enum) "C_num_to_int(")
 	       ((ref)
-		(string-append "*(" (foreign-type-declaration (car type) "*")
+		(string-append "*(" (foreign-type-declaration (cadr type) "*")
 			       ")C_c_pointer_nn("))
 	       ((instance-ref)
 		(string-append "*(" (cadr type) "*)C_c_pointer_nn("))
