@@ -1,7 +1,7 @@
 ;;;; c-platform.scm - Platform specific parameters and definitions
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -63,15 +63,13 @@
      (bound-to-procedure
        ##sys#profile-entry ##sys#profile-exit) ) ) )
 
-(define units-used-by-default '(library eval data-structures ports extras srfi-69)) 
+(define units-used-by-default '(library eval)) 
 (define words-per-flonum 4)
 (define parameter-limit 1024)
 (define small-parameter-limit 128)
 (define unlikely-variables '(unquote unquote-splicing))
 
 (define eq-inline-operator "C_eqp")
-(define optimizable-rest-argument-operators 
-  '(car cadr caddr cadddr length pair? null? list-ref))
 (define membership-test-operators
   '(("C_i_memq" . "C_eqp") ("C_u_i_memq" . "C_eqp") ("C_i_member" . "C_i_equalp")
     ("C_i_memv" . "C_i_eqvp") ) )
@@ -79,23 +77,24 @@
 (define target-include-file "chicken.h")
 
 (define valid-compiler-options
-  '(-help h help version verbose explicit-use 
-	  quiet 			; DEPRECATED
-	  no-trace no-warnings unsafe block 
+  '(-help 
+    h help version verbose explicit-use 
+    no-trace no-warnings unsafe block 
     check-syntax to-stdout no-usual-integrations case-insensitive no-lambda-info 
     profile inline keep-shadowed-macros ignore-repository
     fixnum-arithmetic disable-interrupts optimize-leaf-routines
     lambda-lift compile-syntax tag-pointers accumulate-profile
-    disable-stack-overflow-checks disable-c-syntax-checks unsafe-libraries raw 
+    disable-stack-overflow-checks raw 
     emit-external-prototypes-first release local inline-global
     analyze-only dynamic scrutinize no-argc-checks no-procedure-checks
     no-bound-checks no-procedure-checks-for-usual-bindings no-compiler-syntax
     no-parentheses-synonyms no-symbol-escape r5rs-syntax emit-all-import-libraries
-    setup-mode) )
+    setup-mode unboxing) )
 
 (define valid-compiler-options-with-argument
-  '(debug output-file include-path heap-size stack-size unit uses keyword-style require-extension 
-	  inline-limit profile-name disable-warning parenthesis-synonyms
+  '(debug 
+    output-file include-path heap-size stack-size unit uses keyword-style require-extension 
+    inline-limit profile-name disable-warning parenthesis-synonyms
     prelude postlude prologue epilogue nursery extend feature types
     emit-import-library emit-inline-file static-extension consult-inline-file
     heap-growth heap-shrinkage heap-initial-size ffi-define ffi-include-path) )
@@ -126,7 +125,9 @@
 (define default-extended-bindings
   '(bitwise-and bitwise-ior bitwise-xor bitwise-not add1 sub1 fx+ fx- fx* fx/ fxmod o
     fx= fx> fx< fx>= fx<= fixnum? fxneg fxmax fxmin identity fp+ fp- fp* fp/ fpmin fpmax fpneg
-    fp> fp< fp= fp>= fp<= fxand fxnot fxior fxxor fxshr fxshl bit-set?
+    fp> fp< fp= fp>= fp<= fxand fxnot fxior fxxor fxshr fxshl bit-set? fxodd? fxeven?
+    fpfloor fpceiling fptruncate fpround fpsin fpcos fptan fpasin fpacos fpatan
+    fpatan2 fpexp fpexpt fplog fpsqrt fpabs fpinteger?
     arithmetic-shift void flush-output thread-specific thread-specific-set!
     not-pair? atom? null-list? print print* error cpu-time proper-list? call/cc
     blob-size u8vector->blob/shared s8vector->blob/shared u16vector->blob/shared
@@ -141,11 +142,16 @@
     u8vector-length s8vector-length u16vector-length s16vector-length u32vector-length s32vector-length
     f32vector-length f64vector-length setter
     u8vector-ref s8vector-ref u16vector-ref s16vector-ref u32vector-ref s32vector-ref
-    f32vector-ref f64vector-ref
+    f32vector-ref f64vector-ref f32vector-set! f64vector-set!
     u8vector-set! s8vector-set! u16vector-set! s16vector-set! u32vector-set! s32vector-set!
     locative-ref locative-set! locative->object locative? global-ref
-    null-pointer? pointer->object flonum? finite?
-    printf sprintf format) )
+    null-pointer? pointer->object flonum? finite? address->pointer pointer->address
+    pointer+ pointer=?
+    pointer-u8-ref pointer-s8-ref pointer-u16-ref pointer-s16-ref
+    pointer-u32-ref pointer-s32-ref pointer-f32-ref pointer-f64-ref
+    pointer-u8-set! pointer-s8-set! pointer-u16-set! pointer-s16-set!
+    pointer-u32-set! pointer-s32-set! pointer-f32-set! pointer-f64-set!
+    printf sprintf format get-keyword) )
 
 (define internal-bindings
   '(##sys#slot ##sys#setslot ##sys#block-ref ##sys#block-set!
@@ -157,7 +163,7 @@
     ##sys#fudge ##sys#immediate? ##sys#direct-return ##sys#context-switch
     ##sys#make-structure ##sys#apply ##sys#apply-values ##sys#continuation-graft
     ##sys#bytevector? ##sys#make-vector ##sys#setter ##sys#car ##sys#cdr ##sys#pair?
-    ##sys#eq? ##sys#list? ##sys#vector? ##sys#eqv?
+    ##sys#eq? ##sys#list? ##sys#vector? ##sys#eqv? ##sys#get-keyword
     ##sys#foreign-char-argument ##sys#foreign-fixnum-argument ##sys#foreign-flonum-argument
     ##sys#foreign-block-argument ##sys#foreign-number-vector-argument
     ##sys#foreign-string-argument ##sys#foreign-pointer-argument ##sys#void
@@ -168,7 +174,7 @@
 (define non-foldable-bindings
   '(vector
     cons list string make-vector make-string string->symbol values current-input-port current-output-port
-    read-char write-char printf fprintf
+    read-char write-char printf fprintf format
     apply call-with-current-continuation set-car! set-cdr! write-char newline write display
     peek-char char-ready?
     read read-char for-each map string-set! vector-set! string-fill! vector-fill! open-input-file
@@ -178,10 +184,12 @@
     u8vector->blob/shared s8vector->blob/shared u16vector->blob/shared s16vector->blob/shared u32vector->blob/shared
     f32vector->blob/shared f64vector->blob/shared
     s32vector->blob/shared read-string read-string! o
+    address->pointer pointer->address
     ##sys#make-structure print* ##sys#make-vector ##sys#apply ##sys#setislot ##sys#block-ref
-    ##sys#byte ##sys#setbyte 
+    ##sys#byte ##sys#setbyte ##sys#get-keyword get-keyword
     u8vector-length s8vector-length u16vector-length s16vector-length u32vector-length s32vector-length
     f32vector-length f64vector-length ##sys#apply-values ##sys#setter setter
+    f32vector-set! f64vector-set!
     u8vector-ref s8vector-ref u16vector-ref s16vector-ref u32vector-ref s32vector-ref
     u8vector-set! s8vector-set! u16vector-set! s16vector-set! u32vector-set! s32vector-set!
     ##sys#intern-symbol ##sys#make-symbol make-record-instance error cpu-time ##sys#block-set!) )
@@ -314,6 +322,27 @@
 	     '##core#call '(#t)
 	     (cons* (make-node '##core#proc '("C_quotient" #t) '()) cont callargs) ) ) ) ) )
 
+(let ()
+  ;; (add1 <x>) -> (##core#inline "C_fixnum_increase" <x>)     [fixnum-mode]
+  ;; (add1 <x>) -> (##core#inline "C_u_fixnum_increase" <x>)   [fixnum-mode + unsafe]
+  ;; (add1 <x>) -> (##core#inline_allocate ("C_a_i_plus" 4) <x> 1) 
+  ;; (sub1 <x>) -> (##core#inline "C_fixnum_decrease" <x>)     [fixnum-mode]
+  ;; (sub1 <x>) -> (##core#inline "C_u_fixnum_decrease" <x>)   [fixnum-mode + unsafe]
+  ;; (sub1 <x>) -> (##core#inline_allocate ("C_a_i_minus" 4) <x> 1) 
+  (define ((op1 fiop ufiop aiop) db classargs cont callargs)
+    (and (= (length callargs) 1)
+	 (make-node
+	  '##core#call '(#t)
+	  (list 
+	   cont
+	   (if (eq? 'fixnum number-type)
+	       (make-node '##core#inline (list (if unsafe ufiop fiop)) callargs)
+	       (make-node
+		'##core#inline_allocate (list aiop 4)
+		(list (car callargs) (qnode 1))))))))
+  (rewrite 'add1 8 (op1 "C_fixnum_increase" "C_u_fixnum_increase" "C_a_i_plus"))
+  (rewrite 'sub1 8 (op1 "C_fixnum_decrease" "C_u_fixnum_decrease" "C_a_i_minus")))
+
 (let ()
   (define (eqv?-id db classargs cont callargs)
     ;; (eqv? <var> <var>) -> (quote #t)
@@ -482,133 +511,135 @@
 (rewrite 'locative-ref 13 "C_locative_ref" #t)
 (rewrite '##sys#continuation-graft 13 "C_continuation_graft" #t)
 
-(rewrite 'caar 2 1 "C_u_i_caar" #f #f)
-(rewrite 'cdar 2 1 "C_u_i_cdar" #f #f)
-(rewrite 'cddr 2 1 "C_u_i_cddr" #f #f)
-(rewrite 'caaar 2 1 "C_u_i_caaar" #f #f)
-(rewrite 'cadar 2 1 "C_u_i_cadar" #f #f)
-(rewrite 'caddr 2 1 "C_u_i_caddr" #f #f)
-(rewrite 'cdaar 2 1 "C_u_i_cdaar" #f #f)
-(rewrite 'cdadr 2 1 "C_u_i_cdadr" #f #f)
-(rewrite 'cddar 2 1 "C_u_i_cddar" #f #f)
-(rewrite 'cdddr 2 1 "C_u_i_cdddr" #f #f)
-(rewrite 'caaaar 2 1 "C_u_i_caaaar" #f #f)
-(rewrite 'caadar 2 1 "C_u_i_caadar" #f #f)
-(rewrite 'caaddr 2 1 "C_u_i_caaddr" #f #f)
-(rewrite 'cadaar 2 1 "C_u_i_cadaar" #f #f)
-(rewrite 'cadadr 2 1 "C_u_i_cadadr" #f #f)
-(rewrite 'caddar 2 1 "C_u_i_caddar" #f #f)
-(rewrite 'cadddr 2 1 "C_u_i_cadddr" #f #f)
-(rewrite 'cdaaar 2 1 "C_u_i_cdaaar" #f #f)
-(rewrite 'cdaadr 2 1 "C_u_i_cdaadr" #f #f)
-(rewrite 'cdadar 2 1 "C_u_i_cdadar" #f #f)
-(rewrite 'cdaddr 2 1 "C_u_i_cdaddr" #f #f)
-(rewrite 'cddaar 2 1 "C_u_i_cddaar" #f #f)
-(rewrite 'cddadr 2 1 "C_u_i_cddadr" #f #f)
-(rewrite 'cdddar 2 1 "C_u_i_cdddar" #f #f)
-(rewrite 'cddddr 2 1 "C_u_i_cddddr" #f #f)
+(rewrite 'caar 2 1 "C_u_i_caar" #f)
+(rewrite 'cdar 2 1 "C_u_i_cdar" #f)
+(rewrite 'cddr 2 1 "C_u_i_cddr" #f)
+(rewrite 'caaar 2 1 "C_u_i_caaar" #f)
+(rewrite 'cadar 2 1 "C_u_i_cadar" #f)
+(rewrite 'caddr 2 1 "C_u_i_caddr" #f)
+(rewrite 'cdaar 2 1 "C_u_i_cdaar" #f)
+(rewrite 'cdadr 2 1 "C_u_i_cdadr" #f)
+(rewrite 'cddar 2 1 "C_u_i_cddar" #f)
+(rewrite 'cdddr 2 1 "C_u_i_cdddr" #f)
+(rewrite 'caaaar 2 1 "C_u_i_caaaar" #f)
+(rewrite 'caadar 2 1 "C_u_i_caadar" #f)
+(rewrite 'caaddr 2 1 "C_u_i_caaddr" #f)
+(rewrite 'cadaar 2 1 "C_u_i_cadaar" #f)
+(rewrite 'cadadr 2 1 "C_u_i_cadadr" #f)
+(rewrite 'caddar 2 1 "C_u_i_caddar" #f)
+(rewrite 'cadddr 2 1 "C_u_i_cadddr" #f)
+(rewrite 'cdaaar 2 1 "C_u_i_cdaaar" #f)
+(rewrite 'cdaadr 2 1 "C_u_i_cdaadr" #f)
+(rewrite 'cdadar 2 1 "C_u_i_cdadar" #f)
+(rewrite 'cdaddr 2 1 "C_u_i_cdaddr" #f)
+(rewrite 'cddaar 2 1 "C_u_i_cddaar" #f)
+(rewrite 'cddadr 2 1 "C_u_i_cddadr" #f)
+(rewrite 'cdddar 2 1 "C_u_i_cdddar" #f)
+(rewrite 'cddddr 2 1 "C_u_i_cddddr" #f)
 
-(rewrite 'cddr 2 1 "C_i_cddr" #t #f)
-(rewrite 'cdddr 2 1 "C_i_cdddr" #t #f)
-(rewrite 'cddddr 2 1 "C_i_cddddr" #t #f)
+(rewrite 'cddr 2 1 "C_i_cddr" #t)
+(rewrite 'cdddr 2 1 "C_i_cdddr" #t)
+(rewrite 'cddddr 2 1 "C_i_cddddr" #t)
 
 (rewrite 'cdr 7 1 "C_slot" 1 #f)
-(rewrite 'cdr 2 1 "C_i_cdr" #t #f)
+(rewrite 'cdr 2 1 "C_i_cdr" #t)
 
 (rewrite 'eq? 1 2 "C_eqp")
 (rewrite '##sys#eq? 1 2 "C_eqp")
 (rewrite 'eqv? 1 2 "C_i_eqvp")
 (rewrite '##sys#eqv? 1 2 "C_i_eqvp")
 
-(rewrite 'list-ref 2 2 "C_u_i_list_ref" #f "C_slot")
-(rewrite 'list-ref 2 2 "C_i_list_ref" #t "C_i_vector_ref")
-(rewrite 'null? 2 1 "C_i_nullp" #t "C_vemptyp")
-(rewrite '##sys#null? 2 1 "C_i_nullp" #t "C_vemptyp")
-(rewrite 'length 2 1 "C_i_length" #t "C_block_size")
-(rewrite 'not 2 1 "C_i_not" #t #f)
-(rewrite 'char? 2 1 "C_charp" #t #f)
-(rewrite 'string? 2 1 "C_i_stringp" #t #f)
-(rewrite 'locative? 2 1 "C_i_locativep" #t #f)
-(rewrite 'symbol? 2 1 "C_i_symbolp" #t #f)
-(rewrite 'vector? 2 1 "C_i_vectorp" #t #f)
-(rewrite '##sys#vector? 2 1 "C_i_vectorp" #t #f)
-(rewrite 'pair? 2 1 "C_i_pairp" #t "C_notvemptyp")
-(rewrite '##sys#pair? 2 1 "C_i_pairp" #t "C_notvemptyp")
-(rewrite 'procedure? 2 1 "C_i_closurep" #t #f)
-(rewrite 'port? 2 1 "C_i_portp" #t #f)
-(rewrite 'boolean? 2 1 "C_booleanp" #t #f)
-(rewrite 'number? 2 1 "C_i_numberp" #t #f)
-(rewrite 'complex? 2 1 "C_i_numberp" #t #f)
-(rewrite 'rational? 2 1 "C_i_rationalp" #t #f)
-(rewrite 'real? 2 1 "C_i_numberp" #t #f)
-(rewrite 'integer? 2 1 "C_i_integerp" #t #f)
-(rewrite 'flonum? 2 1 "C_i_flonump" #t #f)
-(rewrite 'fixnum? 2 1 "C_fixnump" #t #f)
-(rewrite 'finite? 2 1 "C_i_finitep" #f #f)
-(rewrite '##sys#pointer? 2 1 "C_anypointerp" #t #f)
-(rewrite '##sys#generic-structure? 2 1 "C_structurep" #t #f)
-(rewrite 'exact? 2 1 "C_fixnump" #f #f)
-(rewrite 'exact? 2 1 "C_i_exactp" #t #f)
-(rewrite 'exact? 2 1 "C_u_i_exactp" #f #f)
-(rewrite 'inexact? 2 1 "C_nfixnump" #f #f)
-(rewrite 'inexact? 2 1 "C_i_inexactp" #t #f)
-(rewrite 'inexact? 2 1 "C_u_i_inexactp" #f #f)
-(rewrite 'list? 2 1 "C_i_listp" #t #f)
-(rewrite 'proper-list? 2 1 "C_i_listp" #t #f)
-(rewrite 'eof-object? 2 1 "C_eofp" #t #f)
-(rewrite 'string-ref 2 2 "C_subchar" #f #f)
-(rewrite 'string-ref 2 2 "C_i_string_ref" #t #f)
-(rewrite 'string-set! 2 3 "C_setsubchar" #f #f)
-(rewrite 'string-set! 2 3 "C_i_string_set" #t #f)
-(rewrite 'vector-ref 2 2 "C_slot" #f #f)
-(rewrite 'vector-ref 2 2 "C_i_vector_ref" #t #f)
-(rewrite 'char=? 2 2 "C_eqp" #t #f)
-(rewrite 'char>? 2 2 "C_fixnum_greaterp" #t #f)
-(rewrite 'char<? 2 2 "C_fixnum_lessp" #t #f)
-(rewrite 'char>=? 2 2 "C_fixnum_greater_or_equal_p" #t #f)
-(rewrite 'char<=? 2 2 "C_fixnum_less_or_equal_p" #t #f)
-(rewrite '##sys#slot 2 2 "C_slot" #t #f)		; consider as safe, the primitive is unsafe anyway.
-(rewrite '##sys#block-ref 2 2 "C_i_block_ref" #t #f) ;*** must be safe for pattern matcher (anymore?)
-(rewrite '##sys#size 2 1 "C_block_size" #t #f)
-(rewrite 'fxnot 2 1 "C_fixnum_not" #t #f)
-(rewrite 'fx* 2 2 "C_fixnum_times" #t #f)
-(rewrite 'fx/ 2 2 "C_fixnum_divide" #f #f)
-(rewrite 'fxmod 2 2 "C_fixnum_modulo" #f #f)
-(rewrite 'fx= 2 2 "C_eqp" #t #f)
-(rewrite 'fx> 2 2 "C_fixnum_greaterp" #t #f)
-(rewrite 'fx< 2 2 "C_fixnum_lessp" #t #f)
-(rewrite 'fx>= 2 2 "C_fixnum_greater_or_equal_p" #t #f)
-(rewrite 'fx<= 2 2 "C_fixnum_less_or_equal_p" #t #f)
-(rewrite 'fp= 2 2 "C_flonum_equalp" #t #f)
-(rewrite 'fp> 2 2 "C_flonum_greaterp" #t #f)
-(rewrite 'fp< 2 2 "C_flonum_lessp" #t #f)
-(rewrite 'fp>= 2 2 "C_flonum_greater_or_equal_p" #t #f)
-(rewrite 'fp<= 2 2 "C_flonum_less_or_equal_p" #t #f)
-(rewrite 'fxmax 2 2 "C_i_fixnum_max" #t #f)
-(rewrite 'fxmin 2 2 "C_i_fixnum_min" #t #f)
-(rewrite 'fpmax 2 2 "C_i_flonum_max" #t #f)
-(rewrite 'fpmin 2 2 "C_i_flonum_min" #t #f)
-(rewrite 'char-numeric? 2 1 "C_u_i_char_numericp" #t #f)
-(rewrite 'char-alphabetic? 2 1 "C_u_i_char_alphabeticp" #t #f)
-(rewrite 'char-whitespace? 2 1 "C_u_i_char_whitespacep" #t #f)
-(rewrite 'char-upper-case? 2 1 "C_u_i_char_upper_casep" #t #f)
-(rewrite 'char-lower-case? 2 1 "C_u_i_char_lower_casep" #t #f)
-(rewrite 'char-upcase 2 1 "C_u_i_char_upcase" #t #f)
-(rewrite 'char-downcase 2 1 "C_u_i_char_downcase" #t #f)
-(rewrite 'list-tail 2 2 "C_i_list_tail" #t #f)
-(rewrite '##sys#structure? 2 2 "C_i_structurep" #t #f)
-(rewrite '##sys#bytevector? 2 2 "C_bytevectorp" #t #f)
-(rewrite 'block-ref 2 2 "C_slot" #f #f)	; ok to be unsafe, lolevel is anyway
-(rewrite 'number-of-slots 2 1 "C_block_size" #f #f)
+(rewrite 'list-ref 2 2 "C_u_i_list_ref" #f)
+(rewrite 'list-ref 2 2 "C_i_list_ref" #t)
+(rewrite 'null? 2 1 "C_i_nullp" #t)
+(rewrite '##sys#null? 2 1 "C_i_nullp" #t)
+(rewrite 'length 2 1 "C_i_length" #t)
+(rewrite 'not 2 1 "C_i_not"#t )
+(rewrite 'char? 2 1 "C_charp" #t)
+(rewrite 'string? 2 1 "C_i_stringp" #t)
+(rewrite 'locative? 2 1 "C_i_locativep" #t)
+(rewrite 'symbol? 2 1 "C_i_symbolp" #t)
+(rewrite 'vector? 2 1 "C_i_vectorp" #t)
+(rewrite '##sys#vector? 2 1 "C_i_vectorp" #t)
+(rewrite 'pair? 2 1 "C_i_pairp" #t)
+(rewrite '##sys#pair? 2 1 "C_i_pairp" #t)
+(rewrite 'procedure? 2 1 "C_i_closurep" #t)
+(rewrite 'port? 2 1 "C_i_portp" #t)
+(rewrite 'boolean? 2 1 "C_booleanp" #t)
+(rewrite 'number? 2 1 "C_i_numberp" #t)
+(rewrite 'complex? 2 1 "C_i_numberp" #t)
+(rewrite 'rational? 2 1 "C_i_rationalp" #t)
+(rewrite 'real? 2 1 "C_i_numberp" #t)
+(rewrite 'integer? 2 1 "C_i_integerp" #t)
+(rewrite 'flonum? 2 1 "C_i_flonump" #t)
+(rewrite 'fixnum? 2 1 "C_fixnump" #t)
+(rewrite 'finite? 2 1 "C_i_finitep" #f)
+(rewrite 'fpinteger? 2 1 "C_u_i_fpintegerp" #f)
+(rewrite '##sys#pointer? 2 1 "C_anypointerp" #t)
+(rewrite 'pointer? 2 1 "C_i_safe_pointerp" #t)
+(rewrite '##sys#generic-structure? 2 1 "C_structurep" #t)
+(rewrite 'exact? 2 1 "C_fixnump" #f)
+(rewrite 'exact? 2 1 "C_i_exactp" #t)
+(rewrite 'exact? 2 1 "C_u_i_exactp" #f)
+(rewrite 'inexact? 2 1 "C_nfixnump" #f)
+(rewrite 'inexact? 2 1 "C_i_inexactp" #t)
+(rewrite 'inexact? 2 1 "C_u_i_inexactp" #f)
+(rewrite 'list? 2 1 "C_i_listp" #t)
+(rewrite 'proper-list? 2 1 "C_i_listp" #t)
+(rewrite 'eof-object? 2 1 "C_eofp" #t)
+(rewrite 'string-ref 2 2 "C_subchar" #f)
+(rewrite 'string-ref 2 2 "C_i_string_ref" #t)
+(rewrite 'string-set! 2 3 "C_setsubchar" #f)
+(rewrite 'string-set! 2 3 "C_i_string_set" #t)
+(rewrite 'vector-ref 2 2 "C_slot" #f)
+(rewrite 'vector-ref 2 2 "C_i_vector_ref" #t)
+(rewrite 'char=? 2 2 "C_eqp" #t)
+(rewrite 'char>? 2 2 "C_fixnum_greaterp" #t)
+(rewrite 'char<? 2 2 "C_fixnum_lessp" #t)
+(rewrite 'char>=? 2 2 "C_fixnum_greater_or_equal_p" #t)
+(rewrite 'char<=? 2 2 "C_fixnum_less_or_equal_p" #t)
+(rewrite '##sys#slot 2 2 "C_slot" #t)		; consider as safe, the primitive is unsafe anyway.
+(rewrite '##sys#block-ref 2 2 "C_i_block_ref" #t) ;*** must be safe for pattern matcher (anymore?)
+(rewrite '##sys#size 2 1 "C_block_size" #t)
+(rewrite 'fxnot 2 1 "C_fixnum_not" #t)
+(rewrite 'fx* 2 2 "C_fixnum_times" #t)
+(rewrite 'fx/ 2 2 "C_fixnum_divide" #f)
+(rewrite 'fxmod 2 2 "C_fixnum_modulo" #f)
+(rewrite 'fx= 2 2 "C_eqp" #t)
+(rewrite 'fx> 2 2 "C_fixnum_greaterp" #t)
+(rewrite 'fx< 2 2 "C_fixnum_lessp" #t)
+(rewrite 'fx>= 2 2 "C_fixnum_greater_or_equal_p" #t)
+(rewrite 'fx<= 2 2 "C_fixnum_less_or_equal_p" #t)
+(rewrite 'fp= 2 2 "C_flonum_equalp" #f)
+(rewrite 'fp> 2 2 "C_flonum_greaterp" #f)
+(rewrite 'fp< 2 2 "C_flonum_lessp" #f)
+(rewrite 'fp>= 2 2 "C_flonum_greater_or_equal_p" #f)
+(rewrite 'fp<= 2 2 "C_flonum_less_or_equal_p" #f)
+(rewrite 'fxmax 2 2 "C_i_fixnum_max" #f)
+(rewrite 'fxmin 2 2 "C_i_fixnum_min" #f)
+(rewrite 'fpmax 2 2 "C_i_flonum_max" #f)
+(rewrite 'fpmin 2 2 "C_i_flonum_min" #f)
+(rewrite 'char-numeric? 2 1 "C_u_i_char_numericp" #t)
+(rewrite 'char-alphabetic? 2 1 "C_u_i_char_alphabeticp" #t)
+(rewrite 'char-whitespace? 2 1 "C_u_i_char_whitespacep" #t)
+(rewrite 'char-upper-case? 2 1 "C_u_i_char_upper_casep" #t)
+(rewrite 'char-lower-case? 2 1 "C_u_i_char_lower_casep" #t)
+(rewrite 'char-upcase 2 1 "C_u_i_char_upcase" #t)
+(rewrite 'char-downcase 2 1 "C_u_i_char_downcase" #t)
+(rewrite 'list-tail 2 2 "C_i_list_tail" #t)
+(rewrite '##sys#structure? 2 2 "C_i_structurep" #t)
+(rewrite '##sys#bytevector? 2 2 "C_bytevectorp" #t)
+(rewrite 'block-ref 2 2 "C_slot" #f)	; ok to be unsafe, lolevel is anyway
+(rewrite 'number-of-slots 2 1 "C_block_size" #f)
 
 (rewrite 'assv 14 'fixnum 2 "C_i_assq" "C_u_i_assq")
-(rewrite 'assv 2 2 "C_i_assv" #t #f)
+(rewrite 'assv 2 2 "C_i_assv" #t)
 (rewrite 'memv 14 'fixnum 2 "C_i_memq" "C_u_i_memq")
-(rewrite 'memv 2 2 "C_i_memv" #t #f)
+(rewrite 'memv 2 2 "C_i_memv" #t)
 (rewrite 'assq 17 2 "C_i_assq" "C_u_i_assq")
 (rewrite 'memq 17 2 "C_i_memq" "C_u_i_memq")
-(rewrite 'assoc 2 2 "C_i_assoc" #t #f)
-(rewrite 'member 2 2 "C_i_member" #t #f)
+(rewrite 'assoc 2 2 "C_i_assoc" #t)
+(rewrite 'member 2 2 "C_i_member" #t)
 
 (rewrite 'set-car! 4 '##sys#setslot 0)
 (rewrite 'set-cdr! 4 '##sys#setslot 1)
@@ -642,47 +673,47 @@
 (rewrite 'atan 16 2 "C_a_i_atan2" #t words-per-flonum)
 
 (rewrite 'zero? 5 "C_eqp" 0 'fixnum)
-(rewrite 'zero? 2 1 "C_i_zerop" #t #f)
-(rewrite 'zero? 2 1 "C_u_i_zerop" #f #f)
+(rewrite 'zero? 2 1 "C_i_zerop" #t)
+(rewrite 'zero? 2 1 "C_u_i_zerop" #f)
 (rewrite 'positive? 5 "C_fixnum_greaterp" 0 'fixnum)
 (rewrite 'positive? 5 "C_flonum_greaterp" 0 'flonum)
-(rewrite 'positive? 2 1 "C_i_positivep" #t #f)
-(rewrite 'positive? 2 1 "C_u_i_positivep" #f #f)
+(rewrite 'positive? 2 1 "C_i_positivep" #t)
+(rewrite 'positive? 2 1 "C_u_i_positivep" #f)
 (rewrite 'negative? 5 "C_fixnum_lessp" 0 'fixnum)
 (rewrite 'negative? 5 "C_flonum_lessp" 0 'flonum)
-(rewrite 'negative? 2 1 "C_i_negativep" #t #f)
-(rewrite 'negative? 2 1 "C_u_i_negativep" #f #f)
+(rewrite 'negative? 2 1 "C_i_negativep" #t)
+(rewrite 'negative? 2 1 "C_u_i_negativep" #f)
 
 (rewrite 'vector-length 6 "C_fix" "C_header_size" #f)
 (rewrite 'string-length 6 "C_fix" "C_header_size" #f)
 (rewrite 'char->integer 6 "C_fix" "C_character_code" #t)
 (rewrite 'integer->char 6 "C_make_character" "C_unfix" #t)
 
-(rewrite 'vector-length 2 1 "C_i_vector_length" #t #f)
-(rewrite '##sys#vector-length 2 1 "C_i_vector_length" #t #f)
-(rewrite 'string-length 2 1 "C_i_string_length" #t #f)
-(rewrite 'inexact->exact 2 1 "C_i_inexact_to_exact" #t #f)
+(rewrite 'vector-length 2 1 "C_i_vector_length" #t)
+(rewrite '##sys#vector-length 2 1 "C_i_vector_length" #t)
+(rewrite 'string-length 2 1 "C_i_string_length" #t)
+(rewrite 'inexact->exact 2 1 "C_i_inexact_to_exact" #t)
 
-(rewrite '##sys#check-exact 2 1 "C_i_check_exact" #t #f)
-(rewrite '##sys#check-number 2 1 "C_i_check_number" #t #f)
-(rewrite '##sys#check-list 2 1 "C_i_check_list" #t #f)
-(rewrite '##sys#check-pair 2 1 "C_i_check_pair" #t #f)
-(rewrite '##sys#check-symbol 2 1 "C_i_check_symbol" #t #f)
-(rewrite '##sys#check-string 2 1 "C_i_check_string" #t #f)
-(rewrite '##sys#check-byte-vector 2 1 "C_i_check_bytevector" #t #f)
-(rewrite '##sys#check-vector 2 1 "C_i_check_vector" #t #f)
-(rewrite '##sys#check-structure 2 2 "C_i_check_structure" #t #f)
-(rewrite '##sys#check-char 2 1 "C_i_check_char" #t #f)
-(rewrite '##sys#check-exact 2 2 "C_i_check_exact_2" #t #f)
-(rewrite '##sys#check-number 2 2 "C_i_check_number_2" #t #f)
-(rewrite '##sys#check-list 2 2 "C_i_check_list_2" #t #f)
-(rewrite '##sys#check-pair 2 2 "C_i_check_pair_2" #t #f)
-(rewrite '##sys#check-symbol 2 2 "C_i_check_symbol_2" #t #f)
-(rewrite '##sys#check-string 2 2 "C_i_check_string_2" #t #f)
-(rewrite '##sys#check-byte-vector 2 2 "C_i_check_bytevector_2" #t #f)
-(rewrite '##sys#check-vector 2 2 "C_i_check_vector_2" #t #f)
-(rewrite '##sys#check-structure 2 3 "C_i_check_structure_2" #t #f)
-(rewrite '##sys#check-char 2 2 "C_i_check_char_2" #t #f)
+(rewrite '##sys#check-exact 2 1 "C_i_check_exact" #t)
+(rewrite '##sys#check-number 2 1 "C_i_check_number" #t)
+(rewrite '##sys#check-list 2 1 "C_i_check_list" #t)
+(rewrite '##sys#check-pair 2 1 "C_i_check_pair" #t)
+(rewrite '##sys#check-symbol 2 1 "C_i_check_symbol" #t)
+(rewrite '##sys#check-string 2 1 "C_i_check_string" #t)
+(rewrite '##sys#check-byte-vector 2 1 "C_i_check_bytevector" #t)
+(rewrite '##sys#check-vector 2 1 "C_i_check_vector" #t)
+(rewrite '##sys#check-structure 2 2 "C_i_check_structure" #t)
+(rewrite '##sys#check-char 2 1 "C_i_check_char" #t)
+(rewrite '##sys#check-exact 2 2 "C_i_check_exact_2" #t)
+(rewrite '##sys#check-number 2 2 "C_i_check_number_2" #t)
+(rewrite '##sys#check-list 2 2 "C_i_check_list_2" #t)
+(rewrite '##sys#check-pair 2 2 "C_i_check_pair_2" #t)
+(rewrite '##sys#check-symbol 2 2 "C_i_check_symbol_2" #t)
+(rewrite '##sys#check-string 2 2 "C_i_check_string_2" #t)
+(rewrite '##sys#check-byte-vector 2 2 "C_i_check_bytevector_2" #t)
+(rewrite '##sys#check-vector 2 2 "C_i_check_vector_2" #t)
+(rewrite '##sys#check-structure 2 3 "C_i_check_structure_2" #t)
+(rewrite '##sys#check-char 2 2 "C_i_check_char_2" #t)
 
 (rewrite '= 9 "C_eqp" "C_i_equalp" #t #t)
 (rewrite '> 9 "C_fixnum_greaterp" "C_flonum_greaterp" #t #f)
@@ -702,7 +733,7 @@
 (rewrite 'list->string 11 1 '##sys#list->string #t)
 
 (rewrite 'vector-set! 11 3 '##sys#setslot #f)
-(rewrite 'vector-set! 2 3 "C_i_vector_set" #t #f)
+(rewrite 'vector-set! 2 3 "C_i_vector_set" #t)
 
 (rewrite '##sys#vector->list 11 1 'vector->list #t)
 (rewrite '##sys#list->vector 11 1 'list->vector #t)
@@ -743,10 +774,6 @@
 (rewrite 'string->number 13 "C_string_to_number" #t)
 (rewrite 'number->string 13 "C_number_to_string" #t)
 (rewrite '##sys#call-with-current-continuation 13 "C_call_cc" #t)
-(rewrite '##sys#floor 13 "C_flonum_floor" #t)
-(rewrite '##sys#ceiling 13 "C_flonum_ceiling" #t)
-(rewrite '##sys#truncate 13 "C_flonum_truncate" #t)
-(rewrite '##sys#round 13 "C_flonum_round" #t)
 (rewrite '##sys#allocate-vector 13 "C_allocate_vector" #t)
 (rewrite '##sys#ensure-heap-reserve 13 "C_ensure_heap_reserve" #t)
 (rewrite 'return-to-host 13 "C_return_to_host" #t)
@@ -756,19 +783,35 @@
 
 (rewrite 'even? 14 'fixnum 1 "C_i_fixnumevenp" "C_i_fixnumevenp")
 (rewrite 'odd? 14 'fixnum 1 "C_i_fixnumoddp" "C_i_fixnumoddp")
-(rewrite 'add1 14 'fixnum 1 "C_fixnum_increase" "C_u_fixnum_increase")
-(rewrite 'sub1 14 'fixnum 1 "C_fixnum_decrease" "C_u_fixnum_decrease")
 (rewrite 'remainder 14 'fixnum 2 "C_fixnum_modulo" "C_fixnum_modulo")
 
-(rewrite 'even? 2 1 "C_i_evenp" #t #f)
-(rewrite 'even? 2 1 "C_u_i_evenp" #f #f)
-(rewrite 'odd? 2 1 "C_i_oddp" #t #f)
-(rewrite 'odd? 2 1 "C_u_i_oddp" #f #f)
+(rewrite 'even? 17 1 "C_i_evenp" "C_u_i_evenp")
+(rewrite 'odd? 17 1 "C_i_oddp" "C_u_i_oddp")
 
-(rewrite 'floor 15 'flonum 'fixnum '##sys#floor #f)
-(rewrite 'ceiling 15 'flonum 'fixnum '##sys#ceiling #f)
-(rewrite 'truncate 15 'flonum 'fixnum '##sys#truncate #f)
-(rewrite 'round 15 'flonum 'fixnum '##sys#round #f)
+(rewrite 'fxodd? 2 1 "C_fixnumoddp" #t)
+(rewrite 'fxeven? 2 1 "C_fixnumevenp" #t)
+
+(rewrite 'floor 15 'flonum 'fixnum 'fpfloor #f)
+(rewrite 'ceiling 15 'flonum 'fixnum 'fpceiling #f)
+(rewrite 'truncate 15 'flonum 'fixnum 'fptruncate #f)
+(rewrite 'round 15 'flonum 'fixnum 'fpround #f)
+
+(rewrite 'fpsin 16 1 "C_a_i_flonum_sin" #f words-per-flonum)
+(rewrite 'fpcos 16 1 "C_a_i_flonum_cos" #f words-per-flonum)
+(rewrite 'fptan 16 1 "C_a_i_flonum_tan" #f words-per-flonum)
+(rewrite 'fpasin 16 1 "C_a_i_flonum_asin" #f words-per-flonum)
+(rewrite 'fpacos 16 1 "C_a_i_flonum_acos" #f words-per-flonum)
+(rewrite 'fpatan 16 1 "C_a_i_flonum_atan" #f words-per-flonum)
+(rewrite 'fpatan2 16 2 "C_a_i_flonum_atan2" #f words-per-flonum)
+(rewrite 'fpexp 16 1 "C_a_i_flonum_exp" #f words-per-flonum)
+(rewrite 'fpexpt 16 2 "C_a_i_flonum_expt" #f words-per-flonum)
+(rewrite 'fplog 16 1 "C_a_i_flonum_log" #f words-per-flonum)
+(rewrite 'fpsqrt 16 1 "C_a_i_flonum_sqrt" #f words-per-flonum)
+(rewrite 'fpabs 16 1 "C_a_i_flonum_abs" #f words-per-flonum)
+(rewrite 'fptruncate 16 1 "C_a_i_flonum_truncate" #f words-per-flonum)
+(rewrite 'fpround 16 1 "C_a_i_flonum_truncate" #f words-per-flonum)
+(rewrite 'fpceiling 16 1 "C_a_i_flonum_truncate" #f words-per-flonum)
+(rewrite 'fpround 16 1 "C_a_i_flonum_truncate" #f words-per-flonum)
 
 (rewrite 'cons 16 2 "C_a_i_cons" #t 3)
 (rewrite '##sys#cons 16 2 "C_a_i_cons" #t 3)
@@ -778,6 +821,27 @@
 (rewrite '##sys#vector 16 #f "C_a_i_vector" #t #t)
 (rewrite '##sys#make-structure 16 #f "C_a_i_record" #t #t)
 (rewrite 'string 16 #f "C_a_i_string" #t #t) ; the last #t is actually too much, but we don't care
+(rewrite 'address->pointer 16 1 "C_a_i_address_to_pointer" #f 2)
+(rewrite 'pointer->address 16 1 "C_a_i_pointer_to_address" #f words-per-flonum)
+(rewrite 'pointer+ 16 2 "C_a_u_i_pointer_inc" #f 2)
+
+(rewrite 'pointer-u8-ref 2 1 "C_u_i_pointer_u8_ref" #f)
+(rewrite 'pointer-s8-ref 2 1 "C_u_i_pointer_s8_ref" #f)
+(rewrite 'pointer-u16-ref 2 1 "C_u_i_pointer_u16_ref" #f)
+(rewrite 'pointer-s16-ref 2 1 "C_u_i_pointer_s16_ref" #f)
+(rewrite 'pointer-u8-set! 2 2 "C_u_i_pointer_u8_set" #f)
+(rewrite 'pointer-s8-set! 2 2 "C_u_i_pointer_s8_set" #f)
+(rewrite 'pointer-u16-set! 2 2 "C_u_i_pointer_u16_set" #f)
+(rewrite 'pointer-s16-set! 2 2 "C_u_i_pointer_s16_set" #f)
+(rewrite 'pointer-u32-set! 2 2 "C_u_i_pointer_u32_set" #f)
+(rewrite 'pointer-s32-set! 2 2 "C_u_i_pointer_s32_set" #f)
+(rewrite 'pointer-f32-set! 2 2 "C_u_i_pointer_f32_set" #f)
+(rewrite 'pointer-f64-set! 2 2 "C_u_i_pointer_f64_set" #f)
+
+(rewrite 'pointer-u32-ref 16 1 "C_a_u_i_pointer_u32_ref" #f words-per-flonum)
+(rewrite 'pointer-s32-ref 16 1 "C_a_u_i_pointer_s32_ref" #f words-per-flonum)
+(rewrite 'pointer-f32-ref 16 1 "C_a_u_i_pointer_f32_ref" #f words-per-flonum)
+(rewrite 'pointer-f64-ref 16 1 "C_a_u_i_pointer_f64_ref" #f words-per-flonum)
 
 (rewrite
  '##sys#setslot 8
@@ -866,31 +930,36 @@
 (rewrite '##sys#foreign-unsigned-integer-argument 17 1 "C_i_foreign_unsigned_integer_argumentp")
 (rewrite '##sys#direct-return 17 2 "C_direct_return")
 
-(rewrite 'blob-size 2 1 "C_block_size" #f #f)
+(rewrite 'blob-size 2 1 "C_block_size" #f)
 
-(rewrite 'u8vector-ref 2 2 "C_u_i_u8vector_ref" #f #f)
-(rewrite 's8vector-ref 2 2 "C_u_i_s8vector_ref" #f #f)
-(rewrite 'u16vector-ref 2 2 "C_u_i_u16vector_ref" #f #f)
-(rewrite 's16vector-ref 2 2 "C_u_i_s16vector_ref" #f #f)
+(rewrite 'u8vector-ref 2 2 "C_u_i_u8vector_ref" #f)
+(rewrite 's8vector-ref 2 2 "C_u_i_s8vector_ref" #f)
+(rewrite 'u16vector-ref 2 2 "C_u_i_u16vector_ref" #f)
+(rewrite 's16vector-ref 2 2 "C_u_i_s16vector_ref" #f)
+
+(rewrite 'f32vector-ref 16 2 "C_a_i_f32vector_ref" #f words-per-flonum)
+(rewrite 'f64vector-ref 16 2 "C_a_i_f64vector_ref" #f words-per-flonum)
 
 (rewrite 'u32vector-ref 22 2 "C_a_i_u32vector_ref" #f words-per-flonum "C_u_i_u32vector_ref")
 (rewrite 's32vector-ref 22 2 "C_a_i_s32vector_ref" #f words-per-flonum "C_u_i_s32vector_ref")
 
-(rewrite 'u8vector-set! 2 3 "C_u_i_u8vector_set" #f #f)
-(rewrite 's8vector-set! 2 3 "C_u_i_s8vector_set" #f #f)
-(rewrite 'u16vector-set! 2 3 "C_u_i_u16vector_set" #f #f)
-(rewrite 's16vector-set! 2 3 "C_u_i_s16vector_set" #f #f)
-(rewrite 'u32vector-set! 2 3 "C_u_i_u32vector_set" #f #f)
-(rewrite 's32vector-set! 2 3 "C_u_i_s32vector_set" #f #f)
+(rewrite 'u8vector-set! 2 3 "C_u_i_u8vector_set" #f)
+(rewrite 's8vector-set! 2 3 "C_u_i_s8vector_set" #f)
+(rewrite 'u16vector-set! 2 3 "C_u_i_u16vector_set" #f)
+(rewrite 's16vector-set! 2 3 "C_u_i_s16vector_set" #f)
+(rewrite 'u32vector-set! 2 3 "C_u_i_u32vector_set" #f)
+(rewrite 's32vector-set! 2 3 "C_u_i_s32vector_set" #f)
+(rewrite 'f32vector-set! 2 3 "C_u_i_f32vector_set" #f)
+(rewrite 'f64vector-set! 2 3 "C_u_i_f64vector_set" #f)
 
-(rewrite 'u8vector-length 2 1 "C_u_i_8vector_length" #f #f)
-(rewrite 's8vector-length 2 1 "C_u_i_8vector_length" #f #f)
-(rewrite 'u16vector-length 2 1 "C_u_i_16vector_length" #f #f)
-(rewrite 's16vector-length 2 1 "C_u_i_16vector_length" #f #f)
-(rewrite 'u32vector-length 2 1 "C_u_i_32vector_length" #f #f)
-(rewrite 's32vector-length 2 1 "C_u_i_32vector_length" #f #f)
-(rewrite 'f32vector-length 2 1 "C_u_i_32vector_length" #f #f)
-(rewrite 'f64vector-length 2 1 "C_u_i_64vector_length" #f #f)
+(rewrite 'u8vector-length 2 1 "C_u_i_8vector_length" #f)
+(rewrite 's8vector-length 2 1 "C_u_i_8vector_length" #f)
+(rewrite 'u16vector-length 2 1 "C_u_i_16vector_length" #f)
+(rewrite 's16vector-length 2 1 "C_u_i_16vector_length" #f)
+(rewrite 'u32vector-length 2 1 "C_u_i_32vector_length" #f)
+(rewrite 's32vector-length 2 1 "C_u_i_32vector_length" #f)
+(rewrite 'f32vector-length 2 1 "C_u_i_32vector_length" #f)
+(rewrite 'f64vector-length 2 1 "C_u_i_64vector_length" #f)
 
 (rewrite 'not-pair? 17 1 "C_i_not_pair_p")
 (rewrite 'atom? 17 1 "C_i_not_pair_p")
@@ -953,6 +1022,7 @@
 			    (let ([var (or rest (second llist))])
 			      (and (not (get db var 'references))
 				   (not (get db var 'assigned)) 
+				   (not (get db var 'inline-transient))
 				   (make-node
 				    '##core#call '(#t)
 				    (list val cont (qnode #f)) ) ) ) ) ) ) ) ) ) ) ) )
@@ -1038,3 +1108,6 @@
 (rewrite 'substring-ci=? 23 2 '##sys#substring-ci=? 0 0 #f)
 (rewrite 'substring-index 23 2 '##sys#substring-index 0)
 (rewrite 'substring-index-ci 23 2 '##sys#substring-index-ci 0)
+
+(rewrite 'get-keyword 7 2 "C_i_get_keyword" #f #t)
+(rewrite '##sys#get-keyword 7 2 "C_i_get_keyword" #f #t)
