@@ -1,7 +1,7 @@
 ;;;; chicken-ffi-syntax.scm
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -34,15 +34,14 @@
  (##sys#er-transformer
   (lambda (form r c)
     (let* ((form (cdr form))
-	   (%quote (r 'quote))
 	   (quals (and (pair? form) (string? (car form))))
 	   (var (and (not quals) (pair? form) (symbol? (car form)))) )
       (cond [var
 	     (##sys#check-syntax 'define-external form '(symbol _ . #(_ 0 1)))
 	     (let ([var (car form)])
 	       `(,(r 'begin)
-		 (,(r 'define-foreign-variable) ,var ,(cadr form))
-		 (,(r 'define-external-variable) ,var ,(cadr form) #t)
+		 (##core#define-foreign-variable ,var ,(cadr form))
+		 (##core#define-external-variable ,var ,(cadr form) #t)
 		 ,@(if (pair? (cddr form))
 		       `((##core#set! ,var ,(caddr form)))
 		       '() ) ) ) ]
@@ -66,6 +65,14 @@
 
 ;;; External locations:
 
+(##sys#extend-macro-environment
+ 'location
+ '()
+ (##sys#er-transformer
+  (lambda (x r c)
+    (##sys#check-syntax 'location x '(location _))
+    `(##core#location ,(cadr x)))))
+
 (##sys#extend-macro-environment
  'define-location
  '()
@@ -77,8 +84,8 @@
 	  (init (optional (cdddr form) #f))
 	  (name (r (gensym))))
       `(,(r 'begin)
-	(,(r 'define-foreign-variable) ,var ,type ,(symbol->string name))
-	(,(r 'define-external-variable) ,var ,type #f ,name)
+	(##core#define-foreign-variable ,var ,type ,(symbol->string name))
+	(##core#define-external-variable ,var ,type #f ,name)
 	,@(if (pair? init)
 	      `((##core#set! ,var ,(car init)))
 	      '() ) ) ) ) ) )
@@ -91,9 +98,8 @@
     (##sys#check-syntax 'let-location form '(_ #((variable _ . #(_ 0 1)) 0) . _))
     (let* ((bindings (cadr form))
 	   (body (cddr form))
-	   (%let (r 'let))
 	   [aliases (map (lambda (_) (r (gensym))) bindings)])
-      `(,%let ,(append-map
+      `(##core#let ,(append-map
 		(lambda (b a)
 		  (if (pair? (cddr b))
 		      (list (cons a (cddr b)))
@@ -111,7 +117,7 @@
 			,(car b)
 			,(cadr b)
 			,rest) ) )
-		`(,%let () ,@body)
+		`(##core#let () ,@body)
 		bindings aliases) ) ) ) ) )
 
 
@@ -124,7 +130,7 @@
   (lambda (form r c)
     (##sys#check-syntax 'foreign-code form '(_ . #(string 0)))
     (let ([tmp (gensym 'code_)])
-      `(,(r 'begin)
+      `(##core#begin
 	 (,(r 'declare)
 	  (foreign-declare
 	   ,(sprintf "static C_word ~A() { ~A\n; return C_SCHEME_UNDEFINED; }\n" 
@@ -140,8 +146,8 @@
     (##sys#check-syntax 'foreign-value form '(_ _ _))
     (let ((tmp (gensym 'code_))
 	  (code (cadr form)))
-      `(,(r 'begin)
-	(,(r 'define-foreign-variable) ,tmp
+      `(##core#begin
+	(##core#define-foreign-variable ,tmp
 	 ,(caddr form)
 	 ,(cond ((string? code) code)
 		((symbol? code) (symbol->string code))
@@ -160,4 +166,56 @@
     `(##core#declare (foreign-declare ,@(cdr form))))))
 
 
+;;; Aliases for internal forms
+
+(##sys#extend-macro-environment
+ 'define-foreign-type
+ '()
+ (##sys#er-transformer
+  (lambda (form r c)
+    `(##core#define-foreign-type ,@(cdr form)))))
+
+(##sys#extend-macro-environment
+ 'define-foreign-variable
+ '()
+ (##sys#er-transformer
+  (lambda (form r c)
+    `(##core#define-foreign-variable ,@(cdr form)))))
+
+(##sys#extend-macro-environment
+ 'foreign-primitive
+ '()
+ (##sys#er-transformer
+  (lambda (form r c)
+    `(##core#foreign-primitive ,@(cdr form)))))
+
+(##sys#extend-macro-environment
+ 'foreign-lambda
+ '()
+ (##sys#er-transformer
+  (lambda (form r c)
+    `(##core#foreign-lambda ,@(cdr form)))))
+
+(##sys#extend-macro-environment
+ 'foreign-lambda*
+ '()
+ (##sys#er-transformer
+  (lambda (form r c)
+    `(##core#foreign-lambda* ,@(cdr form)))))
+
+(##sys#extend-macro-environment
+ 'foreign-safe-lambda
+ '()
+ (##sys#er-transformer
+  (lambda (form r c)
+    `(##core#foreign-safe-lambda ,@(cdr form)))))
+
+(##sys#extend-macro-environment
+ 'foreign-safe-lambda*
+ '()
+ (##sys#er-transformer
+  (lambda (form r c)
+    `(##core#foreign-safe-lambda* ,@(cdr form)))))
+
+
 (##sys#macro-subset me0)))
