@@ -1,7 +1,7 @@
 ;;;; chicken.scm - The CHICKEN Scheme compiler (loader/main-module)
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -26,15 +26,16 @@
 
 
 (declare
-  (uses chicken-syntax srfi-1 srfi-4 utils files support compiler optimizer scrutinizer driver 
-	platform backend srfi-69)
+  (uses chicken-syntax srfi-1 srfi-4 utils files extras data-structures support
+	compiler optimizer unboxing compiler-syntax scrutinizer driver platform backend 
+	srfi-69)
   (compile-syntax) )			
 
 
 (include "compiler-namespace")
 (include "tweaks")
 
-(eval-when (load) 
+(eval-when (load)
   (include "chicken-ffi-syntax") )
 
 
@@ -84,14 +85,22 @@
 		   [(1)
 		    (set! options (cons 'optimize-leaf-routines options)) ]
 		   [(2)
-		    (set! options (cons 'optimize-leaf-routines options)) ] 
+		    (set! options 
+		      (cons* 'optimize-leaf-routines 'inline options)) ] 
 		   [(3)
 		    (set! options
-		      (cons* 'optimize-leaf-routines 'local 'inline options) ) ]
+		      (cons* 'optimize-leaf-routines 'inline 'local options) ) ]
 		   [(4)
 		    (set! options
-		      (cons* 'optimize-leaf-routines 'local 'inline 'unsafe options) ) ]
-		   [else (compiler-warning 'usage "invalid optimization level ~S - ignored" (car rest))] )
+		      (cons* 'optimize-leaf-routines 'inline 'local 'unboxing 'unsafe
+			     options) ) ]
+		   [else
+		    (when (>= level 5)
+		      (set! options 
+			(cons* 'disable-interrupts 'no-trace 'unsafe 'block
+			       'optimize-leaf-routines 'lambda-lift 'no-lambda-info
+			       'inline 'unboxing
+			       options) ) ) ] )
 		 (loop (cdr rest)) ) ]
 	      [(eq? 'debug-level o)
 	       (let ([level (string->number (car rest))])
@@ -101,13 +110,6 @@
 		   [(2) #f]
 		   [else (compiler-warning 'usage "invalid debug level ~S - ignored" (car rest))] )
 		 (loop (cdr rest)) ) ]
-	      [(eq? 'benchmark-mode o)
-	       (set! options 
-		 (cons* 'fixnum-arithmetic 'disable-interrupts 'no-trace 'unsafe
-			'optimize-leaf-routines 'block 'lambda-lift 'no-lambda-info
-			'inline
-			options) )
-	       (loop rest) ]
 	      [(memq o valid-compiler-options) (loop rest)]
 	      [(memq o valid-compiler-options-with-argument)
 	       (if (pair? rest)
