@@ -4,9 +4,9 @@
 ; "This is insane. What we clearly want to do is not exactly clear, and is rooted in NCOMPLR."
 ;
 ;
-;-----------------------------------------------------------------------------------------------------------
+;--------------------------------------------------------------------------------------------------
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -98,22 +98,22 @@
 ; (##core#declare {<spec>})
 ; (##core#immutable <exp>)
 ; (##core#global-ref <variable>)
-; (quote <exp>)
-; (if <exp> <exp> [<exp>])
-; ([##core#]syntax <exp>)
-; ([##core#]let <variable> ({(<variable> <exp>)}) <body>)
-; ([##core#]let ({(<variable> <exp>)}) <body>)
-; ([##core#]letrec ({(<variable> <exp>)}) <body>)
+; (##core#quote <exp>)
+; (##core#syntax <exp>)
+; (##core#if <exp> <exp> [<exp>])
+; (##core#let <variable> ({(<variable> <exp>)}) <body>)
+; (##core#let ({(<variable> <exp>)}) <body>)
+; (##core#letrec ({(<variable> <exp>)}) <body>)
 ; (##core#let-location <symbol> <type> [<init>] <exp>)
-; ([##core#]lambda <variable> <body>)
-; ([##core#]lambda ({<variable>}+ [. <variable>]) <body>)
-; ([##core#]set! <variable> <exp>)
-; ([##core#]begin <exp> ...)
-; (##core#named-lambda <name> <llist> <body>)
+; (##core#lambda <variable> <body>)
+; (##core#lambda ({<variable>}+ [. <variable>]) <body>)
+; (##core#set! <variable> <exp>)
+; (##core#begin <exp> ...)
+; (##core#include <string>)
 ; (##core#loop-lambda <llist> <body>)
 ; (##core#undefined)
 ; (##core#primitive <name>)
-; (##core#inline <op> {<exp>})
+; (##core#inline {<op>} <exp>)
 ; (##core#inline_allocate (<op> <words>) {<exp>})
 ; (##core#inline_ref (<name> <type>))
 ; (##core#inline_update (<name> <type>) <exp>)
@@ -123,30 +123,26 @@
 ; (##core#compiletimeonly <exp>)
 ; (##core#elaborationtimetoo <exp>)
 ; (##core#elaborationtimeonly <exp>)
-; (define-foreign-variable <symbol> <type> [<string>])
-; (define-foreign-type <symbol> <type> [<proc1> [<proc2>]])
-; (foreign-lambda <type> <string> {<type>})
-; (foreign-lambda* <type> ({(<type> <var>)})) {<string>})
-; (foreign-safe-lambda <type> <string> {<type>})
-; (foreign-safe-lambda* <type> ({(<type> <var>)})) {<string>})
-; (foreign-primitive <type> ({(<type> <var>)}) {<string>})
+; (##core#define-foreign-variable <symbol> <type> [<string>])
+; (##core#define-foreign-type <symbol> <type> [<proc1> [<proc2>]])
+; (##core#foreign-lambda <type> <string> {<type>})
+; (##core#foreign-lambda* <type> ({(<type> <var>)})) {<string>})
+; (##core#foreign-safe-lambda <type> <string> {<type>})
+; (##core#foreign-safe-lambda* <type> ({(<type> <var>)})) {<string>})
+; (##core#foreign-primitive <type> ({(<type> <var>)}) {<string>})
 ; (##core#define-inline <name> <exp>)
-; (define-constant <name> <exp>)
+; (##core#define-constant <name> <exp*>)
 ; (##core#foreign-callback-wrapper '<name> <qualifiers> '<type> '({<type>}) <exp>)
-; (##core#define-external-variable (quote <name>) (quote <type>) (quote <bool>))
+; (##core#define-external-variable <name> <type> <bool> [<symbol>])
 ; (##core#check <exp>)
 ; (##core#require-for-syntax <exp> ...)
 ; (##core#require-extension (<id> ...) <bool>)
 ; (##core#app <exp> {<exp>})
-; ([##core#]syntax <exp>)
-; (<exp> {<exp>})
-; (define-syntax <symbol> <expr>)
-; (define-syntax (<symbol> . <llist>) <expr> ...)
-; (define-compiled-syntax <symbol> <expr>)
-; (define-compiled-syntax (<symbol> . <llist>) <expr> ...)
+; (##core#define-syntax <symbol> <expr>)
 ; (##core#define-compiler-syntax <symbol> <expr>)
 ; (##core#let-compiler-syntax ((<symbol> <expr>) ...) <expr> ...)
 ; (##core#module <symbol> #t | (<name> | (<name> ...) ...) <body>)
+; (<exp> {<exp>})
 
 ; - Core language:
 ;
@@ -179,13 +175,17 @@
 ; [if {} <exp> <exp> <exp>]
 ; [quote {<exp>}]
 ; [##core#bind {<count>} <exp-v>... <exp>]
+; [##core#let_unboxed {<name> <utype>} <exp1> <exp2>]
 ; [##core#undefined {}]
+; [##core#unboxed_ref {<name> <utype>}]
+; [##core#unboxed_set! {<name> <utype>} <exp>]
 ; [##core#inline {<op>} <exp>...]
 ; [##core#inline_allocate {<op <words>} <exp>...]
 ; [##core#inline_ref {<name> <type>}]
 ; [##core#inline_update {<name> <type>} <exp>]
 ; [##core#inline_loc_ref {<type>} <exp>]
 ; [##core#inline_loc_update {<type>} <exp> <exp>]
+; [##core#inline_unboxed {<op>} <exp> ...]
 ; [##core#closure {<count>} <exp>...]
 ; [##core#box {} <exp>]
 ; [##core#unbox {} <exp>]
@@ -235,10 +235,10 @@
 ;   standard-binding -> <boolean>            If true: variable names a standard binding
 ;   extended-binding -> <boolean>            If true: variable names an extended binding
 ;   unused -> <boolean>                      If true: variable is a formal parameter that is never used
-;   rest-parameter -> #f | 'vector | 'list   If true: variable holds rest-argument list mode
-;   o-r/access-count -> <n>                  Contains number of references as arguments of optimizable rest operators
+;   rest-parameter -> #f | 'list             If true: variable holds rest-argument list
 ;   constant -> <boolean>                    If true: variable has fixed value
 ;   hidden-refs -> <boolean>                 If true: procedure that refers to hidden global variables
+;   inline-transient -> <boolean>            If true: was introduced during inlining
 ; 
 ; <lambda-id>:
 ;
@@ -251,6 +251,7 @@
 ;   simple -> <boolean>                      If true: procedure only calls its continuation
 ;   explicit-rest -> <boolean>               If true: procedure is called with consed rest list
 ;   captured-variables -> (<var> ...)        List of closed over variables
+;   inline-target -> <boolean>               If true: was target of an inlining operation
 
 
 (declare
@@ -318,7 +319,6 @@
 (define explicit-use-flag #f)
 (define disable-stack-overflow-checking #f)
 (define require-imports-flag #f)
-(define emit-unsafe-marker #f)
 (define external-protos-first #f)
 (define do-lambda-lifting #f)
 (define inline-max-size default-inline-max-size)
@@ -327,6 +327,7 @@
 (define constant-declarations '())
 (define profiled-procedures #f)
 (define import-libraries '())
+(define all-import-libraries #f)
 (define standalone-executable #t)
 (define local-definitions #f)
 (define inline-globally #f)
@@ -350,7 +351,6 @@
 (define current-program-size 0)
 (define line-number-database-2 #f)
 (define immutable-constants '())
-(define rest-parameters-promoted-to-vector '())
 (define inline-table #f)
 (define inline-table-used #f)
 (define constant-table #f)
@@ -364,7 +364,6 @@
 (define foreign-lambda-stubs '())
 (define foreign-callback-stubs '())
 (define external-variables '())
-(define loop-lambda-names '())
 (define profile-lambda-list '())
 (define profile-lambda-index 0)
 (define profile-info-vector-name #f)
@@ -408,6 +407,7 @@
 ;;; Expand macros and canonicalize expressions:
 
 (define (canonicalize-expression exp)
+  (let ((compiler-syntax '()))
 
   (define (find-id id se)		; ignores macro bindings
     (cond ((null? se) #f)
@@ -481,6 +481,29 @@
 	(##sys#current-meta-environment))
        '() ) ))
 
+  (define (emit-import-lib name il)
+    (let* ((fname (if all-import-libraries
+		      (string-append (symbol->string name) ".import.scm")
+		      (cdr il)))
+	   (imps (##sys#compiled-module-registration (##sys#current-module)))
+	   (oldimps
+	    (and (file-exists? fname)
+		 (read-file fname) ) ) )
+      (cond ((equal? imps oldimps)
+	     (when verbose-mode
+	       (print "not generating import library `" fname "' for module `" 
+		      name "' because imports did not change")) )
+	    (else
+	     (when verbose-mode
+	       (print "generating import library `" fname "' for module `"
+		      name "' ..."))
+	     (with-output-to-file fname
+	       (lambda ()
+		 (print ";;;; " fname " - GENERATED BY CHICKEN "
+			(chicken-version) " -*- Scheme -*-\n")
+		 (for-each pretty-print imps)
+		 (print "\n;; END OF FILE"))))) ) )
+
   (define (walk x e se dest)
     (cond ((symbol? x)
 	   (cond ((keyword? x) `(quote ,x))
@@ -498,7 +521,7 @@
 	     (emit-syntax-trace-info x #f)
 	     (unless (proper-list? x)
 	       (if ln
-		   (syntax-error (sprintf "(in line ~s) - malformed expression" ln) x)
+		   (syntax-error (sprintf "(~a) - malformed expression" ln) x)
 		   (syntax-error "malformed expression" x)))
 	     (set! ##sys#syntax-error-culprit x)
 	     (let* ((name0 (lookup (car x) se))
@@ -515,8 +538,7 @@
 		      (when ln (update-line-number-database! xexpanded ln))
 		      (case name
 			
-			((if)
-			 (##sys#check-syntax 'if x '(if _ _ . #(_)) #f se)
+			((##core#if)
 			 `(if
 			   ,(walk (cadr x) e se #f)
 			   ,(walk (caddr x) e se #f)
@@ -524,8 +546,7 @@
 				'(##core#undefined)
 				(walk (cadddr x) e se #f) ) ) )
 
-			((quote syntax ##core#syntax)
-			 (##sys#check-syntax name x '(_ _) #f se)
+			((##core#syntax ##core#quote)
 			 `(quote ,(##sys#strip-syntax (cadr x))))
 
 			((##core#check)
@@ -566,11 +587,13 @@
 			((##core#require-extension)
 			 (let ((imp? (caddr x)))
 			   (walk
-			    (let loop ([ids (cadr x)])
+			    (let loop ([ids (##sys#strip-syntax (cadr x))])
 			      (if (null? ids)
 				  '(##core#undefined)
 				  (let ([id (car ids)])
-				    (let-values ([(exp f) (##sys#do-the-right-thing id #t imp?)])
+				    (let-values ([(exp f)
+						  (##sys#do-the-right-thing
+						   id #t imp?)])
 				      (unless (or f 
 						  (and (symbol? id)
 						       (or (feature? id)
@@ -583,8 +606,7 @@
 				      `(##core#begin ,exp ,(loop (cdr ids))) ) ) ) )
 			    e se dest) ) )
 
-			((let ##core#let)
-			 (##sys#check-syntax 'let x '(_ #((variable _) 0) . #(_ 1)) #f se)
+			((##core#let)
 			 (let* ((bindings (cadr x))
 				(vars (unzip1 bindings))
 				(aliases (map gensym vars))
@@ -598,8 +620,7 @@
 				    (append aliases e)
 				    se2 dest) ) ) )
 
-			((letrec ##core#letrec)
-			 (##sys#check-syntax 'letrec x '(_ #((symbol _) 0) . #(_ 1)))
+			((##core#letrec)
 			 (let ((bindings (cadr x))
 			       (body (cddr x)) )
 			   (walk
@@ -613,8 +634,7 @@
 			      (##core#let () ,@body) )
 			    e se dest)))
 
-			((lambda ##core#lambda)
-			 (##sys#check-syntax 'lambda x '(_ lambda-list . #(_ 1)) #f se)
+			((##core#lambda)
 			 (let ((llist (cadr x))
 			       (obody (cddr x)) )
 			   (when (##sys#extended-lambda-list? llist)
@@ -633,34 +653,25 @@
 				      (build-lambda-list
 				       aliases argc
 				       (and rest (list-ref aliases (posq rest vars))) ) )
-				     (l `(lambda ,llist2 ,body)) )
+				     (l `(##core#lambda ,llist2 ,body)) )
 				(set-real-names! aliases vars)
 				(cond ((or (not dest) 
 					   (assq dest se)) ; not global?
 				       l)
-				      ((and (eq? 'lambda (or (lookup name se) name))
-					    emit-profile
+				      ((and emit-profile
 					    (or (eq? profiled-procedures 'all)
 						(and
 						 (eq? profiled-procedures 'some)
-						 (variable-mark dest '##compiler#profile))))
-				       (expand-profile-lambda dest llist2 body) )
-				      (else
-				       (if (and (> (length body0) 1)
-						(symbol? (car body0))
-						(eq? 'begin (or (lookup (car body0) se) (car body0)))
-						(let ((x1 (cadr body0)))
-						  (or (string? x1)
-						      (and (list? x1)
-							   (= (length x1) 2)
-							   (symbol? (car x1))
-							   (eq? 'quote (or (lookup (car x1) se) (car x1)))))))
-					   (process-lambda-documentation
-					    dest (cadr body) l) 
-					   l))))))))
+						 (variable-mark dest '##compiler#profile)))
+					    (##sys#interned-symbol? dest))
+				       (expand-profile-lambda
+					(if (memq dest e) ;XXX should normally not be the case
+					    e
+					    (##sys#alias-global-hook dest #f))
+					llist2 body) )
+				      (else l)))))))
 			
-			((let-syntax)
-			 (##sys#check-syntax 'let-syntax x '(let-syntax #((variable _) 0) . #(_ 1)) #f se)
+			((##core#let-syntax)
 			 (let ((se2 (append
 				     (map (lambda (b)
 					    (list
@@ -675,8 +686,7 @@
 			    e se2
 			    dest) ) )
 			       
-		       ((letrec-syntax)
-			(##sys#check-syntax 'letrec-syntax x '(letrec-syntax #((variable _) 0) . #(_ 1)) #f se)
+		       ((##core#letrec-syntax)
 			(let* ((ms (map (lambda (b)
 					  (list
 					   (car b)
@@ -693,7 +703,7 @@
 			   (##sys#canonicalize-body (cddr x) se2 compiler-syntax-enabled)
 			   e se2 dest)))
 			       
-		       ((define-syntax define-compiled-syntax)
+		       ((##core#define-syntax)
 			(##sys#check-syntax
 			 (car x) x
 			 (if (pair? (cadr x))
@@ -711,8 +721,7 @@
 			   (##sys#current-environment)
 			   (##sys#er-transformer (eval/meta body)))
 			  (walk
-			   (if (or ##sys#enable-runtime-macros
-				   (eq? 'define-compiled-syntax (car x)))
+			   (if ##sys#enable-runtime-macros
 			       `(##sys#extend-macro-environment
 				 ',var
 				 (##sys#current-environment)
@@ -723,36 +732,46 @@
 		       ((##core#define-compiler-syntax)
 			(let* ((var (cadr x))
 			       (body (caddr x))
-			       (name (##sys#strip-syntax var se #t)))
+			       (name (##sys#strip-syntax var se #f)))
+			  (when body
+			    (set! compiler-syntax
+			      (alist-cons
+			       name
+			       (##sys#get name '##compiler#compiler-syntax) compiler-syntax)))
 			  (##sys#put! 
 			   name '##compiler#compiler-syntax
-			   (##sys#cons
-			    (##sys#er-transformer (eval/meta body))
-			    (##sys#current-environment)))
+			   (and body
+				(##sys#cons
+				 (##sys#er-transformer (eval/meta body))
+				 (##sys#current-environment))))
 			  (walk 
 			   (if ##sys#enable-runtime-macros
 			       `(##sys#put! 
 				(##core#syntax ,name)
 				'##compiler#compiler-syntax
-				(##sys#cons
-				 (##sys#er-transformer ,body)
-				 (##sys#current-environment)))
+				,(and body
+				      `(##sys#cons
+					(##sys#er-transformer ,body)
+					(##sys#current-environment))))
 			       '(##core#undefined) )
 			   e se dest)))
 
 		       ((##core#let-compiler-syntax)
-			(let ((bs (map (lambda (b)
-					 (##sys#check-syntax 'let-compiler-syntax b '(symbol _))
-					 (let ((name (##sys#strip-syntax (car b) se #t)))
-					   (list 
-					    name 
-					    (cons (##sys#er-transformer (eval/meta (cadr b))) se)
-					    (##sys#get name '##compiler#compiler-syntax) ) ) )
-				       (cadr x))))
-			  (dynamic-wind	; this ain't thread safe
+			(let ((bs (map
+				   (lambda (b)
+				     (##sys#check-syntax 'let-compiler-syntax b '(symbol . #(_ 0 1)))
+				     (let ((name (##sys#strip-syntax (car b) se #f)))
+				       (list 
+					name 
+					(and (pair? (cdr b))
+					     (cons (##sys#er-transformer (eval/meta (cadr b))) se))
+					(##sys#get name '##compiler#compiler-syntax) ) ) )
+				   (cadr x))))
+			  (dynamic-wind
 			      (lambda ()
 				(for-each
-				 (lambda (b) (##sys#put! (car b) '##compiler#compiler-syntax (cadr b)))
+				 (lambda (b) 
+				   (##sys#put! (car b) '##compiler#compiler-syntax (cadr b)))
 				 bs) )
 			      (lambda ()
 				(walk 
@@ -760,11 +779,20 @@
 				 e se dest) )
 			      (lambda ()
 				(for-each
-				 (lambda (b) (##sys#put! (car b) '##compiler#compiler-syntax (caddr b)))
+				 (lambda (b)
+				   (##sys#put! (car b) '##compiler#compiler-syntax (caddr b)))
 				 bs) ) ) ) )
 
+		       ((##core#include)
+			(walk
+			 `(##core#begin
+			   ,@(fluid-let ((##sys#default-read-info-hook read-info-hook))
+			       (##sys#include-forms-from-file (cadr x))))
+			 e se dest))
+
 		       ((##core#module)
-			(let* ((name (lookup (cadr x) se))
+			(let* ((x (##sys#strip-syntax x))
+			       (name (##sys#strip-syntax (cadr x)))
 			       (exports 
 				(or (eq? #t (caddr x))
 				    (map (lambda (exp)
@@ -779,9 +807,11 @@
 						  (##sys#syntax-error-hook
 						   'module
 						   "invalid export syntax" exp name))))
-					 (##sys#strip-syntax (caddr x))))))
+					 (##sys#strip-syntax (caddr x)))))
+			       (csyntax compiler-syntax))
 			  (when (##sys#current-module)
-			    (##sys#syntax-error-hook 'module "modules may not be nested" name))
+			    (##sys#syntax-error-hook
+			     'module "modules may not be nested" name))
 			  (let-values (((body mreg)
 					(parameterize ((##sys#current-module 
 							(##sys#register-module name exports) )
@@ -796,17 +826,10 @@
 						      (print-error-message ex (current-error-port))
 						      (exit 1))
 						  (##sys#finalize-module (##sys#current-module)))
-						(cond ((assq name import-libraries) =>
+						(cond ((or all-import-libraries
+							   (assq name import-libraries) ) =>
 						       (lambda (il)
-							 (when verbose-mode
-							   (print "generating import library " (cdr il) " for module "
-								  name " ..."))
-							 (with-output-to-file (cdr il)
-							   (lambda ()
-							     (for-each
-							      pretty-print
-							      (##sys#compiled-module-registration
-							       (##sys#current-module))))) 
+							 (emit-import-lib name il)
 							 (values 
 							  (reverse xs)
 							  '((##core#undefined)))))
@@ -817,19 +840,6 @@
 							    '()
 							    (##sys#compiled-module-registration (##sys#current-module)))))))
 					       (else
-						(when (and (pair? body)
-							   (null? xs)
-							   (pair? (car body))
-							   (symbol? (caar body))
-							   (let ((imp (or (lookup (caar body) se) (caar body))))
-							     (and (not (memq imp '(import import-for-syntax)))
-								  ;; can it get any uglier? yes, it can
-								  (not (eq? imp (cdr (assq 'import ##sys#initial-macro-environment))))
-								  (not (eq? imp (cdr (assq 'import-for-syntax ##sys#initial-macro-environment)))))))
-						  (compiler-warning 
-						   'syntax
-						   "module body of `~s' does not begin with `import' form - maybe unintended?"
-						   name))
 						(loop 
 						 (cdr body)
 						 (cons (walk 
@@ -838,23 +848,26 @@
 							(##sys#current-environment)
 							#f)
 						       xs))))))))
-			    (canonicalize-begin-body
-			     (append
-			      (parameterize ((##sys#current-module #f)
-					     (##sys#macro-environment (##sys#meta-macro-environment)))
-				(map
-				 (lambda (x)
-				   (walk 
-				    x 
-				    e 	;?
-				    (##sys#current-meta-environment) #f) )
-				 mreg))
-			      body)))))
+			    (let ((body
+				   (canonicalize-begin-body
+				    (append
+				     (parameterize ((##sys#current-module #f)
+						    (##sys#macro-environment (##sys#meta-macro-environment)))
+				       (map
+					(lambda (x)
+					  (walk 
+					   x 
+					   e 	;?
+					   (##sys#current-meta-environment) #f) )
+					mreg))
+				     body))))
+			      (do ((cs compiler-syntax (cdr cs)))
+				  ((eq? cs csyntax))
+				(##sys#put! (caar cs) '##compiler#compiler-syntax (cdar cs)))
+			      (set! compiler-syntax csyntax)
+			      body))))
 
-		       ((##core#named-lambda)
-			(walk `(##core#lambda ,@(cddr x)) e se (cadr x)) )
-
-		       ((##core#loop-lambda)
+		       ((##core#loop-lambda) ;XXX is this really needed?
 			(let* ([vars (cadr x)]
 			       [obody (cddr x)]
 			       [aliases (map gensym vars)]
@@ -867,8 +880,7 @@
 			  (set-real-names! aliases vars)
 			  `(##core#lambda ,aliases ,body) ) )
 
-			((set! ##core#set!) 
-			 (##sys#check-syntax 'set! x '(_ variable _) #f se)
+			((##core#set!)
 			 (let* ([var0 (cadr x)]
 				[var (lookup var0 se)]
 				[ln (get-line x)]
@@ -910,7 +922,7 @@
 				    (compiler-warning 
 				     'var "assigned global variable `~S' is a macro ~A"
 				     var
-				     (if ln (sprintf "in line ~S" ln) "") )
+				     (if ln (sprintf "(~a)" ln) "") )
 				    (when undefine-shadowed-macros (##sys#undefine-macro! var) ) )
 				  (when (keyword? var)
 				    (compiler-warning 'syntax "assignment to keyword `~S'" var) )
@@ -945,8 +957,7 @@
 			 (eval/meta (cadr x))
 			 '(##core#undefined) )
 
-			((begin ##core#begin) 
-			 (##sys#check-syntax 'begin x '(_ . #(_ 0)) #f se)
+			((##core#begin) 
 			 (if (pair? (cdr x))
 			     (canonicalize-begin-body
 			      (let fold ([xs (cdr x)])
@@ -957,22 +968,22 @@
 				      (cons (walk x e se #f) (fold r)) ) ) ) )
 			     '(##core#undefined) ) )
 
-			((foreign-lambda)
+			((##core#foreign-lambda)
 			 (walk (expand-foreign-lambda x #f) e se dest) )
 
-			((foreign-safe-lambda)
+			((##core#foreign-safe-lambda)
 			 (walk (expand-foreign-lambda x #t) e se dest) )
 
-			((foreign-lambda*)
+			((##core#foreign-lambda*)
 			 (walk (expand-foreign-lambda* x #f) e se dest) )
 
-			((foreign-safe-lambda*)
+			((##core#foreign-safe-lambda*)
 			 (walk (expand-foreign-lambda* x #t) e se dest) )
 
-			((foreign-primitive)
+			((##core#foreign-primitive)
 			 (walk (expand-foreign-primitive x) e se dest) )
 
-			((define-foreign-variable)
+			((##core#define-foreign-variable)
 			 (let* ([var (##sys#strip-syntax (second x))]
 				[type (##sys#strip-syntax (third x))]
 				[name (if (pair? (cdddr x))
@@ -986,7 +997,7 @@
 				   foreign-variables))
 			   '(##core#undefined) ) )
 
-			((define-foreign-type)
+			((##core#define-foreign-type)
 			 (let ([name (second x)]
 			       [type (##sys#strip-syntax (third x))] 
 			       [conv (cdddr x)] )
@@ -1009,7 +1020,7 @@
 				  (##sys#hash-table-set! foreign-type-table name type)
 				  '(##core#undefined) ] ) ) )
 
-			((define-external-variable)
+			((##core#define-external-variable)
 			 (let* ([sym (second x)]
 				[name (symbol->string sym)]
 				[type (third x)] 
@@ -1056,7 +1067,7 @@
 			     (set! inline-table-used #t)
 			     '(##core#undefined)))
 
-			((define-constant)
+			((##core#define-constant)
 			 (let* ([name (second x)]
 				[valexp (third x)]
 				[val (handle-exceptions ex
@@ -1163,37 +1174,32 @@
 					   rtype) ) )
 				      e se #f) ) ) ) )
 
-			(else
-			 (let ([handle-call
-				(lambda ()
-				  (let* ([x2 (mapwalk x e se)]
-					 [head2 (car x2)]
-					 [old (##sys#hash-table-ref line-number-database-2 head2)] )
-				    (when ln
-				      (##sys#hash-table-set!
-				       line-number-database-2
-				       head2
-				       (cons name (alist-cons x2 ln (if old (cdr old) '()))) ) )
-				    x2) ) ] )
-
-			   (cond [(eq? 'location name)
-				  (##sys#check-syntax 'location x '(location _) #f se)
-				  (let ([sym (cadr x)])
-				    (if (symbol? sym)
-					(cond [(assq (lookup sym se) location-pointer-map)
-					       => (lambda (a)
-						    (walk
-						     `(##sys#make-locative ,(second a) 0 #f 'location)
-						     e se #f) ) ]
-					      [(assq sym external-to-pointer) 
-					       => (lambda (a) (walk (cdr a) e se #f)) ]
-					      [(memq sym callback-names)
-					       `(##core#inline_ref (,(symbol->string sym) c-pointer)) ]
-					      [else 
-					       (walk `(##sys#make-locative ,sym 0 #f 'location) e se #f) ] )
-					(walk `(##sys#make-locative ,sym 0 #f 'location) e se #f) ) ) ]
+			((##core#location)
+			 (let ([sym (cadr x)])
+			   (if (symbol? sym)
+			       (cond [(assq (lookup sym se) location-pointer-map)
+				      => (lambda (a)
+					   (walk
+					    `(##sys#make-locative ,(second a) 0 #f 'location)
+					    e se #f) ) ]
+				     [(assq sym external-to-pointer) 
+				      => (lambda (a) (walk (cdr a) e se #f)) ]
+				     [(memq sym callback-names)
+				      `(##core#inline_ref (,(symbol->string sym) c-pointer)) ]
+				     [else 
+				      (walk `(##sys#make-locative ,sym 0 #f 'location) e se #f) ] )
+			       (walk `(##sys#make-locative ,sym 0 #f 'location) e se #f) ) ) )
 				 
-				 [else (handle-call)] ) ) ) ) ] ) ) ) )
+			(else
+			 (let* ([x2 (mapwalk x e se)]
+				[head2 (car x2)]
+				[old (##sys#hash-table-ref line-number-database-2 head2)] )
+			   (when ln
+			     (##sys#hash-table-set!
+			      line-number-database-2
+			      head2
+			      (cons name (alist-cons x2 ln (if old (cdr old) '()))) ) )
+			   x2) ) ) ] ) ) ) )
 
 	  ((not (proper-list? x))
 	   (syntax-error "malformed expression" x) )
@@ -1205,20 +1211,12 @@
 
 	  (else
 	   (emit-syntax-trace-info x #f)
-	   (let ((x (mapwalk x e se)))
-	     (if (and (pair? (car x))
-		      (symbol? (caar x))
-		      (memq (or (lookup (caar x) se) (caar x)) '(lambda ##core#lambda)))
-		 (let ((lexp (car x))
-		       (args (cdr x)) )
-		   (##sys#check-syntax 'lambda lexp '(_ lambda-list . #(_ 1)) #f se)
-		   (let ((llist (cadr lexp)))
-		     (if (and (proper-list? llist) (= (llist-length llist) (length args)))
-			 `(let ,(map list llist args) ,@(cddr lexp)) 
-			 (let ((var (gensym 't)))
-			   `(let ((,var ,(car x)))
-			     (,var ,@(cdr x)) ) ) ) ) ) 
-		 x))) ) )
+	   (let ((tmp (gensym)))
+	     (walk
+	      `(##core#let 
+		((,tmp ,(car x)))
+		(,tmp ,@(cdr x)))
+	      e se dest)))))
   
   (define (mapwalk xs e se)
     (map (lambda (x) (walk x e se #f)) xs) )
@@ -1234,8 +1232,7 @@
      ,(begin
 	(set! extended-bindings (append internal-bindings extended-bindings))
 	exp) )
-   '() (##sys#current-environment)
-   #f) )
+   '() (##sys#current-environment) #f) ) )
 
 
 (define (process-declaration spec se)	; se unused in the moment
@@ -1320,15 +1317,7 @@
 	(let ([fds (cdr spec)])
 	  (if (every string? fds)
 	      (set! foreign-declarations (append foreign-declarations fds))
-	      (syntax-error "invalid declaration" spec) ) ) )
-       ((c-options)
-	(emit-control-file-item `(c-options ,@(strip (cdr spec)))) )
-       ((link-options)
-	(emit-control-file-item `(link-options ,@(strip (cdr spec))) ) )
-       ((post-process)
-	(emit-control-file-item
-	 (let ([file (pathname-strip-extension source-filename)])
-	   `(post-process ,@(map (cut string-substitute "\\$@" file <>) (cdr spec))) ) ) )
+	      (syntax-error 'declare "invalid declaration" spec) ) ) )
        ((block) (set! block-compilation #t))
        ((separate) (set! block-compilation #f))
        ((keep-shadowed-macros) (set! undefine-shadowed-macros #f))
@@ -1377,8 +1366,7 @@
 	       [(interrupts-enabled) (set! insert-timer-checks #f)]
 	       [(safe) (set! unsafe #t)]
 	       [else (compiler-warning 'syntax "illegal declaration specifier `~s'" id)]))]))
-       ((compile-syntax 
-	 run-time-macros)		; DEPRECATED
+       ((compile-syntax )
 	(set! ##sys#enable-runtime-macros #t))
        ((block-global hide) 
 	(let ([syms (stripa (cdr spec))])
@@ -1465,7 +1453,7 @@
      '(##core#undefined) ) ) )
 
 
-;;; Expand "foreign-lambda"/"foreign-callback-lambda" forms and add item to stub-list:
+;;; Expand "foreign-lambda"/"foreign-safe-lambda" forms and add item to stub-list:
 
 (define-record-type foreign-stub
   (make-foreign-stub id return-type name argument-types argument-names body cps callback)
@@ -1486,7 +1474,7 @@
 	 [f-id (gensym 'stub)]
 	 [bufvar (gensym)] 
 	 [rsize (estimate-foreign-result-size rtype)] )
-    (set-real-name! f-id #t)
+    (when sname (set-real-name! f-id (string->symbol sname)))
     (set! foreign-lambda-stubs 
       (cons (make-foreign-stub f-id rtype sname argtypes argnames body cps callback)
 	    foreign-lambda-stubs) )
@@ -1579,16 +1567,17 @@
 	((if) (let* ((t1 (gensym 'k))
 		     (t2 (gensym 'r))
 		     (k1 (lambda (r) (make-node '##core#call '(#t) (list (varnode t1) r)))) )
-		(make-node 'let
-			   (list t1)
-			   (list (make-node '##core#lambda (list (gensym-f-id) #f (list t2) 0) 
-					    (list (k (varnode t2))) )
-				 (walk (car subs)
-				       (lambda (v)
-					 (make-node 'if '()
-						    (list v
-							  (walk (cadr subs) k1)
-							  (walk (caddr subs) k1) ) ) ) ) ) ) ) )
+		(make-node 
+		 'let
+		 (list t1)
+		 (list (make-node '##core#lambda (list (gensym-f-id) #f (list t2) 0) 
+				  (list (k (varnode t2))) )
+		       (walk (car subs)
+			     (lambda (v)
+			       (make-node 'if '()
+					  (list v
+						(walk (cadr subs) k1)
+						(walk (caddr subs) k1) ) ) ) ) ) ) ) )
 	((let)
 	 (let loop ((vars params) (vals subs))
 	   (if (null? vars)
@@ -1706,7 +1695,8 @@
 	     (ref var n)
 	     (unless (memq var localenv)
 	       (grow 1)
-	       (cond ((memq var env) (put! db var 'captured #t))
+	       (cond ((memq var env) 
+		      (put! db var 'captured #t))
 		     ((not (get db var 'global)) 
 		      (put! db var 'global #t) ) ) ) ) )
 	  
@@ -1723,18 +1713,9 @@
 	  ((##core#call)
 	   (grow 1)
 	   (let ([fun (car subs)])
-	     (if (eq? '##core#variable (node-class fun))
-		 (let ([name (first (node-parameters fun))])
-		   (collect! db name 'call-sites (cons here n))
-		   ;; If call to standard-binding & optimizable rest-arg operator: decrease access count:
-		   (if (and (intrinsic? name)
-			    (memq name optimizable-rest-argument-operators) )
-		       (for-each
-			(lambda (arg)
-			  (and-let* ([(eq? '##core#variable (node-class arg))]
-				     [var (first (node-parameters arg))] )
-			    (when (get db var 'rest-parameter) (count! db var 'o-r/access-count)) ) )
-			(cdr subs) ) ) ) )
+	     (when (eq? '##core#variable (node-class fun))
+	       (let ([name (first (node-parameters fun))])
+		 (collect! db name 'call-sites (cons here n))))
 	     (walk (first subs) env localenv here #t)
 	     (walkeach (cdr subs) env localenv here #f) ) )
 
@@ -1750,9 +1731,9 @@
 		     (walk val env localenv here #f) 
 		     (loop (cdr vars) (cdr vals)) ) ) ) ) )
 
-	  ((lambda)
-	   (grow 1)
-	   (decompose-lambda-list
+	  ((lambda)			; why do we have 2 cases for lambda?
+	   (grow 1)			; the reason for this should be tracked down
+	   (decompose-lambda-list	; and this case removed
 	    (first params)
 	    (lambda (vars argc rest)
 	      (for-each 
@@ -1779,10 +1760,7 @@
 		   (put! db var 'unknown #t) )
 		 vars)
 		(when rest
-		  (put! db rest 'rest-parameter
-			(if (memq rest rest-parameters-promoted-to-vector)
-			    'vector
-			    'list) ) )
+		  (put! db rest 'rest-parameter 'list) )
 		(when (simple-lambda-node? n) (put! db id 'simple #t))
 		(let ([tl toplevel-scope])
 		  (unless toplevel-lambda-id (set! toplevel-lambda-id id))
@@ -1803,10 +1781,12 @@
 		 ((extended)
 		  (compiler-warning 'redef "redefinition of extended binding `~S'" var) ) )
 	       (put! db var 'potential-value val) )
-	     (when (and (not (memq var localenv)) 
-			(not (memq var env)) )
+	     (unless (memq var localenv)
 	       (grow 1)
-	       (put! db var 'global #t) )
+	       (cond ((memq var env) 
+		      (put! db var 'captured #t))
+		     ((not (get db var 'global)) 
+		      (put! db var 'global #t) ) ) )
 	     (assign var val (append localenv env) here)
 	     (unless toplevel-scope (put! db var 'assigned-locally #t))
 	     (put! db var 'assigned #t)
@@ -1826,7 +1806,7 @@
     (define (assign var val env here)
       (cond ((eq? '##core#undefined (node-class val))
 	     (put! db var 'undefined #t) )
-	    ((and (eq? '##core#variable (node-class val))
+	    ((and (eq? '##core#variable (node-class val)) ; assignment to itself
 		  (eq? var (first (node-parameters val))) ) )
 	    ((or (memq var env)
 		 (variable-mark var '##compiler#constant)
@@ -1843,9 +1823,11 @@
 		      (variable-mark var '##compiler#local))
 		  (not (get db var 'unknown)))
 	     (let ((home (get db var 'home)))
-	       (if (or (not home) (eq? here home))
-		   (put! db var 'local-value val)	       
-		   (put! db var 'unknown #t))))
+	       (cond ((get db var 'local-value)
+		      (put! db var 'unknown #t))
+		     ((or (not home) (eq? here home))
+		      (put! db var 'local-value val)	       )
+		     (else (put! db var 'unknown #t)))))
 	    (else (put! db var 'unknown #t)) ) )
     
     (define (ref var node)
@@ -1881,7 +1863,6 @@
 	     [assigned-locally #f]
 	     [undefined #f]
 	     [global #f]
-	     [o-r/access-count 0]
 	     [rest-parameter #f] 
 	     [nreferences 0]
 	     [ncall-sites 0] )
@@ -1904,11 +1885,11 @@
 	      [(global) (set! global #t)]
 	      [(value) (set! value (cdr prop))]
 	      [(local-value) (set! local-value (cdr prop))]
-	      [(o-r/access-count) (set! o-r/access-count (cdr prop))]
 	      [(rest-parameter) (set! rest-parameter #t)] ) )
 	  plist)
 
 	 (set! value (and (not unknown) value))
+	 (set! local-value (and (not unknown) local-value))
 
 	 ;; If this is the first analysis, register known local or potentially known global lambda-value id's
 	 ;;  along with their names:
@@ -1928,7 +1909,8 @@
 	     (compiler-warning 'var "local assignment to unused variable `~S' may be unintended" sym) )
 	   (when (and (not (variable-visible? sym))
 		      (not (variable-mark sym '##compiler#constant)) )
-	     (compiler-warning 'var "global variable `~S' is never used" sym) ) )
+	     (##sys#notice 
+	      (sprintf "global variable `~S' is never used" sym) ) ) )
 
  	 ;; Make 'boxed, if 'assigned & 'captured:
 	 (when (and assigned captured)
@@ -1970,7 +1952,6 @@
 			   (mark-variable sym '##compiler#inline-global 'no))
 			  (else
 			   (let ((lparams (node-parameters n)))
-			     (put! db (first lparams) 'simple #t) ;XXX hack
 			     (quick-put! plist 'inlinable #t)
 			     (quick-put! plist 'local-value n))))))))
 
@@ -2067,25 +2048,10 @@
 				    (eq? (first llist) (first (node-parameters v2))) )
 			   (let ([kvar (first (node-parameters v1))])
 			     (quick-put! plist 'replacable kvar)
-			     (put! db kvar 'replacing #t) ) ) ) ) ) ) ) ) ) )
-
-	 ;; If a rest-argument, convert 'rest-parameter property to 'vector, if the variable is never
-	 ;;  assigned, and the number of references is identical to the number of accesses in optimizable
-	 ;;  rest-argument operators:
-	 ;; - Add variable to "rest-parameters-promoted-to-vector", because subsequent optimization will
-	 ;;   change variables context (operators applied to it).
-	 (when (and rest-parameter
-		    (not assigned)
-		    (= nreferences o-r/access-count) )
-	   (set! rest-parameters-promoted-to-vector (lset-adjoin eq? rest-parameters-promoted-to-vector sym))
-	   (put! db sym 'rest-parameter 'vector) ) ) )
+			     (put! db kvar 'replacing #t) ) ) ) ) ) ) ) ) ) ) ) )
 
      db)
 
-    ;; Remove explicitly consed rest parameters from promoted ones:
-    (set! rest-parameters-promoted-to-vector
-      (lset-difference eq? rest-parameters-promoted-to-vector explicitly-consed) )
-
     ;; Set original program-size, if this is the first analysis-pass:
     (unless original-program-size
       (set! original-program-size current-program-size) )
@@ -2150,8 +2116,9 @@
 						     custom
 						     (not (= (llist-length llist) (length (cdr subs)))))
 					    (quit
-					     "known procedure called with wrong number of arguments: ~A" 
-					     (source-info->string name) ) )
+					     "~a: procedure `~a' called with wrong number of arguments" 
+					     (source-info->line name)
+					     (if (pair? name) (cadr name) name)))
 					  (register-direct-call! id)
 					  (when custom (register-customizable! varname id)) 
 					  (list id custom) )
@@ -2351,36 +2318,38 @@
 
 (define-record-type lambda-literal
   (make-lambda-literal id external arguments argument-count rest-argument temporaries
-		       callee-signatures allocated directly-called closure-size
-		       looping customizable rest-argument-mode body direct)
+		       unboxed-temporaries callee-signatures allocated directly-called
+		       closure-size looping customizable rest-argument-mode body direct)
   lambda-literal?
   (id lambda-literal-id)			       ; symbol
   (external lambda-literal-external)		       ; boolean
-  (arguments lambda-literal-arguments)		       ; (symbol...)
+  (arguments lambda-literal-arguments)		       ; (symbol ...)
   (argument-count lambda-literal-argument-count)       ; integer
   (rest-argument lambda-literal-rest-argument)	       ; symbol | #f
   (temporaries lambda-literal-temporaries)	       ; integer
-  (callee-signatures lambda-literal-callee-signatures) ; (integer...)
+  (unboxed-temporaries lambda-literal-unboxed-temporaries) ; ((sym . utype) ...)
+  (callee-signatures lambda-literal-callee-signatures) ; (integer ...)
   (allocated lambda-literal-allocated)		       ; integer
   (directly-called lambda-literal-directly-called)     ; boolean
   (closure-size lambda-literal-closure-size)	       ; integer
   (looping lambda-literal-looping)		       ; boolean
   (customizable lambda-literal-customizable)	       ; boolean
-  (rest-argument-mode lambda-literal-rest-argument-mode) ; #f | LIST | VECTOR | UNUSED
+  (rest-argument-mode lambda-literal-rest-argument-mode) ; #f | LIST | NONE
   (body lambda-literal-body)				 ; expression
   (direct lambda-literal-direct))			 ; boolean
   
 (define (prepare-for-code-generation node db)
-  (let ([literals '()]
-	[lambda-info-literals '()]
-        [lambdas '()]
-        [temporaries 0]
-        [allocated 0]
-	[looping 0]
-        [signatures '()] 
-	[fastinits 0] 
-	[fastrefs 0] 
-	[fastsets 0] )
+  (let ((literals '())
+	(lambda-info-literals '())
+        (lambdas '())
+        (temporaries 0)
+	(ubtemporaries '())
+        (allocated 0)
+	(looping 0)
+        (signatures '()) 
+	(fastinits 0) 
+	(fastrefs 0) 
+	(fastsets 0) )
 
     (define (walk-var var e sf)
       (cond [(posq var e) => (lambda (i) (make-node '##core#local (list i) '()))]
@@ -2457,12 +2426,14 @@
 	      subs) ) )
 
 	  ((##core#lambda ##core#direct_lambda) 
-	   (let ([temps temporaries]
-		 [sigs signatures]
-		 [lping looping]
-		 [alc allocated] 
-		 [direct (eq? class '##core#direct_lambda)] )
+	   (let ((temps temporaries)
+		 (ubtemps ubtemporaries)
+		 (sigs signatures)
+		 (lping looping)
+		 (alc allocated) 
+		 (direct (eq? class '##core#direct_lambda)) )
 	     (set! temporaries 0)
+	     (set! ubtemporaries '())
 	     (set! allocated 0)
 	     (set! signatures '())
 	     (set! looping 0)
@@ -2483,9 +2454,8 @@
 				  vars)
 			      id
 			      '()) ] )
-		  (case rest-mode
-		    [(none) (debugging 'o "unused rest argument" rest id)]
-		    [(vector) (debugging 'o "rest argument accessed as vector" rest id)] )
+		  (when (eq? rest-mode 'none)
+		    (debugging 'o "unused rest argument" rest id))
 		  (when (and direct rest)
 		    (bomb "bad direct lambda" id allocated rest) )
 		  (set! lambdas
@@ -2496,6 +2466,7 @@
 			   argc
 			   rest
 			   (add1 temporaries)
+			   ubtemporaries
 			   signatures
 			   allocated
 			   (or direct (memq id direct-call-ids))
@@ -2512,6 +2483,7 @@
 			  lambdas) )
 		  (set! looping lping)
 		  (set! temporaries temps)
+		  (set! ubtemporaries ubtemps)
 		  (set! allocated alc)
 		  (set! signatures sigs)
 		  (make-node '##core#proc (list (first params)) '()) ) ) ) ) )
@@ -2522,10 +2494,19 @@
 		  [boxvars (if (eq? '##core#box (node-class val)) (list var) '())] )
 	     (set! temporaries (add1 temporaries))
 	     (make-node
-	      '##core#bind (list 1)
+	      '##core#bind (list 1)	; is actually never used with more than 1 variable
 	      (list (walk val e here boxes)
 		    (walk (second subs) (append e params) here (append boxvars boxes)) ) ) ) )
 
+	  ((##core#let_unboxed)
+	   (let* ((var (first params))
+		  (val (first subs)) )
+	     (set! ubtemporaries (alist-cons var (second params) ubtemporaries))
+	     (make-node
+	      '##core#let_unboxed params
+	      (list (walk val e here boxes)
+		    (walk (second subs) e here boxes) ) ) ) )
+
 	  ((set!)
 	   (let ([var (first params)]
 		 [val (first subs)] )
