@@ -1,7 +1,7 @@
 ;;;; csc.scm - Driver program for the CHICKEN compiler - felix -*- Scheme -*-
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -32,8 +32,10 @@
 (define-foreign-variable INSTALL_BIN_HOME c-string "C_INSTALL_BIN_HOME")
 (define-foreign-variable INSTALL_CC c-string "C_INSTALL_CC")
 (define-foreign-variable INSTALL_CXX c-string "C_INSTALL_CXX")
+(define-foreign-variable INSTALL_RC_COMPILER c-string "C_INSTALL_RC_COMPILER")
 (define-foreign-variable TARGET_CC c-string "C_TARGET_CC")
 (define-foreign-variable TARGET_CXX c-string "C_TARGET_CXX")
+(define-foreign-variable TARGET_RC_COMPILER c-string "C_TARGET_RC_COMPILER")
 (define-foreign-variable TARGET_CFLAGS c-string "C_TARGET_CFLAGS")
 (define-foreign-variable INSTALL_CFLAGS c-string "C_INSTALL_CFLAGS")
 (define-foreign-variable TARGET_LDFLAGS c-string "C_TARGET_LDFLAGS")
@@ -53,7 +55,9 @@
 (define-foreign-variable TARGET_STATIC_LIB_HOME c-string "C_TARGET_STATIC_LIB_HOME")
 (define-foreign-variable TARGET_RUN_LIB_HOME c-string "C_TARGET_RUN_LIB_HOME")
 (define-foreign-variable CHICKEN_PROGRAM c-string "C_CHICKEN_PROGRAM")
+(define-foreign-variable CSC_PROGRAM c-string "C_CSC_PROGRAM")
 (define-foreign-variable WINDOWS_SHELL bool "C_WINDOWS_SHELL")
+(define-foreign-variable BINARY_VERSION int "C_BINARY_VERSION")
 
 
 ;;; Parameters:
@@ -61,11 +65,14 @@
 (define mingw (eq? (build-platform) 'mingw32))
 (define msvc (eq? (build-platform) 'msvc))
 (define osx (eq? (software-version) 'macosx))
-(define hpux-hppa (and (eq? (software-version) 'hpux)
-                       (eq? (machine-type) 'hppa)))
+(define win (or mingw msvc))
+(define netbsd (eq? (software-version) 'netbsd))
+
+(define elf
+  (memq (software-version) '(linux netbsd freebsd solaris openbsd)))
 
 (define (quit msg . args)
-  (fprintf (current-error-port) "csc: ~?~%" msg args)
+  (fprintf (current-error-port) "~a: ~?~%" CSC_PROGRAM msg args)
   (exit 64) )
 
 (define chicken-prefix (get-environment-variable "CHICKEN_PREFIX"))
@@ -89,11 +96,12 @@
   (quotewrap 
    (prefix "chicken" "bin"
 	   (make-pathname
-	    (if host-mode INSTALL_BIN_HOME TARGET_BIN_HOME)
+	    INSTALL_BIN_HOME
 	    CHICKEN_PROGRAM))))
 
 (define compiler (quotewrap (if host-mode INSTALL_CC TARGET_CC)))
 (define c++-compiler (quotewrap (if host-mode INSTALL_CXX TARGET_CXX)))
+(define rc-compiler (quotewrap (if host-mode INSTALL_RC_COMPILER TARGET_RC_COMPILER)))
 (define linker (quotewrap (if msvc "link" (if host-mode INSTALL_CC TARGET_CC))))
 (define c++-linker (quotewrap (if msvc "link" (if host-mode INSTALL_CXX TARGET_CXX))))
 (define object-extension (if msvc "obj" "o"))
@@ -106,15 +114,12 @@
 (define default-translation-optimization-options '())
 (define pic-options (if (or mingw msvc) '("-DPIC") '("-fPIC" "-DPIC")))
 (define windows-shell WINDOWS_SHELL)
+(define generate-manifest #f)
 
-(define default-library (string-append
-                         (if msvc "libchicken-static." "libchicken.")
-                         library-extension))
-(define default-unsafe-library (string-append
-                                (if msvc "libuchicken-static." "libuchicken.")
-                                library-extension))
-
-(define cleanup-filename quotewrap)
+(define default-library
+  (string-append
+   (if msvc "libchicken-static." "libchicken.")
+   library-extension))
 
 (define default-compilation-optimization-options (string-split (if host-mode INSTALL_CFLAGS TARGET_CFLAGS)))
 (define best-compilation-optimization-options default-compilation-optimization-options)
@@ -124,13 +129,13 @@
 (define-constant simple-options
   '(-explicit-use -no-trace -no-warnings -no-usual-integrations -optimize-leaf-routines -unsafe
     -block -disable-interrupts -fixnum-arithmetic -to-stdout -profile -raw -accumulate-profile
-    -check-syntax -case-insensitive -benchmark-mode -shared -compile-syntax -no-lambda-info
+    -check-syntax -case-insensitive -shared -compile-syntax -no-lambda-info
     -lambda-lift -dynamic -disable-stack-overflow-checks -local
     -emit-external-prototypes-first -inline -release -scrutinize
     -analyze-only -keep-shadowed-macros -inline-global -ignore-repository
     -no-symbol-escape -no-parentheses-synonyms -r5rs-syntax
     -no-argc-checks -no-bound-checks -no-procedure-checks -no-compiler-syntax
-    -emit-all-import-libraries -setup-mode
+    -emit-all-import-libraries -setup-mode -unboxing -no-elevation
     -no-procedure-checks-for-usual-bindings))
 
 (define-constant complex-options
@@ -146,7 +151,6 @@
     (-S "-scrutinize")
     (|-P| "-check-syntax")
     (|-V| "-version")
-    (|-Ob| "-benchmark-mode")
     (-f "-fixnum-arithmetic")
     (|-D| "-feature")
     (-i "-case-insensitive")
@@ -161,7 +165,7 @@
     (-b "-block") ) )
 
 (define short-options
-  (string->list "PHhsfiENxubvwAOeWkctgS") )
+  (string->list "PHhsfiENxubvwAOeWkctgSJ") )
 
 
 ;;; Variables:
@@ -169,7 +173,9 @@
 (define scheme-files '())
 (define generated-scheme-files '())
 (define c-files '())
+(define rc-files '())
 (define generated-c-files '())
+(define generated-rc-files '())
 (define object-files '())
 (define generated-object-files '())
 (define cpp-mode #f)
@@ -180,15 +186,21 @@
 (define show-ldflags #f)
 (define show-libs #f)
 (define dry-run #f)
+(define gui #f)
+(define deploy #f)
+(define deployed #f)
+(define rpath #f)
 
 (define extra-libraries
   (if host-mode
       INSTALL_MORE_STATIC_LIBS
       TARGET_MORE_STATIC_LIBS))
+
 (define extra-shared-libraries 
   (if host-mode 
       INSTALL_MORE_LIBS
       TARGET_MORE_LIBS))
+
 (define default-library-files 
   (list
    (quotewrap
@@ -196,21 +208,12 @@
 	    (string-append
 	     (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
 	     (string-append "/" default-library)))) ))
-(define default-shared-library-files (if msvc
-                                         (list (string-append "libchicken." library-extension))
-                                         '("-lchicken")))
-(define unsafe-library-files
-  (list
-   (quotewrap 
-    (prefix default-unsafe-library "lib"
-	    (string-append 
-	     (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
-	     (string-append "/" default-unsafe-library)))) ))
-(define unsafe-shared-library-files (if msvc
-                                        (list (string-append "libuchicken." library-extension))
-                                        '("-luchicken")))
-(define gui-library-files default-library-files)
-(define gui-shared-library-files default-shared-library-files)
+
+(define default-shared-library-files 
+  (if msvc
+      (list (string-append "libchicken." library-extension))
+      '("-lchicken")))
+
 (define library-files default-library-files)
 (define shared-library-files default-shared-library-files)
 
@@ -223,8 +226,9 @@
 	 id) ) )
 
 (define compile-options '())
+
 (define builtin-compile-options
-  (if include-dir (list (conc "-I" (quotewrap include-dir))) '()))
+  (if include-dir (list (conc "-I\"" include-dir "\"")) '()))
 
 (define compile-only-flag "-c")
 (define translation-optimization-options default-translation-optimization-options)
@@ -238,19 +242,22 @@
              TARGET_LIB_HOME)) )
 
 (define link-options '())
-(define builtin-link-options
-  (cond ((or osx hpux-hppa mingw)
-	 (list (conc "-L" (quotewrap library-dir))))
-        (msvc
-         (list (conc "-LIBPATH:" (quotewrap library-dir))))
-	(else 
-	 (list
-	  (conc "-L" (quotewrap library-dir))
-	  (conc " -Wl,-R" (quotewrap (prefix "" "lib"
-					     (if host-mode
-						 INSTALL_LIB_HOME
-						 TARGET_RUN_LIB_HOME)))) ) ) ) )
 
+(define (builtin-link-options)
+  (cond (elf
+	 (list
+	  (conc "-L\"" library-dir "\"")
+	  (conc " -Wl,-R\""
+		(if (and deployed (not netbsd))
+		    "\\$ORIGIN"
+		    (prefix "" "lib"
+			    (if host-mode
+				INSTALL_LIB_HOME
+				TARGET_RUN_LIB_HOME)))
+		"\"")) )
+	(else
+	 (list (conc "-L\"" library-dir "\"")))))
+	
 (define target-filename #f)
 (define verbose #f)
 (define keep-files #f)
@@ -262,16 +269,16 @@
 (define static-libs #f)
 (define static-extensions '())
 (define required-extensions '())
-(define gui #f)
 
 
 ;;; Display usage information:
 
 (define (usage)
-  (display #<<EOF
-Usage: csc FILENAME | OPTION ...
+  (let ((csc CSC_PROGRAM))
+    (print #<#EOF
+Usage: #{csc} FILENAME | OPTION ...
 
-  `csc' is a driver program for the CHICKEN compiler. Files given on the
+  `#{csc}' is a driver program for the CHICKEN compiler. Files given on the
   command line are translated, compiled or linked as needed.
 
   FILENAME is a Scheme source file name with optional extension or a
@@ -307,7 +314,7 @@ Usage: csc FILENAME | OPTION ...
   Syntax related options:
 
     -i -case-insensitive           don't preserve case of read symbols    
-    -k  -keyword-style STYLE       enable alternative keyword-syntax
+    -K  -keyword-style STYLE       enable alternative keyword-syntax
                                     (prefix, suffix or none)
         -no-parentheses-synonyms   disables list delimiter synonyms
         -no-symbol-escape          disables support for escaped symbols
@@ -343,7 +350,7 @@ Usage: csc FILENAME | OPTION ...
 
   Optimization options:
 
-    -O -O1 -O2 -O3 -O4 -optimize-level NUMBER
+    -O -O1 -O2 -O3 -O4 -O5 -optimize-level NUMBER
                                    enable certain sets of optimization options
     -optimize-leaf-routines        enable leaf routine optimization
     -N  -no-usual-integrations     standard procedures may be redefined
@@ -353,15 +360,12 @@ Usage: csc FILENAME | OPTION ...
     -b  -block                     enable block-compilation
     -disable-interrupts            disable interrupts in compiled code
     -f  -fixnum-arithmetic         assume all numbers are fixnums
-    -Ob  -benchmark-mode           equivalent to '-block -optimize-level 4
-                                    -debug-level 0 -fixnum-arithmetic
-                                    -lambda-lift -inline -disable-interrupts'
     -lambda-lift                   perform lambda-lifting
-    -unsafe-libraries              link with unsafe runtime system
     -disable-stack-overflow-checks disables detection of stack-overflows
     -inline                        enable inlining
     -inline-limit                  set inlining threshold
     -inline-global                 enable cross-module inlining
+    -unboxing                      use unboxed temporaries if possible
     -n -emit-inline-file FILENAME  generate file with globally inlinable
                                     procedures (implies -inline -local)
     -consult-inline-file FILENAME  explicitly load inline file
@@ -391,11 +395,12 @@ Usage: csc FILENAME | OPTION ...
 
     -e  -embedded                  compile as embedded
                                     (don't generate `main()')
-    -W  -windows                   compile as Windows GUI application
+    -gui                           compile as GUI application
     -R  -require-extension NAME    require extension and import in compiled
                                     code
     -dll -library                  compile multiple units into a dynamic
                                     library
+    -deploy                        deploy self-contained application bundle
 
   Options to other passes:
 
@@ -442,7 +447,6 @@ Usage: csc FILENAME | OPTION ...
 
     -debug MODES                   display debugging output for the given modes
     -compiler PATHNAME             use other compiler than default `chicken'
-    -disable-c-syntax-checks       disable syntax checks of C code fragments
     -raw                           do not generate implicit init- and exit code
     -emit-external-prototypes-first
                                    emit prototypes for callbacks before foreign
@@ -451,6 +455,11 @@ Usage: csc FILENAME | OPTION ...
     -keep-shadowed-macros          do not remove shadowed macro
     -host                          compile for host when configured for
                                     cross-compiling
+    -private-repository            load extensions from executable path
+    -deployed                      compile support file to be used from a deployed 
+                                    executable
+    -no-elevation                  embed manifest on Windows to supress elevation
+                                    warnings for programs named `install' or `setup'
 
   Options can be collapsed if unambiguous, so
 
@@ -461,10 +470,10 @@ Usage: csc FILENAME | OPTION ...
     -v -k -fixnum-arithmetic -optimize
 
   The contents of the environment variable CSC_OPTIONS are implicitly passed to
-  every invocation of `csc'.
+  every invocation of `#{csc}'.
 
 EOF
-) )
+  ) ) )
 
 
 ;;; Parse arguments:
@@ -488,24 +497,26 @@ EOF
              (else "-shared")) link-options))
     (set! shared #t) )
 
+  (define (use-private-repository)
+    (set! compile-options (cons "-DC_PRIVATE_REPOSITORY" compile-options))
+    (when osx
+      (set! link-options (cons "-framework CoreFoundation" link-options))))
+
   (let loop ([args args])
     (cond [(null? args)
-           ;Builtin search directory options do not override explict options
+	   ;; Builtin search directory options do not override explict options
            (set! compile-options (append compile-options builtin-compile-options))
-           (set! link-options (append link-options builtin-link-options))
-           ;
+           (set! link-options (append link-options (builtin-link-options)))
+	   ;;
 	   (when inquiry-only
 	     (when show-cflags (print* (compiler-options) #\space))
 	     (when show-ldflags (print* (linker-options) #\space))
 	     (when show-libs (print* (linker-libraries #t) #\space))
 	     (newline)
 	     (exit) )
-	   #; ;UNUSED
-	   (when (null? scheme-files)
-	     (set! scheme-files c-files)
-	     (set! c-files '()) )
 	   (cond [(null? scheme-files)
-		  (when (and (null? c-files) (null? object-files))
+		  (when (and (null? c-files) 
+			     (null? object-files))
 		    (quit "no source files specified") )
 		  (let ((f0 (last (if (null? c-files) object-files c-files))))
 		    (unless target-filename
@@ -522,17 +533,20 @@ EOF
 			  (pathname-replace-extension (first scheme-files) shared-library-extension)
 			  (pathname-replace-extension (first scheme-files) executable-extension) ) ) )
 		  (run-translation) ] )
+	   (when (and deploy (not shared))
+	     (use-private-repository))
 	   (unless translate-only 
 	     (run-compilation)
 	     (unless compile-only
 	       (when (member target-filename scheme-files)
 		 (printf "Warning: output file will overwrite source file `~A' - renaming source to `~A.old'~%"
 			 target-filename target-filename)
-		 (unless (zero? ($system (sprintf "~A ~A ~A" 
-						  (if *windows-shell* "move" "mv")
-						  (quotewrap target-filename)
-						  (quotewrap (string-append target-filename ".old")))))
-		   (exit last-exit-code) ) )
+		 (command 
+		  (sprintf
+		      "~A ~A ~A" 
+		      (if windows-shell "move" "mv")
+		    (quotewrap target-filename)
+		    (quotewrap (string-append target-filename ".old")))))
 	       (run-linking)) ) ]
 	  [else
 	   (let* ([arg (car args)]
@@ -580,15 +594,6 @@ EOF
 		       (t-options "-verbose") 
 		       (set! verbose 2)) 
 		      (else (set! verbose #t))) ]
-	       [(-v2 -verbose)		; DEPRECATED
-		(set! verbose #t)
-		(t-options "-verbose") ]
-	       [(-v3)			; DEPRECATED
-		(set! verbose #t)
-		(t-options "-verbose")
-                (if (not msvc)
-                    (set! compile-options (cons* "-v" "-Q" compile-options)))
-		(set! link-options (cons (if msvc "-VERBOSE" "-v") link-options)) ]
 	       [(-w -no-warnings)
 		(set! compile-options (cons "-w" compile-options))
 		(t-options "-no-warnings") ]
@@ -614,18 +619,33 @@ EOF
 		(set! static-extensions (append static-extensions (list (car rest))))
 		(t-options "-static-extension" (car rest))
 		(set! rest (cdr rest)) ]
-	       [(-windows |-W|)
+	       ((-private-repository)
+		(use-private-repository))
+	       ((-no-elevation)
+		(set! generate-manifest #t))
+	       [(-gui
+		 -windows |-W|)		;DEPRECATED
 		(set! gui #t)
-		(cond
-                 (mingw
-		  (set! link-options
-		    (cons* "-lkernel32" "-luser32" "-lgdi32" "-mwindows"
-			   link-options))
-		  (set! compile-options (cons "-DC_WINDOWS_GUI" compile-options)))
-                 (msvc
-                  (set! link-options
-                    (cons* "kernel32.lib" "user32.lib" "gdi32.lib" link-options))
-		  (set! compile-options (cons "-DC_WINDOWS_GUI" compile-options)))) ]
+		(set! compile-options (cons "-DC_GUI" compile-options))
+		(set! object-files 
+		  (cons (make-pathname 
+			 INSTALL_SHARE_HOME "chicken.rc"
+			 object-extension) 
+			object-files))
+		(when (or msvc mingw)
+		  (cond
+		   (mingw
+		    (set! link-options
+		      (cons* "-lkernel32" "-luser32" "-lgdi32" "-mwindows"
+			     link-options)))
+		   (msvc
+		    (set! link-options
+		      (cons* "kernel32.lib" "user32.lib" "gdi32.lib" link-options)))))]
+	       ((-deploy)
+		(set! deploy #t)
+		(set! deployed #t))
+	       ((-deployed)
+		(set! deployed #t))
 	       [(-framework)
 		(check s rest)
 		(when osx 
@@ -640,6 +660,11 @@ EOF
 	       [(|-O2|) (set! rest (cons* "-optimize-level" "2" rest))]
 	       [(|-O3|) (set! rest (cons* "-optimize-level" "3" rest))]
 	       [(|-O4|) (set! rest (cons* "-optimize-level" "4" rest))]
+	       [(|-O5|)
+		(set! rest (cons* "-optimize-level" "5" rest))
+		(when (memq (build-platform) '(mingw32 cygwin gnu clang))
+		  (set! compile-options 
+		    (cons* "-O3" "-fomit-frame-pointer" compile-options)) ) ]
 	       [(-d0) (set! rest (cons* "-debug-level" "0" rest))]
 	       [(-d1) (set! rest (cons* "-debug-level" "1" rest))]
 	       [(-d2) (set! rest (cons* "-debug-level" "2" rest))]
@@ -679,24 +704,18 @@ EOF
 		(check s rest)
 		(set! link-options (append link-options (string-split (car rest))))
 		(set! rest (cdr rest)) ]
-	       [(-unsafe-libraries)
-		(t-options arg)
-		(set! library-files unsafe-library-files)
-		(set! shared-library-files unsafe-shared-library-files) ]
 	       [(-rpath)
 		(check s rest)
-		(when (eq? 'gnu (build-platform))
-		  (set! link-options (append link-options (list (string-append "-Wl,-R" (car rest)))))
+		(set! rpath (car rest))
+		(when (and (memq (build-platform) '(gnu clang))
+			   (not mingw))
+		  (set! link-options (append link-options (list (string-append "-Wl,-R" rpath))))
 		  (set! rest (cdr rest)) ) ]
 	       [(-host) #f]
 	       [(-) 
 		(set! target-filename (make-pathname #f "a" executable-extension))
 		(set! scheme-files (append scheme-files '("-")))]
 	       [else
-		(when (memq s '(-unsafe -benchmark-mode))
-		  (when (eq? s '-benchmark-mode)
-		    (set! library-files unsafe-library-files)
-		    (set! shared-library-files unsafe-shared-library-files) ) )
 		(when (eq? s '-to-stdout) 
 		  (set! to-stdout #t)
 		  (set! translate-only #t) )
@@ -737,9 +756,12 @@ EOF
 			     [else (quit "invalid option `~A'" s)] ) ]
 		      [(file-exists? arg)
 		       (let-values ([(dirs name ext) (decompose-pathname arg)])
-			 (cond [(not ext) (set! scheme-files (append scheme-files (list arg)))]
+			 (cond [(not ext)
+				(set! scheme-files (append scheme-files (list arg)))]
 			       [(member ext '("h" "c"))
 				(set! c-files (append c-files (list arg))) ]
+			       ((string-ci=? ext "rc")
+				(set! rc-files (append rc-files (list arg))) )
 			       [(member ext '("cpp" "C" "cc" "cxx" "hpp"))
 				(when osx (set! compile-options (cons "-no-cpp-precomp" compile-options)))
 				(set! cpp-mode #t)
@@ -771,45 +793,57 @@ EOF
 		(cond (cpp-mode "cpp")
 		      (objc-mode "m")
 		      (else "c") ) ) ] )
-       (unless (zero?
-		($system 
-		 (string-intersperse 
-		  (cons* translator (cleanup-filename f) 
-			 (append 
-			  (if to-stdout 
-			      '("-to-stdout")
-			      `("-output-file" ,(cleanup-filename fc)) )
-			  (map quote-option (append translate-options translation-optimization-options)) ) )
-		  " ") ) )
-	 (exit last-exit-code) )
+       (command
+	(string-intersperse 
+	 (cons* translator (quotewrap f) 
+		(append 
+		 (if to-stdout 
+		     '("-to-stdout")
+		     `("-output-file" ,(quotewrap fc)) )
+		 (map quote-option (append translate-options translation-optimization-options)) ) )
+	 " ") )
        (set! c-files (append (list fc) c-files))
        (set! generated-c-files (append (list fc) generated-c-files))))
    scheme-files)
   (unless keep-files (for-each $delete-file generated-scheme-files)) )
 
 
-;;; Compile all C files:
+;;; Compile all C  and .rc files:
 
 (define (run-compilation)
   (let ((ofiles '()))
     (for-each
      (lambda (f)
        (let ([fo (pathname-replace-extension f object-extension)])
-	 (unless (zero?
-		  ($system
-		   (string-intersperse
-		    (list (cond (cpp-mode c++-compiler)
-				(else compiler) )
-			  (cleanup-filename f)
-			  (string-append compile-output-flag (cleanup-filename fo)) 
-			  compile-only-flag
-			  (compiler-options) ) ) ) )
-	   (exit last-exit-code) )
+	 (command
+	  (string-intersperse
+	   (list (cond (cpp-mode c++-compiler)
+		       (else compiler) )
+		 (quotewrap f)
+		 (string-append compile-output-flag (quotewrap fo)) 
+		 compile-only-flag
+		 (compiler-options) ) ) )
 	 (set! generated-object-files (cons fo generated-object-files))
 	 (set! ofiles (cons fo ofiles))))
      c-files)
+    (when (and generate-manifest (eq? 'windows (software-type)))
+      (let ((rcf (pathname-replace-extension target-filename "rc")))
+	(create-win-manifest (pathname-file target-filename) rcf)
+	(set! rc-files (cons rcf rc-files))
+	(set! generated-rc-files (cons rcf generated-rc-files))))
+    (for-each
+     (lambda (f)
+       (let ((fo (string-append f "." object-extension)))
+	 (command
+	  (string-intersperse
+	   (list rc-compiler (quotewrap f) (quotewrap fo))))
+	 (set! generated-object-files (cons fo generated-object-files))
+	 (set! ofiles (cons fo ofiles))))
+     rc-files)
     (set! object-files (append (reverse ofiles) object-files)) ; put generated object files first
-    (unless keep-files (for-each $delete-file generated-c-files)) ) )
+    (unless keep-files 
+      (for-each $delete-file generated-c-files)
+      (for-each $delete-file generated-rc-files))))
 
 (define (compiler-options)
   (string-intersperse
@@ -823,37 +857,90 @@ EOF
 ;;; Link object files and libraries:
 
 (define (run-linking)
-  (let ((files (map cleanup-filename
-		    (append object-files
-			    (nth-value 0 (static-extension-info)) ) ) )
-	(target (cleanup-filename target-filename)))
-    (unless (zero?
-	     ($system
-	      (string-intersperse 
-	       (cons* (cond (cpp-mode c++-linker)
-			    (else linker) )
-		      (append
-		       files
-		       (list (string-append link-output-flag target)
-			     (linker-options)
-			     (linker-libraries #f) ) ) ) ) ) )
-      (exit last-exit-code) )
+  (let* ((files (map quotewrap
+		     (append object-files
+			     (nth-value 0 (static-extension-info)) ) ) )
+	 (target (quotewrap target-filename))
+	 (targetdir #f))
+    (when deploy
+      (set! targetdir (pathname-strip-extension target-filename))
+      (when (and osx gui)
+	(set! targetdir (make-pathname #f targetdir "app"))
+	(command (sprintf "mkdir -p ~a" (quotewrap (make-pathname targetdir "Contents/MacOS"))))
+	(command (sprintf "mkdir -p ~a" (quotewrap (make-pathname targetdir "Contents/Resources")))))
+      (set! target-filename
+	(make-pathname
+	 targetdir
+	 (if (and osx gui)
+	     (string-append "Contents/MacOS/" (pathname-file target-filename))
+	     (pathname-file target-filename))))
+      (set! target (quotewrap target-filename))
+      (unless (directory-exists? targetdir)
+	(when verbose
+	  (print "mkdir " targetdir))
+	(create-directory targetdir)))
+    (command
+     (string-intersperse 
+      (cons* (cond (cpp-mode c++-linker)
+		   (else linker) )
+	     (append
+	      files
+	      (list (string-append link-output-flag (quotewrap target-filename))
+		    (linker-options)
+		    (linker-libraries #f) ) ) ) ) )
     (when (and osx (or (not cross-chicken) host-mode))
-      (unless (zero? ($system 
-		      (string-append
-		       "install_name_tool -change libchicken.dylib "
-		       (quotewrap 
-			(make-pathname
-			 (prefix "" "lib"
-				 (if host-mode
-				     INSTALL_LIB_HOME
-				     TARGET_RUN_LIB_HOME))
-			 "libchicken.dylib") )
-		       " " 
-		       target) ) )
-	(exit last-exit-code) ) )
+      (command
+       (string-append
+	"install_name_tool -change libchicken.dylib "
+	(quotewrap 
+	 (let ((lib "libchicken.dylib"))
+	   (if deployed
+	       (make-pathname "@executable_path" lib)
+	       (make-pathname
+		(lib-path)
+		lib))))
+	" " 
+	target) )
+      (when (and gui (not deploy))
+	(rez target)))
+    (when (and deploy (not (or static static-libs)))
+      (copy-libraries 
+       (if (and osx gui)
+	   (make-pathname targetdir "Contents/MacOS")
+	   targetdir))
+      (when (and osx gui)
+	(create-mac-bundle
+	 (pathname-file target-filename)
+	 targetdir)))
     (unless keep-files (for-each $delete-file generated-object-files)) ) )
 
+(define (lib-path)
+  (prefix "" 
+	  "lib"
+	  (if win
+	      INSTALL_BIN_HOME
+	      (if host-mode
+		  INSTALL_LIB_HOME
+		  TARGET_RUN_LIB_HOME))))
+
+(define (copy-libraries targetdir)
+  (let ((lib (make-pathname
+	      (lib-path) 
+	      "libchicken"
+	      (cond (osx "dylib")
+		    (win "dll")
+		    (else (conc "so." BINARY_VERSION))))))
+    (copy-files lib targetdir)))
+
+(define (copy-files from to)
+  (command
+   (sprintf "~a ~a ~a"
+     (if windows-shell 
+	 "copy /Y"
+	 "cp")
+     (quotewrap from)
+     (quotewrap to))))
+
 (define (static-extension-info)
   (let ((rpath (repository-path)))
     (if (and rpath (pair? static-extensions))
@@ -882,8 +969,8 @@ EOF
    (append
     (if staticexts (nth-value 0 (static-extension-info)) '())
     (if (or static static-libs)
-        (if gui gui-library-files library-files)
-        (if gui gui-shared-library-files shared-library-files))
+        library-files
+        shared-library-files)
     (if (or static static-libs)
         (list extra-libraries)
         (list extra-shared-libraries)))))
@@ -909,11 +996,12 @@ EOF
 	s) ) )
 
 (define (quote-option x)
-  (if (string-any (lambda (c)
-		    (or (char-whitespace? c) (memq c +hairy-chars+)) )
-		  x)
-      (cleanup x)
-      x) )
+  (cond ((string-any (cut char=? #\" <>) x) x)
+	((string-any (lambda (c)
+		       (or (char-whitespace? c) (memq c +hairy-chars+)) )
+		     x)
+	 (cleanup x))
+	(else x) ))
 
 (define last-exit-code #f)
 
@@ -922,20 +1010,89 @@ EOF
   (let ((str (if windows-shell
 		 (string-append "\"" str "\"")
 		 str)))
-    (set! last-exit-code
-      (if dry-run 
-	  0
-	  (system str)))
-    (unless (zero? last-exit-code)
-      (printf "\nError: shell command terminated with non-zero exit status ~S: ~A~%" last-exit-code str) )
-    last-exit-code))
+    (let ((raw-exit-code (if dry-run 0 (system str))))
+      (unless (zero? raw-exit-code)
+	(printf "\nError: shell command terminated with non-zero exit status ~S: ~A~%" raw-exit-code str))
+      (set! last-exit-code
+	(if (zero? raw-exit-code) 0 1))
+      last-exit-code)))
+
+(define (command str)
+  (unless (zero? ($system str))
+    (exit last-exit-code)))
 
 (define ($delete-file str)
   (when verbose 
     (print "rm " str) )
   (unless dry-run (delete-file str) ))
 
+(define (rez file)
+  ;; see also: http://www.cocan.org/getting_started_with_ocaml_on_mac_os_x
+  (command 
+   (sprintf "/Developer/Tools/Rez -t APPL -o ~a ~a"
+     (quotewrap file)
+     (quotewrap (make-pathname home "mac.r")))))
+
+(define (create-mac-bundle prg dname)
+  (let* ((d0 (make-pathname dname "Contents"))
+	 (d (make-pathname dname "Contents/MacOS"))
+	 (d2 (make-pathname dname "Contents/Resources")))
+    (let ((icons (make-pathname d2 "CHICKEN.icns")))
+      (unless (file-exists? icons)
+	(copy-files 
+	 (make-pathname home "chicken/CHICKEN.icns") 
+	 d2)))
+    (let ((pl (make-pathname d0 "Info.plist")))
+      (unless (file-exists? pl)
+	(when verbose (print "generating " pl))
+	(with-output-to-file pl
+	  (cut print #<#EOF
+<?xml version="1.0" encoding="UTF-8"?>
+<!DOCTYPE plist SYSTEM "file://localhost/System/Library/DTDs/PropertyList.dtd">
+<plist version="0.9">
+<dict>
+	<key>CFBundlePackageType</key>
+	<string>APPL</string>
+	<key>CFBundleIconFile</key>
+	<string>CHICKEN.icns</string>
+        <key>CFBundleGetInfoString</key>
+	<string>Created by CHICKEN</string>
+	<key>CFBundleSignature</key>
+	<string>????</string>
+	<key>CFBundleExecutable</key>
+	<string>#{prg}</string>
+</dict>
+</plist>
+EOF
+)))
+      d)))
+
+(define (create-win-manifest prg rcfname)
+  (when verbose (print "generating " rcfname))
+  (with-output-to-file rcfname
+    (lambda ()
+      (print #<#EOF
+1 24 MOVEABLE PURE
+BEGIN
+  "<?xml version=""1.0"" encoding=""UTF-8"" standalone=""yes""?>\r\n"
+  "<assembly xmlns=""urn:schemas-microsoft-com:asm.v1"" manifestVersion=""1.0"">\r\n"
+  "  <assemblyIdentity version=""1.0.0.0"" processorArchitecture=""*"" name=""#{prg}"" type=""win32""/>\r\n"
+  "  <ms_asmv2:trustInfo xmlns:ms_asmv2=""urn:schemas-microsoft-com:asm.v2"">\r\n"
+  "    <ms_asmv2:security>\r\n"
+  "      <ms_asmv2:requestedPrivileges>\r\n"
+  "        <ms_asmv2:requestedExecutionLevel level=""asInvoker"" uiAccess=""false""/>\r\n"
+  "      </ms_asmv2:requestedPrivileges>\r\n"
+  "    </ms_asmv2:security>\r\n"
+  "  </ms_asmv2:trustInfo>\r\n"
+  "</assembly>\r\n"
+END
+EOF
+) ) ) )
+
 
 ;;; Run it:
 
-(run (append (string-split (or (get-environment-variable "CSC_OPTIONS") "")) arguments))
+(run
+ (append 
+  (string-split (or (get-environment-variable "CSC_OPTIONS") "")) 
+  arguments))
