@@ -1,7 +1,7 @@
 ;;;; csi.scm - Interpreter stub for CHICKEN
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -30,7 +30,7 @@
   (usual-integrations)
   (disable-interrupts)
   (disable-warning var)
-  (run-time-macros)			;*** later: compile-syntax
+  (compile-syntax)
   (foreign-declare #<<EOF
 #if (defined(_MSC_VER) && defined(_WIN32)) || defined(HAVE_DIRECT_H)
 # include <direct.h>
@@ -48,15 +48,14 @@ EOF
   parse-option-string chop-separator lookup-script-file
   report describe dump hexdump bytevector-data get-config
   deldups tty-input?
-  history-list history-count history-add history-ref
-  trace-indent trace-indent-level traced-procedure-entry traced-procedure-exit)
+  history-list history-count history-add history-ref)
 
 (declare
   (always-bound
     ##sys#windows-platform)
-  (hide parse-option-string bytevector-data member* canonicalize-args do-trace do-untrace
-	traced-procedures describer-table dirseparator? resolve-var
-	findall trace-indent command-table do-break do-unbreak broken-procedures) )
+  (hide parse-option-string bytevector-data member* canonicalize-args 
+	describer-table dirseparator? resolve-var
+	findall command-table) )
 
 
 ;;; Parameters:
@@ -65,6 +64,7 @@ EOF
 
 (set! ##sys#repl-print-length-limit 2048)
 (set! ##sys#features (cons #:csi ##sys#features))
+(set! ##sys#notices-enabled #t)
 
 
 ;;; Print all sorts of information:
@@ -98,7 +98,7 @@ EOF
   (display  #<<EOF
     -b  -batch                    terminate after command-line processing
     -w  -no-warnings              disable all warnings
-    -k  -keyword-style STYLE      enable alternative keyword-syntax
+    -K  -keyword-style STYLE      enable alternative keyword-syntax
                                    (prefix, suffix or none)
         -no-parentheses-synonyms  disables list delimiter synonyms
         -no-symbol-escape         disables support for escaped symbols
@@ -254,7 +254,6 @@ EOF
   (let ((eval eval)
 	(load-noisily load-noisily)
 	(read read)
-	(singlestep singlestep)
 	(read-line read-line)
 	(length length)
 	(display display)
@@ -266,7 +265,6 @@ EOF
 	(integer? integer?)
 	(values values) )
     (lambda (form)
-      (set! trace-indent-level 0)
       (cond ((eof-object? form) (exit))
 	    ((and (pair? form)
 		  (eq? 'unquote (##sys#slot form 0)) )
@@ -315,34 +313,10 @@ EOF
 			   (receive rs (time (eval x))
 			     (history-add rs)
 			     (apply values rs) ) ) )
-			((tr) (do-trace (map resolve-var (string-split (read-line)))))
-			((utr) (do-untrace (map resolve-var (string-split (read-line)))))
-			((br) (do-break (map resolve-var (string-split (read-line)))))
-			((ubr) (do-unbreak (map resolve-var (string-split (read-line)))))
-			((uba) (do-unbreak-all))
-			((breakall) 
-			 (set! ##sys#break-in-thread #f) ) 
-			((breakonly)
-			 (set! ##sys#break-in-thread (eval (read))) )
-			((info)
-			 (when (pair? traced-procedures)
-			   (printf "Traced: ~s~%" (map car traced-procedures)) )
-			 (when (pair? broken-procedures)
-			   (printf "Breakpoints: ~s~%" (map car broken-procedures)) ) )
-			((c)
-			 (cond (##sys#last-breakpoint
-				(let ((exn ##sys#last-breakpoint))
-				  (set! ##sys#last-breakpoint #f)
-				  (##sys#break-resume exn) ) )
-			       (else (display "no breakpoint pending\n") ) ) )
 			((exn) 
 			 (when ##sys#last-exception
 			   (history-add (list ##sys#last-exception))
 			   (describe ##sys#last-exception) ) )
-			((step)
-			 (let ((x (read)))
-			   (read-line)
-			   (singlestep (eval `(lambda () ,x))) ) )
 			((s)
 			 (let* ((str (read-line))
 				(r (system str)) )
@@ -362,16 +336,6 @@ EOF
  ,ln FILENAME ...  Load one or more files and print result of each top-level expression
  ,r                Show system information
  ,s TEXT ...       Execute shell-command
- ,tr NAME ...      Trace procedures
- ,utr NAME ...     Untrace procedures
- ,br NAME ...      Set breakpoints
- ,ubr NAME ...     Remove breakpoints
- ,uba              Remove all breakpoints
- ,breakall         Break in all threads (default)
- ,breakonly THREAD Break only in specified thread
- ,c                Continue from breakpoint
- ,info             List traced procedures and breakpoints
- ,step EXPR        Execute EXPR in single-stepping mode
  ,exn              Describe last exception
  ,t EXP            Evaluate form and print elapsed time
  ,x EXP            Pretty print expanded expression EXP\n")
@@ -395,131 +359,6 @@ EOF
   (##sys#strip-syntax (string->symbol str) (##sys#current-environment) #t))
 
 
-;;; Tracing:
-
-(define (del x lst tst)
-  (let loop ([lst lst])
-    (if (null? lst)
-	'()
-	(let ([y (car lst)])
-	  (if (tst x y)
-	      (cdr lst)
-	      (cons y (loop (cdr lst))) ) ) ) ) )
-
-(define trace-indent-level 0)
-(define traced-procedures '())
-(define broken-procedures '())
-
-(define trace-indent
-  (lambda ()
-    (write-char #\|)
-    (do ((i trace-indent-level (sub1 i)))
-	((<= i 0))
-      (write-char #\space) ) ) )
-
-(define traced-procedure-entry
-  (lambda (name args)
-    (trace-indent)
-    (set! trace-indent-level (add1 trace-indent-level))
-    (write (cons name args))
-    (##sys#write-char-0 #\newline ##sys#standard-output) 
-    (flush-output) ) )
-
-(define traced-procedure-exit
-  (lambda (name results)
-    (set! trace-indent-level (sub1 trace-indent-level))
-    (trace-indent)
-    (write name)
-    (display " -> ")
-    (for-each
-     (lambda (x)
-       (write x)
-       (write-char #\space) )
-     results)
-    (##sys#write-char-0 #\newline ##sys#standard-output) 
-    (flush-output) ) )
-
-(define do-trace
-  (lambda (names)
-    (if (null? names)
-	(for-each (lambda (a) (print (car a))) traced-procedures) 
-	(for-each
-	 (lambda (s)
-	   (let ((s (expand s)))
-	     (cond ((assq s traced-procedures)
-		    (##sys#warn "procedure already traced" s) )
-		   ((assq s broken-procedures)
-		    (##sys#warn "procedure already has breakpoint") )
-		   (else
-		    (let ((old (##sys#slot s 0)))
-		      (cond ((not (procedure? old)) (##sys#error "cannot trace non-procedure" s))
-			    (else
-			     (set! traced-procedures (cons (cons s old) traced-procedures))
-			     (##sys#setslot
-			      s 0
-			      (lambda args
-				(traced-procedure-entry s args)
-				(call-with-values (lambda () (apply old args))
-				  (lambda results
-				    (traced-procedure-exit s results) 
-				    (apply values results) ) ) ) ) ) ) ) ) ) ) )
-	 names) ) ) )
-
-(define do-untrace 
-  (lambda (names)
-    (for-each
-     (lambda (s)
-       (let* ((s (expand s))
-	      (p (assq s traced-procedures)) )
-	 (cond ((not p) (##sys#warn "procedure not traced" s))
-	       (else
-		(##sys#setslot s 0 (cdr p))
-		(set! traced-procedures (del p traced-procedures eq?) ) ) ) ) )
-     names) ) )
-
-(define do-break
-  (lambda (names)
-    (if (null? names)
-	(for-each (lambda (b) (print (car a))) broken-procedures) 
-	(for-each
-	 (lambda (s)
-	   (let* ((s (expand s))
-		  (a (assq s traced-procedures)))
-	     (when a
-	       (##sys#warn "un-tracing procedure" s)
-	       (##sys#setslot s 0 (cdr a))
-	       (set! traced-procedures (del a traced-procedures eq?)) )
-	     (let ((old (##sys#slot s 0)))
-	       (cond ((not (procedure? old)) (##sys#error "cannot set breakpoint on non-procedure" s))
-		     (else
-		      (set! broken-procedures (cons (cons s old) broken-procedures))
-		      (##sys#setslot
-		       s 0
-		       (lambda args
-			 (##sys#break-entry s args)
-			 (##sys#apply old args) ) ) ) ) ) ) )
-	 names) ) ) )
-
-(define do-unbreak
-  (lambda (names)
-    (for-each
-     (lambda (s)
-       (let* ((s (expand s))
-	      (p (assq s broken-procedures)) )
-	 (cond ((not p) (##sys#warn "procedure has no breakpoint" s))
-	       (else
-		(##sys#setslot s 0 (cdr p))
-		(set! broken-procedures (del p broken-procedures eq?) ) ) ) ) )
-     names) ) )
-
-(define do-unbreak-all
-  (lambda ()
-    (for-each (lambda (bp)
-                (##sys#setslot (car bp) 0 (cdr bp)))
-              broken-procedures)
-    (set! broken-procedures '())
-    (##sys#void)))
-
 ;;; Parse options from string:
 
 (define (parse-option-string str)
@@ -543,7 +382,10 @@ EOF
 	(chop chop)
 	(sort sort)
 	(with-output-to-port with-output-to-port)
-	(current-output-port current-output-port) )
+	(current-output-port current-output-port) 
+	(prefix
+	 (or (get-environment-variable "CHICKEN_PREFIX")
+	     (foreign-value "C_INSTALL_PREFIX" c-string) ) ))
     (lambda port
       (with-output-to-port (if (pair? port) (car port) (current-output-port))
 	(lambda ()
@@ -551,7 +393,7 @@ EOF
 	  (let ([sinfo (##sys#symbol-table-info)]
 		[minfo (memory-statistics)] )
 	    (define (shorten n) (/ (truncate (* n 100)) 100))
-	    (printf "Features:")
+	    (printf "Features:\n")
 	    (for-each
 	     (lambda (lst) 
 	       (display "\n  ")
@@ -560,11 +402,13 @@ EOF
 		  (printf "~a~a" f (make-string (fxmax 1 (fx- 16 (string-length f))) #\space)) )
 		lst) )
 	     (chop (sort (map keyword->string ##sys#features) string<?) 5))
-	    (printf "~%~
+	    (printf "~%~%~
                    Machine type:    \t~A ~A~%~
                    Software type:   \t~A~%~
                    Software version:\t~A~%~
                    Build platform:  \t~A~%~
+                   Installation prefix:\t~A~%~
+                   Extension path:  \t~A~%~
                    Include path:    \t~A~%~
                    Symbol-table load:\t~S~%  ~
                      Avg bucket length:\t~S~%  ~
@@ -576,6 +420,8 @@ EOF
 		    (software-type)
 		    (software-version)
 		    (build-platform)
+		    prefix
+		    (repository-path)
 		    ##sys#include-pathnames
 		    (shorten (vector-ref sinfo 0))
 		    (shorten (vector-ref sinfo 1))
@@ -652,16 +498,22 @@ EOF
 	     (##sys#write-char-0 #\newline ##sys#standard-output) ]
 	    [(eq? x (##sys#slot '##sys#arbitrary-unbound-symbol 0))
 	     (fprintf out "unbound value~%") ]
-	    [(##sys#number? x) (fprintf out "number ~S~%" x)]
+	    [(flonum? x) (fprintf out "inexact number ~S~%" x)]
+	    [(number? x) (fprintf out "number ~S~%" x)]
 	    [(string? x) (descseq "string" ##sys#size string-ref 0)]
 	    [(vector? x) (descseq "vector" ##sys#size ##sys#slot 0)]
 	    [(symbol? x)
-	     (unless (##sys#symbol-has-toplevel-binding? x) (display "unbound " out))
+	     (unless (##sys#symbol-has-toplevel-binding? x)
+	       (display "unbound " out))
 	     (when (and (symbol? x) (fx= 0 (##sys#byte (##sys#slot x 1) 0)))
 	       (display "keyword " out) )
-	     (fprintf out "~asymbol with name ~S~%"
-		      (if (##sys#interned-symbol? x) "" "uninterned ")
-		      (##sys#symbol->string x))
+	     (let ((q (##sys#qualified-symbol? x)))
+	       (fprintf out "~a~asymbol with name ~S~%"
+		 (if (##sys#interned-symbol? x) "" "uninterned ")
+		 (if q "qualified " "")
+		 (if q 
+		     (##sys#symbol->qualified-string x)
+		     (##sys#symbol->string x))))
 	     (let ((plist (##sys#slot x 2)))
 	       (unless (null? plist)
 		 (display "  \nproperties:\n\n" out)
@@ -813,6 +665,15 @@ EOF
 
 ;;; Start interpreting:
 
+(define (del x lst tst)
+  (let loop ([lst lst])
+    (if (null? lst)
+	'()
+	(let ([y (car lst)])
+	  (if (tst x y)
+	      (cdr lst)
+	      (cons y (loop (cdr lst))) ) ) ) ) )
+
 (define (deldups lis . maybe-=)
   (let ((elt= (optional maybe-= equal?)))
     (let recur ((lis lis))
@@ -870,13 +731,13 @@ EOF
     "-ss" "-sx" "-s" "-script") )
 
 (define-constant complex-options
-  '("-D" "-feature" "-I" "-include-path" "-k" "-keyword-style") )
+  '("-D" "-feature" "-I" "-include-path" "-K" "-keyword-style") )
 
 (define (run)
   (let* ([extraopts (parse-option-string (or (get-environment-variable "CSI_OPTIONS") ""))]
 	 [args (canonicalize-args (command-line-arguments))]
 	 ; Check for these before 'args' is updated by any 'extraopts'
-	 [kwstyle (member* '("-k" "-keyword-style") args)]
+	 [kwstyle (member* '("-K" "-keyword-style") args)]
 	 [script (member* '("-ss" "-sx" "-s" "-script") args)])
     (cond [script
 	   (when (or (not (pair? (cdr script)))
@@ -920,7 +781,8 @@ EOF
 	  (do ([x (read in) (read in)])
 	      ((eof-object? x))
 	    (rec (receive (eval x))) ) ) )
-      (when quietflag (set! ##sys#eval-debug-level 0))
+      (when quietflag
+	(set! ##sys#eval-debug-level 0))
       (when (member* '("-h" "-help" "--help") args)
 	(print-usage)
 	(exit 0) )
@@ -973,6 +835,8 @@ EOF
 	(parentheses-synonyms #f)
 	(symbol-escape #f) )
       (unless (or (member* '("-n" "-no-init") args) script) (loadinit))
+      (when batch 
+	(set! ##sys#notices-enabled #f))
       (do ([args args (cdr args)])
 	  ((null? args)
 	   (unless batch 
@@ -1000,9 +864,18 @@ EOF
 		    arg 
 		    (and (equal? "-sx" scr)
 			 (lambda (x)
-			   (pretty-print x ##sys#standard-error)
-			   (newline ##sys#standard-error)
-			   (eval x)))
+			   (let* ((str (with-output-to-string (cut pretty-print x)))
+				  (len (string-length str)))
+			     (flush-output ##sys#standard-output)
+			     (display "\n; " ##sys#standard-error)
+			     (do ((i 0 (fx+ i 1)))
+				 ((fx>= i len))
+			       (let ((c (string-ref str i)))
+				 (write-char c ##sys#standard-error)
+				 (when (char=? #\newline c)
+				   (display "; " ##sys#standard-error))))
+			     (newline ##sys#standard-error)
+			     (eval x))))
 		    #f)
 		   (when (equal? "-ss" scr)
 		     (call-with-values (cut main (command-line-arguments))
