@@ -1,7 +1,7 @@
 ;;;; eval.scm - Interpreter for CHICKEN
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -31,7 +31,10 @@
   (disable-warning var)
   (hide ##sys#split-at-separator
 	##sys#r4rs-environment ##sys#r5rs-environment 
-	##sys#interaction-environment pds pdss pxss) )
+	##sys#interaction-environment pds pdss pxss d) 
+  (not inline ##sys#repl-eval-hook ##sys#repl-read-hook ##sys#repl-print-hook 
+       ##sys#read-prompt-hook ##sys#alias-global-hook ##sys#user-read-hook
+       ##sys#syntax-error-hook))
 
 (define (d arg1 . more)
   (if (null? more)
@@ -48,6 +51,10 @@
 #ifndef C_INSTALL_SHARE_HOME
 # define C_INSTALL_SHARE_HOME NULL
 #endif
+
+#ifndef C_BINARY_VERSION
+# define C_BINARY_VERSION      0
+#endif
 <#
 
 (cond-expand
@@ -58,9 +65,10 @@
     (no-procedure-checks-for-usual-bindings)
     (bound-to-procedure 
      ##sys#check-char ##sys#check-exact ##sys#check-port ##sys#check-string ##sys#load-library
+     ##sys#load-library-0
      ##sys#for-each ##sys#map ##sys#setslot ##sys#allocate-vector ##sys#check-pair ##sys#error-not-a-proper-list
-     ##sys#check-symbol ##sys#check-vector ##sys#floor ##sys#ceiling ##sys#truncate ##sys#round 
-     ##sys#check-number ##sys#cons-flonum ##sys#copy-env-table
+     ##sys#check-symbol ##sys#check-vector 
+     ##sys#check-number ##sys#copy-env-table
      ##sys#flonum-fraction ##sys#make-port ##sys#fetch-and-check-port-arg ##sys#print ##sys#check-structure 
      ##sys#make-structure ##sys#feature?
      ##sys#error-handler ##sys#hash-symbol ##sys#check-syntax
@@ -73,7 +81,7 @@
      ##sys#ensure-heap-reserve ##sys#syntax-error-hook ##sys#read-prompt-hook
      ##sys#repl-eval-hook ##sys#append ##sys#eval-decorator
      open-output-string get-output-string make-parameter software-type software-version machine-type
-     build-platform set-extensions-specifier! ##sys#string->symbol list->vector get-environment-variable
+     build-platform ##sys#string->symbol list->vector get-environment-variable
      extension-information syntax-error ->string chicken-home ##sys#expand-curried-define
      vector->list store-string open-input-string eval ##sys#gc
      with-exception-handler print-error-message read-char read ##sys#read-error
@@ -94,10 +102,11 @@
 
 (define-foreign-variable install-egg-home c-string "C_INSTALL_EGG_HOME")
 (define-foreign-variable installation-home c-string "C_INSTALL_SHARE_HOME")
+(define-foreign-variable binary-version int "C_BINARY_VERSION")
 
 (define ##sys#core-library-modules
-  '(extras lolevel utils files tcp regex regex-extras posix srfi-1 srfi-4 srfi-13 
-	   srfi-14 srfi-18 data-structures ports chicken-syntax))
+  '(extras lolevel utils files tcp regex posix srfi-1 srfi-4 srfi-13 
+	   srfi-14 srfi-18 srfi-69 data-structures ports chicken-syntax))
 
 (define ##sys#explicit-library-modules '())
 
@@ -112,18 +121,17 @@
 (define-constant setup-file-extension "setup-info")
 (define-constant repository-environment-variable "CHICKEN_REPOSITORY")
 (define-constant prefix-environment-variable "CHICKEN_PREFIX")
-(define-constant default-binary-version 4)
 
 ; these are actually in unit extras, but that is used by default
 ; srfi-12 in unit library
-; srfi-98 partically in unit posix
+; srfi-98 partially in unit posix
 
 (define-constant builtin-features
   '(chicken srfi-2 srfi-6 srfi-10 srfi-12 srfi-23 srfi-28 srfi-30 srfi-31 srfi-39 
-	    srfi-69 srfi-98) )
+	    srfi-55 srfi-88 srfi-98) )
 
 (define-constant builtin-features/compiled
-  '(srfi-6 srfi-8 srfi-9 srfi-11 srfi-15 srfi-16 srfi-17 srfi-26 srfi-55) )
+  '(srfi-6 srfi-8 srfi-9 srfi-11 srfi-15 srfi-16 srfi-17 srfi-26) )
 
 (define ##sys#chicken-prefix
   (let ((prefix (and-let* ((p (get-environment-variable prefix-environment-variable)))
@@ -219,6 +227,7 @@
 
 (define ##sys#unbound-in-eval #f)
 (define ##sys#eval-debug-level 1)
+(define ##sys#unsafe-eval #f)
 
 (define ##sys#compile-to-closure
   (let ([write write]
@@ -292,20 +301,21 @@
 			  (if ##sys#eval-environment
 			      (let ([loc (##sys#hash-table-location ##sys#eval-environment var #t)])
 				(unless loc (##sys#syntax-error-hook "reference to undefined identifier" var))
-				(cond-expand 
-				 [unsafe (lambda v (##sys#slot loc 1))]
-				 [else
-				  (lambda v 
-				    (let ([val (##sys#slot loc 1)])
-				      (if (eq? unbound val)
-					  (##sys#error "unbound variable" var)
-					  val) ) ) ] ) )
-			      (cond-expand
-			       [unsafe (lambda v (##core#inline "C_slot" var 0))]
-			       [else
-				(when (and ##sys#unbound-in-eval (not (##sys#symbol-has-toplevel-binding? var)))
-				  (set! ##sys#unbound-in-eval (cons (cons var cntr) ##sys#unbound-in-eval)) )
-				(lambda v (##core#inline "C_retrieve" var))] ) ) ) ]
+				(if ##sys#unsafe-eval 
+				    (lambda v (##sys#slot loc 1))
+				    (lambda v 
+				      (let ([val (##sys#slot loc 1)])
+					(if (eq? unbound val)
+					    (##sys#error "unbound variable" var)
+					    val) ) ) ))
+			      (cond (##sys#unsafe-eval
+				     (lambda v (##core#inline "C_slot" var 0)))
+				    (else
+				     (when (and ##sys#unbound-in-eval
+						(not (##sys#symbol-has-toplevel-binding? var)))
+				       (set! ##sys#unbound-in-eval
+					 (cons (cons var cntr) ##sys#unbound-in-eval)) )
+				     (lambda v (##core#inline "C_retrieve" var))))))]
 		       [(zero? i) (lambda (v) (##sys#slot (##sys#slot v 0) j))]
 		       [else (lambda (v) (##sys#slot (##core#inline "C_u_i_list_ref" v i) j))] ) ) )
 	      [(##sys#number? x)
@@ -335,8 +345,7 @@
 		       ;; a normal walking of the operator)
 		       (case head
 
-			 [(quote)
-			  (##sys#check-syntax 'quote x '(quote _) #f se)
+			 [(##core#quote)
 			  (let* ((c (##sys#strip-syntax (cadr x))))
 			    (case c
 			      [(-1) (lambda v -1)]
@@ -348,7 +357,7 @@
 			      [(()) (lambda v '())]
 			      [else (lambda v c)] ) ) ]
 
-			 ((syntax ##core#syntax)
+			 ((##core#syntax)
 			  (let ((c (cadr x)))
 			    (lambda v c)))
 
@@ -367,8 +376,7 @@
 		   
 			 [(##core#undefined) (lambda (v) (##core#undefined))]
 
-			 [(if)
-			  (##sys#check-syntax 'if x '(if _ _ . #(_)) #f se)
+			 [(##core#if)
 			  (let* ([test (compile (cadr x) e #f tf cntr se)]
 				 [cns (compile (caddr x) e #f tf cntr se)]
 				 [alt (if (pair? (cdddr x))
@@ -376,10 +384,9 @@
 					  (compile '(##core#undefined) e #f tf cntr se) ) ] )
 			    (lambda (v) (if (##core#app test v) (##core#app cns v) (##core#app alt v))) ) ]
 
-			 [(begin ##core#begin)
-			  (##sys#check-syntax 'begin x '(_ . #(_ 0)) #f se)
-			  (let* ([body (##sys#slot x 1)]
-				 [len (length body)] )
+			 [(##core#begin)
+			  (let* ((body (##sys#slot x 1))
+				 (len (length body)) )
 			    (case len
 			      [(0) (compile '(##core#undefined) e #f tf cntr se)]
 			      [(1) (compile (##sys#slot body 0) e #f tf cntr se)]
@@ -392,8 +399,7 @@
 				      [x3 (compile `(##core#begin ,@(##sys#slot (##sys#slot body 1) 1)) e #f tf cntr se)] )
 				 (lambda (v) (##core#app x1 v) (##core#app x2 v) (##core#app x3 v)) ) ] ) ) ]
 
-			 [(set! ##core#set!)
-			  (##sys#check-syntax 'set! x '(_ variable _) #f se)
+			 [(##core#set!)
 			  (let ((var (cadr x)))
 			    (receive (i j) (lookup var e se)
 			      (let ((val (compile (caddr x) e var tf cntr se)))
@@ -416,8 +422,7 @@
 					 (##sys#setslot
 					  (##core#inline "C_u_i_list_ref" v i) j (##core#app val v)) ) ] ) ) ) ) ]
 
-			 [(let ##core#let)
-			  (##sys#check-syntax 'let x '(_ #((variable _) 0) . #(_ 1)) #f se)
+			 [(##core#let)
 			  (let* ([bindings (cadr x)]
 				 [n (length bindings)] 
 				 [vars (map (lambda (x) (car x)) bindings)] 
@@ -468,8 +473,7 @@
 				       (##sys#setslot v2 i (##core#app (##sys#slot vlist 0) v)) )
 				     (##core#app body (cons v2 v)) ) ) ) ] ) ) ]
 
-			 ((letrec ##core#letrec)
-			  (##sys#check-syntax 'letrec x '(_ #((symbol _) 0) . #(_ 1)))
+			 ((##core#letrec)
 			  (let ((bindings (cadr x))
 				(body (cddr x)) )
 			    (compile
@@ -483,7 +487,7 @@
 			       (##core#let () ,@body) )
 			     e h tf cntr se)))
 
-			 [(lambda ##core#lambda)
+			 [(##core#lambda)
 			  (##sys#check-syntax 'lambda x '(_ lambda-list . #(_ 1)) #f se)
 			  (let* ([llist (cadr x)]
 				 [body (cddr x)] 
@@ -578,8 +582,7 @@
 						   (##core#app body (##sys#cons (apply ##sys#vector as) v)))))
 					   info h cntr) ) ) ] ) ) ) ) ) ]
 
-			 ((let-syntax)
-			  (##sys#check-syntax 'let-syntax x '(let-syntax #((variable _) 0) . #(_ 1)) #f se)
+			 ((##core#let-syntax)
 			  (let ((se2 (append
 				      (map (lambda (b)
 					     (list
@@ -593,8 +596,7 @@
 			     (##sys#canonicalize-body (cddr x) se2 #f)
 			     e #f tf cntr se2)))
 			       
-			 ((letrec-syntax)
-			  (##sys#check-syntax 'letrec-syntax x '(letrec-syntax #((variable _) 0) . #(_ 1)) #f se)
+			 ((##core#letrec-syntax)
 			  (let* ((ms (map (lambda (b)
 					    (list
 					     (car b)
@@ -611,17 +613,9 @@
 			     (##sys#canonicalize-body (cddr x) se2 #f)
 			     e #f tf cntr se2)))
 			       
-			 ((define-syntax define-compiled-syntax)
-			  (##sys#check-syntax
-			   'define-syntax x
-			   (if (and (pair? (cdr x)) (pair? (cadr x)))
-			       '(_ (variable . lambda-list) . #(_ 1))
-			       '(_ variable _))
-			   #f se)
-			  (let* ((var (if (pair? (cadr x)) (caadr x) (cadr x)))
-				 (body (if (pair? (cadr x))
-					   `(,(rename 'lambda se) ,(cdadr x) ,@(cddr x))
-					   (caddr x)))
+			 ((##core#define-syntax)
+			  (let* ((var (cadr x))
+				 (body (caddr x))
 				 (name (rename var se)))
 			    (##sys#register-syntax-export 
 			     name (##sys#current-module)
@@ -640,8 +634,15 @@
 			   (##sys#canonicalize-body (cddr x) se #f)
 			   e #f tf cntr se))
 
+			 ((##core#include)
+			  (compile
+			   `(##core#begin
+			     ,@(##sys#include-forms-from-file (cadr x)))
+			   e #f tf cntr se))
+
 			 ((##core#module)
-			  (let* ((name (rename (cadr x) se))
+			  (let* ((x (##sys#strip-syntax x))
+				 (name (##sys#strip-syntax (cadr x)))
 				 (exports 
 				  (or (eq? #t (caddr x))
 				      (map (lambda (exp)
@@ -688,9 +689,6 @@
 			 [(##core#loop-lambda)
 			  (compile `(,(rename 'lambda se) ,@(cdr x)) e #f tf cntr se) ]
 
-			 [(##core#named-lambda)
-			  (compile `(,(rename 'lambda se) ,@(cddr x)) e (cadr x) tf cntr se) ]
-
 			 [(##core#require-for-syntax)
 			  (let ([ids (map (lambda (x)
 					    (eval/meta x))
@@ -706,11 +704,12 @@
 			 [(##core#require-extension)
 			  (let ((imp? (caddr x)))
 			    (compile
-			     (let loop ([ids (cadr x)])
+			     (let loop ([ids (##sys#strip-syntax (cadr x))])
 			       (if (null? ids)
 				   '(##core#undefined)
-				   (let-values ([(exp _)
-						 (##sys#do-the-right-thing (car ids) #f imp?)])
+				   (let-values ([(exp _) 
+						 (##sys#do-the-right-thing
+						  (car ids) #f imp?)])
 				     `(##core#begin ,exp ,(loop (cdr ids))) ) ) )
 			     e #f tf cntr se) ) ]
 
@@ -727,27 +726,25 @@
 			 [(##core#declare)
 			  (if (memq #:compiling ##sys#features)
 			      (for-each (lambda (d) (##compiler#process-declaration d se)) (cdr x)) 
-			      (##sys#warn "declarations are ignored in interpreted code" x) )
+			      (##sys#notice
+			       "declarations are ignored in interpreted code"
+			       x) )
 			  (compile '(##core#undefined) e #f tf cntr se) ]
 
-			 [(define-inline define-constant)
+			 [(##core#define-inline ##core#define-constant)
 			  (compile `(,(rename 'define se) ,@(cdr x)) e #f tf cntr se) ]
                    
 			 [(##core#primitive ##core#inline ##core#inline_allocate ##core#foreign-lambda 
 					    ##core#define-foreign-variable 
 					    ##core#define-external-variable ##core#let-location
-					    ##core#foreign-primitive
+					    ##core#foreign-primitive ##core#location
 					    ##core#foreign-lambda* ##core#define-foreign-type)
 			  (##sys#syntax-error-hook "cannot evaluate compiler-special-form" x) ]
 
 			 [(##core#app)
 			  (compile-call (cdr x) e tf cntr se) ]
 
-			 [else
-			  (cond [(eq? head 'location)
-				 (##sys#syntax-error-hook "cannot evaluate compiler-special-form" x) ]
-
-				[else (compile-call x e tf cntr se)] ) ] ) ) ) ) ]
+			 [else (compile-call x e tf cntr se)] ) ) ) ) ]
 	      
 	      [else
 	       (emit-syntax-trace-info tf x cntr)
@@ -823,8 +820,8 @@
 	     (##sys#check-structure env 'environment)
 	     (set! e (##sys#slot env 1)) 
 	     (set! mut (##sys#slot env 2)) ) ) )
-       ((fluid-let ([##sys#environment-is-mutable mut]
-		    [##sys#eval-environment e] )
+       ((fluid-let ((##sys#environment-is-mutable mut)
+		    (##sys#eval-environment e) )
 	  (##sys#compile-to-closure x '() (##sys#current-environment)) )
 	'() ) ) ) ) )
 
@@ -906,24 +903,24 @@
 	(set! input (##sys#expand-home-path input)) )
       (let* ([isdir #f]
 	     [fname 
-	     (cond [(port? input) #f]
-		   [(not (string? input)) (badfile input)]
-		   [(and-let* ([info (##sys#file-info input)]
-			       [id (##sys#slot info 4)] )
-		      (set! isdir (eq? 1 id)) 
-		      (not id) )
-		    input]
-		   [else
-		    (let ([fname2 (##sys#string-append input ##sys#load-dynamic-extension)])
-		      (if (and (not ##sys#dload-disabled)
-			       (##sys#fudge 24) ; dload?
-			       (##sys#file-info fname2))
-			  fname2
-			  (let ([fname3 (##sys#string-append input source-file-extension)])
-			    (if (##sys#file-info fname3)
-				fname3
-				(and (not isdir) input) ) ) ) ) ] ) ]
-	    [evproc (or evaluator eval)] )
+	      (cond [(port? input) #f]
+		    [(not (string? input)) (badfile input)]
+		    [(and-let* ([info (##sys#file-info input)]
+				[id (##sys#slot info 4)] )
+		       (set! isdir (eq? 1 id)) 
+		       (not isdir) )
+		     input]
+		    [else
+		     (let ([fname2 (##sys#string-append input ##sys#load-dynamic-extension)])
+		       (if (and (not ##sys#dload-disabled)
+				(##sys#fudge 24) ; dload?
+				(##sys#file-info fname2))
+			   fname2
+			   (let ([fname3 (##sys#string-append input source-file-extension)])
+			     (if (##sys#file-info fname3)
+				 fname3
+				 (and (not isdir) input) ) ) ) ) ] ) ]
+	     [evproc (or evaluator eval)] )
 	(cond [(and (string? input) (not fname))
 	       (##sys#signal-hook #:file-error 'load "cannot open file" input) ]
 	      [(and (load-verbose) fname)
@@ -937,18 +934,18 @@
 			  (##sys#dload (##sys#make-c-string (##sys#string-append "./" fname)) topentry #t) ) ) )
 	    (call-with-current-continuation
 	     (lambda (abrt)
-	       (fluid-let ([##sys#read-error-with-line-number #t]
-			   [##sys#current-source-filename fname]
-			   [##sys#current-load-path
+	       (fluid-let ((##sys#read-error-with-line-number #t)
+			   (##sys#current-source-filename fname)
+			   (##sys#current-load-path
 			    (and fname
 				 (let ((i (has-sep? fname)))
-				   (if i (##sys#substring fname 0 (fx+ i 1)) "") ) ) ]
-			   [##sys#abort-load (lambda () (abrt #f))] )
-		 (let ([in (if fname (open-input-file fname) input)])
+				   (if i (##sys#substring fname 0 (fx+ i 1)) "") ) ) )
+			   (##sys#abort-load (lambda () (abrt #f))) )
+		 (let ((in (if fname (open-input-file fname) input)))
 		   (##sys#dynamic-wind
 		    (lambda () #f)
 		    (lambda ()
-		      (let ([c1 (peek-char in)])
+		      (let ((c1 (peek-char in)))
 			(when (char=? c1 (integer->char 127))
 			  (##sys#error 'load "unable to load compiled module" fname _dlerror) ) )
 		      (let ((x1 (read in)))
@@ -998,13 +995,23 @@
     (else default-dynamic-load-libraries) ) )
 
 (define dynamic-load-libraries 
-  (make-parameter
-   (map (cut ##sys#string-append <> ##sys#load-library-extension) ##sys#default-dynamic-load-libraries)
-   (lambda (x)
-     (##sys#check-list x)
-     x) ) )
+  (let ((ext
+	 (if (and (memq (software-version) '(linux netbsd openbsd freebsd))
+		  (not (zero? binary-version))) ; allow "configless" build
+	     (string-append
+	      ##sys#load-library-extension
+	      "." 
+	      (number->string binary-version))
+	     ##sys#load-library-extension)))
+    (define complete
+      (cut ##sys#string-append <> ext))
+    (make-parameter
+     (map complete ##sys#default-dynamic-load-libraries)
+     (lambda (x)
+       (##sys#check-list x)
+       x) ) ) )
 
-(define ##sys#load-library
+(define ##sys#load-library-0
   (let ([load-verbose load-verbose]
 	[string-append string-append]
 	[dynamic-load-libraries dynamic-load-libraries]
@@ -1034,12 +1041,14 @@
 		       #t]
 		      [else (loop (##sys#slot libs 1))] ) ) ) ) ) ) ) )
 
-(define load-library
+(define ##sys#load-library
   (lambda (uname . lib)
     (##sys#check-symbol uname 'load-library)
-    (or (##sys#load-library uname (and (pair? lib) (car lib)))
+    (or (##sys#load-library-0 uname (and (pair? lib) (car lib)))
 	(##sys#error 'load-library "unable to load library" uname _dlerror) ) ) )
 
+(define load-library ##sys#load-library)
+
 (define ##sys#split-at-separator
   (let ([reverse reverse] )
     (lambda (str sep)
@@ -1052,6 +1061,23 @@
 		   (loop (cons (##sys#substring str j i) items) i2 i2) ) ]
 		[else (loop items (fx+ i 1) j)] ) ) ) ) ) )
 
+(define ##sys#include-forms-from-file
+  (let ((load-verbose load-verbose)
+	(print print)
+	(with-input-from-file with-input-from-file)
+	(read read)
+	(reverse reverse))
+    (lambda (fname)
+      (let ((path (##sys#resolve-include-filename fname #t)))
+	(when (load-verbose) (print "; including " path " ..."))
+	(with-input-from-file path
+	  (lambda ()
+	    (fluid-let ((##sys#current-source-filename path))
+	      (do ((x (read) (read))
+		   (xs '() (cons x xs)) )
+		  ((eof-object? x) 
+		   (reverse xs))) ) ) ) ) ) ) )
+
 
 ;;; Extensions:
 
@@ -1085,13 +1111,19 @@
 		  [else p] ) ) ) ) ) ) )
 
 (define ##sys#repository-path
-  (make-parameter 
-   (or (get-environment-variable repository-environment-variable)
-       (##sys#chicken-prefix 
-	(##sys#string-append 
-	 "lib/chicken/"
-	 (##sys#number->string (or (##sys#fudge 42) default-binary-version)) ) )
-       install-egg-home) ) )
+  (let ((rpath
+	 (if (##sys#fudge 22)		; private repository?
+	     (foreign-value "C_private_repository_path()" c-string)
+	     (or (get-environment-variable repository-environment-variable)
+		 (##sys#chicken-prefix 
+		  (##sys#string-append 
+		   "lib/chicken/"
+		   (##sys#number->string (##sys#fudge 42))) )
+		 install-egg-home))))
+    (lambda (#!optional val)
+      (if val
+	  (set! rpath val)
+	  rpath))))
 
 (define repository-path ##sys#repository-path)
 
@@ -1124,20 +1156,22 @@
 
 (define ##sys#load-extension
   (let ((string->symbol string->symbol))
-    (lambda (id loc . err?)
+    (lambda (id loc #!optional (err? #t))
       (cond ((string? id) (set! id (string->symbol id)))
 	    (else (##sys#check-symbol id loc)) )
       (let ([p (##sys#canonicalize-extension-path id loc)])
 	(cond ((member p ##sys#loaded-extensions))
 	      ((memq id ##sys#core-library-modules)
-	       (##sys#load-library id #f) )
+	       (or (##sys#load-library-0 id #f)
+		   (and err?
+			(##sys#error loc "cannot load core library" id))))
 	      (else
 	       (let ([id2 (##sys#find-extension p #t)])
 		 (cond (id2
 			(##sys#load id2 #f #f)
 			(set! ##sys#loaded-extensions (cons p ##sys#loaded-extensions)) 
 			#t)
-		       ((optional err? #t) (##sys#error loc "cannot load extension" id))
+		       (err? (##sys#error loc "cannot load extension" id))
 		       (else #f) ) ) ) ) ) ) ) )
 
 (define (##sys#provide . ids)
@@ -1202,132 +1236,95 @@
 	  (##sys#hash-table-update!
 	   ##compiler#file-requirements
 	   (if syntax? 'dynamic/syntax 'dynamic)
-	   (cut lset-adjoin eq? <> id)	;XXX assumes compiler has srfi-1 loaded
+	   (cut lset-adjoin eq? <> id) ;XXX assumes compiler has srfi-1 loaded
 	   (lambda () (list id)))))
       (define (impform x id builtin?)
 	`(##core#begin
-	   ,x
-	   ,@(if (and imp? (or (not builtin?) (##sys#current-module)))
-		 `((import ,id))	;XXX make hygienic
-		 '())))
-      (define (doit id)
+	  ,x
+	  ,@(if (and imp? (or (not builtin?) (##sys#current-module)))
+		`((import ,id))		;XXX make hygienic
+		'())))
+      (define (doit id impid)
 	(cond ((or (memq id builtin-features)
 		   (if comp?
 		       (memq id builtin-features/compiled)
 		       (##sys#feature? id) ) )
-	       (values (impform '(##core#undefined) id #t) #t) )
+	       (values (impform '(##core#undefined) impid #t) #t) )
 	      ((memq id ##sys#core-library-modules)
 	       (values
 		(impform
 		 (if comp?
 		     `(##core#declare (uses ,id))
 		     `(##sys#load-library ',id #f) )
-		 id #t)
+		 impid #t)
 		#t) )
 	      ((memq id ##sys#explicit-library-modules)
 	       (let* ((info (##sys#extension-information id 'require-extension))
+		      (nr (assq 'import-only info))
 		      (s (assq 'syntax info)))
 		 (values
 		  `(##core#begin
-		     ,@(if s `((##core#require-for-syntax ',id)) '())
-		     ,(impform
-		       (if comp?
-			   `(##core#declare (uses ,id)) 
-			   `(##sys#load-library ',id #f) )
-		       id #f))
+		    ,@(if s `((##core#require-for-syntax ',id)) '())
+		    ,(impform
+		      (if (not nr)
+			  (if comp?
+			      `(##core#declare (uses ,id)) 
+			      `(##sys#load-library ',id #f) )
+			  '(##core#undefined))
+		      impid #f))
 		  #t) ) )
 	      (else
 	       (let ((info (##sys#extension-information id 'require-extension)))
 		 (cond (info
 			(let ((s (assq 'syntax info))
+			      (nr (assq 'import-only info))
 			      (rr (assq 'require-at-runtime info)) )
 			  (when s (add-req id #t))
 			  (values 
 			   (impform
 			    `(##core#begin
-			       ,@(if s `((##core#require-for-syntax ',id)) '())
-			       ,@(if (and (not rr) s)
-				     '()
-				     `((##sys#require
-					,@(map (lambda (id) `',id)
-					       (cond (rr (cdr rr))
-						     (else (list id)) ) ) ) ) ) )
-			    id #f)
+			      ,@(if s `((##core#require-for-syntax ',id)) '())
+			      ,@(if (or nr (and (not rr) s))
+				    '()
+				    `((##sys#require
+				       ,@(map (lambda (id) `',id)
+					      (cond (rr (cdr rr))
+						    (else (list id)) ) ) ) ) ) )
+			    impid #f)
 			   #t) ) )
 		       (else
 			(add-req id #f)
 			(values
 			 (impform
 			  `(##sys#require ',id) 
-			  id #f)
+			  impid #f)
 			 #f)))))))
-      (if (and (pair? id) (symbol? (car id)))
-	  (let ((a (assq (##sys#slot id 0) ##sys#extension-specifiers)))
-	    (if a
-		(let ((a ((##sys#slot a 1) id)))
-		  (cond ((string? a) (values `(load ,a) #f))
-			((vector? a) 
-			 (let loop ((specs (vector->list a))
-				    (exps '())
-				    (f #f) )
-			   (if (null? specs)
-			       (values `(##core#begin ,@(reverse exps)) f)
-			       (let-values (((exp fi) (##sys#do-the-right-thing (car specs) comp? imp?)))
-				 (loop (cdr specs)
-				       (cons exp exps)
-				       (or fi f) ) ) ) ) )
-			(else (##sys#do-the-right-thing a comp? imp?)) ) )
-		(##sys#error "undefined extension specifier" id) ) )
-	  (if (symbol? id)
-	      (doit id) 
-	      (##sys#error "invalid extension specifier" id) ) ) ) ) )
-
-(define ##sys#extension-specifiers '())
-
-(define (set-extension-specifier! name proc)
-  (##sys#check-symbol name 'set-extension-specifier!)
-  (let ([a (assq name ##sys#extension-specifiers)])
-    (if a
-	(let ([old (##sys#slot a 1)])
-	  (##sys#setslot a 1 (lambda (spec) (proc spec old))) )
-	(set! ##sys#extension-specifiers
-	  (cons (cons name (lambda (spec) (proc spec #f)))
-		##sys#extension-specifiers)) ) ) )
-
-
-;;; SRFI-55
-
-(set-extension-specifier!
- 'srfi 
- (let ([list->vector list->vector])
-   (lambda (spec old)
-     (list->vector
-      (let loop ([ids (cdr spec)])
-	(if (null? ids)
-	    '()
-	    (let ([id (car ids)])
-	      (##sys#check-exact id 'require-extension)
-	      (cons (##sys#string->symbol (##sys#string-append "srfi-" (number->string id)))
-		    (loop (cdr ids)) ) ) ) ) ) ) ) )
-
-
-;;; Version checking
-
-(set-extension-specifier!
- 'version
- (lambda (spec _)
-   (define (->string x)
-     (cond ((string? x) x)
-	   ((symbol? x) (##sys#slot x 1))
-	   ((number? x) (##sys#number->string x))
-	   (else (error "invalid extension version" x)) ) )
-   (if (and (list? spec) (fx= 3 (length spec)))
-       (let* ((info (extension-information (cadr spec)))
-	      (vv (and info (assq 'version info))) )
-	 (unless (and vv (string>=? (->string (car vv)) (->string (caddr spec))))
-	   (error "installed extension does not match required version" id vv (caddr spec)))
-	 id) 
-       (##sys#syntax-error-hook "invalid version specification" spec)) ) )
+      (cond ((and (pair? id) (symbol? (car id)))
+	     (case (car id)
+	       ((srfi)
+		(let* ((f #f)
+		       (exp
+			`(##core#begin
+			  ,@(map (lambda (n)
+				   (unless (fixnum? n)
+				     (##sys#syntax-error 'require-extension "invalid SRFI number" n))
+				   (let ((rid (string->symbol (string-append "srfi-" (number->string n)))))
+				     (let-values (((exp f2) (doit rid rid)))
+				       (set! f (or f f2))
+				       exp)))
+				 (cdr id)))))
+		  (values exp f)))
+	       ((rename except only prefix)
+		(doit
+		 (let follow ((id2 id))
+		   (if (and (pair? id2) (pair? (cdr id2)))
+		       (follow (cadr id2))
+		       id2))
+		 id))
+	       (else (##sys#error "invalid extension specifier" id) ) ) )
+	    ((symbol? id)
+	     (doit id id))
+	    (else (##sys#error "invalid extension specifier" id) ) ) )))
 
 
 ;;; Convert string into valid C-identifier:
@@ -1567,8 +1564,10 @@
 	(for-each (cut ##sys#repl-print-hook <> ##sys#standard-error) xs) )
 
       (define (write-results xs)
-	(unless (or (null? xs) (eq? (##core#undefined) (car xs)))
-	  (for-each (cut ##sys#repl-print-hook <> ##sys#standard-output) xs) ) )
+	(cond ((null? xs)
+	       (##sys#print "; no values\n" #f ##sys#standard-output))
+	      ((not (eq? (##core#undefined) (car xs)))
+	       (for-each (cut ##sys#repl-print-hook <> ##sys#standard-output) xs) ) ) )
 
       (let ((stdin ##sys#standard-input)
 	    (stdout ##sys#standard-output)
@@ -1718,7 +1717,7 @@
   #t)
 
 (define-external (CHICKEN_yield) bool
-  (run-safe (lambda () (begin (thread-yield!) #t))) )
+  (run-safe (lambda () (begin (##sys#thread-yield!) #t))) )
 
 (define-external (CHICKEN_eval (scheme-object exp) ((c-pointer "C_word") result)) bool
   (run-safe
