@@ -1,7 +1,7 @@
 ;;; extras.scm - Optional non-standard extensions
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -42,13 +42,12 @@
       ##sys#substring ##sys#for-each ##sys#map ##sys#setslot
       ##sys#allocate-vector ##sys#check-pair ##sys#error-not-a-proper-list
       ##sys#member ##sys#assoc ##sys#error ##sys#signal-hook ##sys#read-string!
-      ##sys#check-symbol ##sys#check-vector ##sys#floor ##sys#ceiling
-      ##sys#truncate ##sys#round ##sys#check-number ##sys#cons-flonum
+      ##sys#check-symbol ##sys#check-vector##sys#check-number
       ##sys#flonum-fraction ##sys#make-port ##sys#fetch-and-check-port-arg
       ##sys#print ##sys#check-structure ##sys#make-structure make-parameter
       ##sys#flush-output ##sys#write-char-0 ##sys#number->string
       ##sys#fragments->string ##sys#symbol->qualified-string
-      reverse-string-append ##sys#number? ##sys#procedure->string
+      ##sys#number? ##sys#procedure->string
       ##sys#pointer->string ##sys#user-print-hook ##sys#peek-char-0
       ##sys#read-char-0 ##sys#write-char ##sys#string-append ##sys#gcd ##sys#lcm
       ##sys#fudge ##sys#check-list ##sys#user-read-hook ##sys#check-closure ##sys#check-inexact
@@ -59,7 +58,7 @@
 
 (declare
   (hide
-    fprintf0 generic-write reverse-string-append) )
+    fprintf0 generic-write) )
 
 (include "unsafe-declarations.scm")
 
@@ -85,48 +84,23 @@
 
 ;;; Random numbers:
 
-(cond-expand
-  (unix
+(define random-seed
+  (let ((srand (foreign-lambda void "srand" unsigned-integer)))
+    (lambda n
+      (let ((t (if (null? n) (current-seconds) (car n))))
+	(##sys#check-integer t 'random-seed)
+	(srand t) ) ) ) )
 
-    (define random-seed)
-    (define randomize)
+(define (randomize . n)
+  (let ((nn (if (null? n) (##sys#fudge 2) (car n))))
+    (##sys#check-exact nn 'randomize)
+    (##core#inline "C_randomize" nn) ) )
 
-    (let ((srandom (foreign-lambda void "srandom" unsigned-integer)))
-
-      (set! random-seed
-        (lambda (#!optional (seed (current-seconds)))
-          (##sys#check-integer seed 'random-seed)
-          (srandom seed) ) )
-
-      (set! randomize
-        (lambda (#!optional (seed (##sys#fudge 2)))
-          (##sys#check-exact seed 'randomize)
-          (srandom seed) ) ) )
-
-    (define (random n)
-      (##sys#check-integer n 'random)
-      (if (eq? 0 n)
-          0
-          ((foreign-lambda* long ((integer64 n)) "return( random() % ((uint64_t) n) );") n) ) ) )
-  (else
-
-    (define random-seed
-      (let ((srand (foreign-lambda void "srand" unsigned-integer)))
-        (lambda n
-          (let ((t (if (null? n) (current-seconds) (car n))))
-            (##sys#check-integer t 'random-seed)
-            (srand t) ) ) ) )
-
-    (define (randomize . n)
-      (let ((nn (if (null? n) (##sys#fudge 2) (car n))))
-        (##sys#check-exact nn 'randomize)
-        (##core#inline "C_randomize" nn) ) )
-
-    (define (random n)
-      (##sys#check-exact n 'random)
-      (if (eq? n 0)
-          0
-          (##core#inline "C_random_fixnum" n) ) ) ) )
+(define (random n)
+  (##sys#check-exact n 'random)
+  (if (eq? n 0)
+      0
+      (##core#inline "C_random_fixnum" n) ) )
 
 
 ;;; Line I/O:
@@ -602,25 +576,6 @@
 	  (out (make-string 1 #\newline) (pp obj 0))
 	  (wr obj 0)))) )
 
-; (reverse-string-append l) = (apply string-append (reverse l))
-
-(define (reverse-string-append l)
-
-  (define (rev-string-append l i)
-    (if (pair? l)
-      (let* ((str (car l))
-	     (len (string-length str))
-	     (result (rev-string-append (cdr l) (+ i len))))
-	(let loop ((j 0) (k (- (- (string-length result) i) len)))
-	  (if (< j len)
-	    (begin
-	      (string-set! result k (string-ref str j))
-	      (loop (+ j 1) (+ k 1)))
-	    result)))
-      (make-string i)))
-
-  (rev-string-append l 0))
-
 ; (pretty-print obj port) pretty prints 'obj' on 'port'.  The current
 ; output port is used if 'port' is not specified.
 
@@ -656,7 +611,7 @@
 	      (set! index (fx+ index 1))
 	      c) )
 	  (define (next)
-	    (if (cond-expand [unsafe #f] [else (##core#inline "C_eqp" args '())])
+	    (if (##core#inline "C_eqp" args '())
 		(##sys#error loc "too few arguments to formatted output procedure")
 		(let ((x (##sys#slot args 0)))
 		  (set! args (##sys#slot args 1)) 
