@@ -1,7 +1,7 @@
 ;;;; files.scm - File and pathname operations
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
@@ -39,7 +39,7 @@
   (uses regex data-structures)
   (usual-integrations)
   (fixnum)
-  (hide chop-pds)
+  (hide chop-pds absolute-pathname-root root-origin root-directory split-directory)
   (disable-interrupts) )
 
 (cond-expand
@@ -68,7 +68,7 @@
   (let ([file-exists? file-exists?]
 	[delete-file delete-file] )
     (lambda (file)
-      (and (file-exists? file) (delete-file file) #t) ) ) )
+      (and (file-exists? file) (delete-file file)) ) ) )
 
 ;;; file-copy and file-move : they do what you'd think.
 (define (file-copy origfile newfile #!optional (clobber #f) (blocksize 1024))
@@ -164,16 +164,31 @@
 
 ;;; Pathname operations:
 
-(define absolute-pathname?
-  (let ([string-match string-match]
-        [regexp regexp]
-        [string-append string-append])
-    (let* ([drv (if ##sys#windows-platform "([A-Za-z]:)?" "")]
-           [patt (string-append drv "[\\/\\\\].*")]
-	   [rx (regexp patt)] )
-      (lambda (pn)
-        (##sys#check-string pn 'absolute-pathname?)
-        (pair? (string-match rx pn)) ) ) ) )
+;; Platform specific absolute pathname operations:
+;; absolute-pathname-root => #f or (<match> [<origin>] <root>)
+;;
+;; Not for general consumption
+
+(define absolute-pathname-root)
+(define root-origin)
+(define root-directory)
+
+(let ((string-match string-match))
+  (if ##sys#windows-platform
+      (let ((rx (regexp "([A-Za-z]:)?([\\/\\\\]).*")))
+        (set! absolute-pathname-root (lambda (pn) (string-match rx pn)))
+        (set! root-origin (lambda (rt) (and rt (cadr rt))))
+        (set! root-directory (lambda (rt) (and rt (caddr rt)))) )
+      (let ((rx (regexp "([\\/\\\\]).*")))
+        (set! absolute-pathname-root (lambda (pn) (string-match rx pn)))
+        (set! root-origin (lambda (rt) #f))
+        (set! root-directory (lambda (rt) (and rt (cadr rt)))) ) ) )
+
+(define (absolute-pathname? pn)
+  (##sys#check-string pn 'absolute-pathname?)
+  (pair? (absolute-pathname-root pn)) )
+
+(define-inline (*char-pds? ch) (memq ch '(#\\ #\/)))
 
 (define (chop-pds str pds)
   (and str
@@ -182,8 +197,7 @@
 	 (if (and (fx>= len 1)
 		  (if pds
 		      (##core#inline "C_substring_compare" str pds (fx- len pdslen) 0 pdslen)
-		      (memq (##core#inline "C_subchar" str (fx- len pdslen))
-			    '(#\/ #\\) ) ) )
+		      (*char-pds? (##core#inline "C_subchar" str (fx- len pdslen)) ) ) )
 	     (##sys#substring str 0 (fx- len pdslen))
 	     str) ) ) )
 
@@ -224,7 +238,7 @@
        (if (and (fx>= (##sys#size file) pdslen)
 		(if pds
                     (##core#inline "C_substring_compare" pds file 0 0 pdslen)
-                    (memq (##core#inline "C_subchar" file 0) '(#\\ #\/))))
+                    (*char-pds? (##core#inline "C_subchar" file 0))))
 	   (##sys#substring file pdslen (##sys#size file))
 	   file)
        (if (and (fx> (##sys#size ext) 0)
@@ -234,11 +248,11 @@
        ext) ) )
 
   (set! make-pathname
-    (lambda (dirs file #!optional ext pds)
+    (lambda (dirs file #!optional ext pds) ; The 'pds' argument is DEPRECATED
       (_make-pathname 'make-pathname (canonicalize-dirs dirs pds) file ext pds)))
 
   (set! make-absolute-pathname
-    (lambda (dirs file #!optional ext pds)
+    (lambda (dirs file #!optional ext pds) ; The 'pds' argument is DEPRECATED
       (_make-pathname
        'make-absolute-pathname
        (let ([dir (canonicalize-dirs dirs pds)])
@@ -248,19 +262,17 @@
        file ext pds) ) ) )
 
 (define decompose-pathname
-  (let ([string-match string-match]
-        [regexp regexp]
-        [string-append string-append])
+  (let ((string-match string-match))
     (let* ([patt1 "^(.*[\\/\\\\])?([^\\/\\\\]+)(\\.([^\\/\\\\.]+))$"]
 	   [patt2 "^(.*[\\/\\\\])?((\\.)?[^\\/\\\\]+)$"]
 	   [rx1 (regexp patt1)]
 	   [rx2 (regexp patt2)]
 	   [strip-pds
 	     (lambda (dir)
-	        (and dir
-		     (if (member dir '("/" "\\"))
-		         dir
-		         (chop-pds dir #f) ) ) )] )
+	       (and dir
+		    (if (member dir '("/" "\\"))
+		        dir
+		        (chop-pds dir #f) ) ) )] )
       (lambda (pn)
         (##sys#check-string pn 'decompose-pathname)
         (if (fx= 0 (##sys#size pn))
@@ -350,16 +362,14 @@
 	(get-output-string get-output-string)
 	(get-environment-variable get-environment-variable)
 	(reverse reverse)
-	(display display))
-    (lambda (path #!optional (platform (if (memq (build-platform) '(msvc mingw32)) 'windows 'unix)))
+	(display display)
+	(bldplt (if (memq (build-platform) '(msvc mingw32)) 'windows 'unix)) )
+    (define (addpart part parts)
+      (cond ((string=? "." part)        parts )
+            ((string=? ".." part)       (if (null? parts) '("..") (cdr parts)) )
+            (else                       (cons part parts) ) ) )
+    (lambda (path #!optional (platform bldplt))
       (let ((sep (if (eq? platform 'windows) #\\ #\/)))
-	(define (addpart part parts)
-	  (cond ((string=? "." part) parts)
-		((string=? ".." part) 
-		 (if (null? parts)
-		     '("..")
-		     (cdr parts)))
-		(else (cons part parts))))
 	(##sys#check-string path 'normalize-pathname)
 	(let ((len (##sys#size path))
 	      (abspath #f)
@@ -369,7 +379,9 @@
 		   (when (fx> i prev)
 		     (set! parts (addpart (##sys#substring path prev i) parts)))
 		   (if (null? parts)
-		       (##sys#string-append "." (string sep))
+		       (if abspath
+		           (##sys#string-append (string sep) ".")
+		           (##sys#string-append "." (string sep)) )
 		       (let ((out (open-output-string))
 			     (parts (reverse parts)))
 			 (display (car parts) out)
@@ -387,7 +399,7 @@
 			     (when drive
 			       (set! r (##sys#string-append drive r))))
 			   r))))
-		  ((memq (string-ref path i) '(#\\ #\/))
+		  ((*char-pds? (string-ref path i))
 		   (when (and (null? parts) (fx= i prev))
 		     (set! abspath #t))
 		   (if (fx= i prev)
@@ -400,19 +412,49 @@
 			(eq? 'windows platform))
 		   (set! drive (##sys#substring path 0 (fx+ i 1)))
 		   (loop (fx+ i 1) (fx+ i 1) '()))
-		  (else (loop (fx+ i 1) prev parts)))))))))
+		  (else (loop (fx+ i 1) prev parts)) ) ) ) ) ) ) )
 
 
+;; directory pathname => list of strings
+;; does arg check
+
+(define split-directory
+  (let ((string-split string-split) )
+    (lambda (loc dir keep?)
+      (##sys#check-string dir loc)
+      (string-split dir "/\\" keep?) ) ) )
+
 ;; Directory string or list only contains path-separators
-;; and/or current-directory names.
+;; and/or current-directory (".") names.
 
 (define (directory-null? dir)
-  (let loop ([lst
-              (if (list? dir)
-                  dir ; Don't bother to check for strings here
-                  (begin
-                    (##sys#check-string dir 'directory-null?)
-                    (string-split dir "/\\" #t)))])
-    (or (null? lst)
-        (and (member (car lst) '("" "."))
-             (loop (cdr lst)) ) ) ) )
+  (let loop ((ls (if (list? dir) dir (split-directory 'directory-null? dir #t))))
+    (or (null? ls)
+        (and (member (car ls) '("" "."))
+             (loop (cdr ls)) ) ) ) )
+
+;; Directory string => {<origin> <root> <directory-list>}
+;; where any maybe #f when missing
+
+(define (decompose-directory dir)
+  (define (strip-origin-prefix org decomp)
+    #;(assert (or (not org) decomp)) ;cannot have an "origin" but no "decomp"
+    (if (not org)
+        decomp
+        (let ((1st (car decomp)))
+          (let ((olen (##sys#size org)))
+            (if (not (##core#inline "C_substring_compare" org 1st 0 0 olen))
+                ; then origin is not a prefix (really shouldn't happen)
+                decomp
+                ; else is a prefix
+                (let ((rst (cdr decomp))
+                      (elen (##sys#size 1st)) )
+                  (if (fx= olen (##sys#size elen))
+                      ; then origin is a list prefix
+                      rst
+                      ; else origin is a string prefix
+                      (cons (##sys#substring 1st olen elen) rst) ) ) ) ) ) ) )
+  (let* ((ls (split-directory 'decompose-directory dir #f))
+         (rt (absolute-pathname-root dir))
+         (org (root-origin rt)) )
+    (values org (root-directory rt) (strip-origin-prefix org (and (not (null? ls)) ls))) ) )
