@@ -1,7 +1,7 @@
 ;;;; lolevel.scm - Low-level routines for CHICKEN
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -30,11 +30,12 @@
   (uses srfi-69)
   (usual-integrations)
   (disable-warning var redef)
-  (hide ipc-hook-0 xproc-tag
+  (hide ipc-hook-0 *set-invalid-procedure-call-handler! xproc-tag
    ##sys#check-block
    ##sys#check-become-alist
    ##sys#check-generic-structure
    ##sys#check-generic-vector )
+  (not inline ipc-hook-0 ##sys#invalid-procedure-call-hook)
   (foreign-declare #<<EOF
 #if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
 # include <sys/types.h>
@@ -44,7 +45,6 @@
 #endif
 
 #define C_w2b(x)                   C_fix(C_wordstobytes(C_unfix(x)))
-#define C_pointer_eqp(x, y)        C_mk_bool(C_c_pointer_nn(x) == C_c_pointer_nn(y))
 #define C_memmove_o(to, from, n, toff, foff) C_memmove((char *)(to) + (toff), (char *)(from) + (foff), (n))
 EOF
 ) )
@@ -57,7 +57,7 @@ EOF
     (no-procedure-checks-for-usual-bindings)
     (bound-to-procedure
      ##sys#check-pointer ##sys#check-closure ##sys#check-integer ##sys#check-special
-     ##sys#error ##sys#signal-hook 
+     ##sys#error ##sys#signal-hook ##sys#error-hook 
      ##sys#error-not-a-proper-list
      make-hash-table hash-table-ref/default hash-table-set!
      ##sys#make-pointer ##sys#make-tagged-pointer ##sys#make-locative ##sys#locative?
@@ -77,7 +77,7 @@ EOF
 ;;; Helpers:
 
 (define-inline (%pointer? x)
-  (and (##core#inline "C_blockp" x) (##core#inline "C_anypointerp" x)) )
+  (##core#inline "C_i_safe_pointerp" x))
 
 (define-inline (%generic-pointer? x)
   (or (%pointer? x)
@@ -141,25 +141,6 @@ EOF
      (and (pair? loc) (car loc))
      "bad argument type - not a pointer" x) ) )
 
-(cond-expand
-  [unsafe
-   (define-syntax ##sys#check-pointer
-     (syntax-rules ()
-       ((_ . _) (##core#undefined))))
-   (define-syntax ##sys#check-block
-     (syntax-rules ()
-       ((_ . _) (##core#undefined))))
-   (define-syntax ##sys#check-become-alist
-     (syntax-rules ()
-       ((_ . _) (##core#undefined))))
-   (define-syntax ##sys#check-generic-structure
-     (syntax-rules ()
-       ((_ . _) (##core#undefined))))
-   (define-syntax ##sys#check-generic-vector
-     (syntax-rules ()
-       ((_ . _) (##core#undefined)))) ]
-  [else] )
-
 
 ;;; Move arbitrary blocks of memory around:
 
@@ -184,17 +165,21 @@ EOF
 	(apply ##sys#error 'move-memory! "number of bytes to move too large" from to args))
       ;
       (define (checkn1 n nmax off)
-	(if (cond-expand [unsafe #t] [else (fx<= n (fx- nmax off))])
+	(if (fx<= n (fx- nmax off))
 	    n
 	    (sizerr n nmax) ) )
       ;
       (define (checkn2 n nmax nmax2 off1 off2)
-	(if (cond-expand [unsafe #t] [else (and (fx<= n (fx- nmax off1)) (fx<= n (fx- nmax2 off2)))])
+	(if (and (fx<= n (fx- nmax off1)) (fx<= n (fx- nmax2 off2)))
 	    n
 	    (sizerr n nmax nmax2) ) )
       ;
       (##sys#check-block from 'move-memory!)
       (##sys#check-block to 'move-memory!)
+      (when (fx< foffset 0)
+	(##sys#error 'move-memory! "negative source offset" foffset))
+      (when (fx< toffset 0)
+	(##sys#error 'move-memory! "negative destination offset" toffset))
       (let move ([from from] [to to])
 	(cond [(##sys#generic-structure? from)
 	       (if (memq (##sys#slot from 0) slot1structs)
@@ -262,7 +247,7 @@ EOF
 
 (define (null-pointer? ptr)
   (##sys#check-special ptr 'null-pointer?)
-  (eq? 0 (##sys#pointer->address ptr) ) )
+  (##core#inline "C_null_pointerp" ptr))
 
 (define (object->pointer x)
   (and (##core#inline "C_blockp" x)
@@ -277,10 +262,12 @@ EOF
   (##sys#check-special p2 'pointer=?)
   (##core#inline "C_pointer_eqp" p1 p2) )
 
-(define pointer-offset
+(define pointer+
   (foreign-lambda* nonnull-c-pointer ([c-pointer ptr] [integer off])
     "return((unsigned char *)ptr + off);") )
 
+(define pointer-offset pointer+)	; DEPRECATED
+
 (define align-to-word
   (let ([align (foreign-lambda integer "C_align" integer)])
     (lambda (x)
@@ -348,54 +335,54 @@ EOF
 
 ;;; SRFI-4 number-vector:
 
-(define pointer-u8-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((unsigned char *)p) = n;"))
-(define pointer-s8-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((char *)p) = n;"))
-(define pointer-u16-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((unsigned short *)p) = n;"))
-(define pointer-s16-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((short *)p) = n;"))
-(define pointer-u32-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((C_u32 *)p) = n;"))
-(define pointer-s32-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((C_s32 *)p) = n;"))
-(define pointer-f32-set! (foreign-lambda* void ([c-pointer p] [double n]) "*((float *)p) = n;"))
-(define pointer-f64-set! (foreign-lambda* void ([c-pointer p] [float n]) "*((double *)p) = n;"))
+(define (pointer-u8-set! p n) (##core#inline "C_u_i_pointer_u8_set" p n))
+(define (pointer-s8-set! p n) (##core#inline "C_u_i_pointer_s8_set" p n))
+(define (pointer-u16-set! p n) (##core#inline "C_u_i_pointer_u16_set" p n))
+(define (pointer-s16-set! p n) (##core#inline "C_u_i_pointer_s16_set" p n))
+(define (pointer-u32-set! p n) (##core#inline "C_u_i_pointer_u32_set" p n))
+(define (pointer-s32-set! p n) (##core#inline "C_u_i_pointer_s32_set" p n))
+(define (pointer-f32-set! p n) (##core#inline "C_u_i_pointer_f32_set" p n))
+(define (pointer-f64-set! p n) (##core#inline "C_u_i_pointer_f64_set" p n))
 
 (define pointer-u8-ref
   (getter-with-setter
-   (foreign-lambda* int ([c-pointer p]) "return(*((unsigned char *)p));")
-   pointer-u8-set!) )
+   (lambda (p) (##core#inline "C_u_i_pointer_u8_ref" p))
+   pointer-u8-set!))
 
 (define pointer-s8-ref
   (getter-with-setter
-   (foreign-lambda* int ([c-pointer p]) "return(*((signed char *)p));")
-   pointer-s8-set!) )
+   (lambda (p) (##core#inline "C_u_i_pointer_s8_ref" p))
+   pointer-s8-set!))
 
 (define pointer-u16-ref
   (getter-with-setter
-   (foreign-lambda* int ([c-pointer p]) "return(*((unsigned short *)p));")
-   pointer-u16-set!) )
+   (lambda (p) (##core#inline "C_u_i_pointer_u16_ref" p))
+   pointer-u16-set!))
 
 (define pointer-s16-ref
   (getter-with-setter
-   (foreign-lambda* int ([c-pointer p]) "return(*((short *)p));")
-   pointer-s6-set!) )
+   (lambda (p) (##core#inline "C_u_i_pointer_s16_ref" p))
+   pointer-s16-set!))
 
 (define pointer-u32-ref
   (getter-with-setter
-   (foreign-lambda* integer ([c-pointer p]) "return(*((C_u32 *)p));")
-   pointer-u32-set!) )
+   (lambda (p) (##core#inline_allocate ("C_a_u_i_pointer_u32_ref" 4) p)) ;XXX hardcoded size
+   pointer-u32-set!))
 
 (define pointer-s32-ref
   (getter-with-setter
-   (foreign-lambda* integer ([c-pointer p]) "return(*((C_s32 *)p));")
-   pointer-s32-set!) )
+   (lambda (p) (##core#inline_allocate ("C_a_u_i_pointer_s32_ref" 4) p)) ;XXX hardcoded size
+   pointer-s32-set!))
 
 (define pointer-f32-ref
   (getter-with-setter
-   (foreign-lambda* float ([c-pointer p]) "return(*((float *)p));")
-   pointer-f32-set!) )
+   (lambda (p) (##core#inline_allocate ("C_a_u_i_pointer_f32_ref" 4) p)) ;XXX hardcoded size
+   pointer-f32-set!))
 
 (define pointer-f64-ref
   (getter-with-setter
-   (foreign-lambda* double ([c-pointer p]) "return(*((double *)p));")
-   pointer-f64-set!) )
+   (lambda (p) (##core#inline_allocate ("C_a_u_i_pointer_f64_ref" 4) p)) ;XXX hardcoded size
+   pointer-f64-set!))
 
 
 ;;; Procedures extended with data:
@@ -439,9 +426,6 @@ EOF
 (define block-set! ##sys#block-set!)
 (define block-ref (getter-with-setter ##sys#block-ref ##sys#block-set!))
 
-(define (vector-like? x)
-  (%generic-vector? x) )
-
 (define (number-of-slots x)
   (##sys#check-generic-vector x 'number-of-slots)
   (##sys#size x) )
@@ -532,7 +516,7 @@ EOF
 	       y ) ] ) ) ) )
 
 (define (object-evict-to-location x ptr . limit)
-  (cond-expand [(not unsafe) (##sys#check-special ptr 'object-evict-to-location)] [else])
+  (##sys#check-special ptr 'object-evict-to-location)
   (let* ([limit (and (pair? limit)
 		     (let ([limit (car limit)])
 		       (##sys#check-exact limit 'object-evict-to-location)
@@ -633,7 +617,7 @@ EOF
 ;;; `become':
 
 (define (object-become! alst)
-  (cond-expand [(not unsafe) (##sys#check-become-alist alst 'object-become!)] [else])
+  (##sys#check-become-alist alst 'object-become!)
   (##sys#become! alst) )
 
 (define (mutate-procedure old proc)
@@ -646,23 +630,6 @@ EOF
     new ) )
 
 
-;;; Hooks:
-
-(define ipc-hook-0 #f)			; we need this because `##sys#invalid-procedure-call-hook' may not have free variables.
-
-(define (set-invalid-procedure-call-handler! proc)
-  (##sys#check-closure proc 'set-invalid-procedure-call-handler!)
-  (set! ipc-hook-0 proc)
-  (set! ##sys#invalid-procedure-call-hook 
-    (lambda args
-      (ipc-hook-0 ##sys#last-invalid-procedure args) ) ) )
-
-(define (unbound-variable-value . val)
-  (set! ##sys#unbound-variable-value-hook 
-    (and (pair? val)
-	 (vector (car val)))) )
-
-
 ;;; Access computed globals:
 
 (define (global-ref sym)
