@@ -1,7 +1,7 @@
 ;;;; optimizer.scm - The CHICKEN Scheme compiler (optimizations)
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -25,7 +25,9 @@
 ; POSSIBILITY OF SUCH DAMAGE.
 
 
-(declare (unit optimizer))
+(declare
+  (unit optimizer)
+  (not inline ##sys#compiler-syntax-hook) )
 
 
 (include "compiler-namespace")
@@ -37,50 +39,72 @@
 ;;; Scan toplevel expressions for assignments:
 
 (define (scan-toplevel-assignments node)
-  (let ([safe '()]
-	[unsafe '()] )
+  (let ((safe '())
+	(unsafe '()) 
+	(escaped #f)
+	(previous '()))
 
     (define (mark v)
-      (if (not (memq v unsafe)) (set! safe (cons v safe))) )
+      (when (and (not escaped)
+		 (not (memq v unsafe)))
+	(set! safe (cons v safe))) )
+
+    (define (remember v x)
+      (set! previous (alist-update! v x previous)))
+
+    (define (touch)
+      (set! escaped #t)
+      (set! previous '()))
+
+    (define (scan-each ns e)
+      (for-each (lambda (n) (scan n e)) ns) )
+
+    (define (scan n e)
+      (let ([params (node-parameters n)]
+	    [subs (node-subexpressions n)] )
+	(case (node-class n)
+
+	  [(##core#variable)
+	   (let ((var (first params)))
+	     (when (and (not (memq var e)) 
+			(not (memq var safe)))
+	       (set! unsafe (cons var unsafe)) )
+	     (set! previous (remove (lambda (p) (eq? (car p) var)) previous)))]
+
+	  [(if ##core#cond ##core#switch)
+	   (scan (first subs) e)
+	   (touch)
+	   (scan-each (cdr subs) e)]
+
+	  [(let)
+	   (scan (first subs) e)
+	   (scan (second subs) (append params e)) ]
+
+	  [(lambda ##core#lambda ##core#callunit) #f]
+
+	  [(##core#call) (touch)]
+
+	  [(set!)
+	   (let ((var (first params))
+		 (val (first subs)))
+	     (scan val e)
+	     (let ((p (alist-ref var previous)))
+	       (when (and p (not (memq var unsafe)))
+		 (##sys#notice
+		  (sprintf "dropping assignment of unused value to global variable `~s'"
+		    var))
+		 (copy-node!
+		  (make-node '##core#undefined '() '())
+		  p))
+	       (unless (memq var e) (mark var))
+	       (remember var n) ) ) ]
+
+	  [else (scan-each subs e)] ) ) )
 
     (debugging 'p "scanning toplevel assignments...")
-    (call-with-current-continuation
-     (lambda (return)
-
-       (define (scan-each ns e)
-	 (for-each (lambda (n) (scan n e)) ns) )
-
-       (define (scan n e)
-	 (let ([params (node-parameters n)]
-	       [subs (node-subexpressions n)] )
-	   (case (node-class n)
-
-	     [(##core#variable)
-	      (let ([var (first params)])
-		(if (and (not (memq var e)) (not (memq var safe)))
-		    (set! unsafe (cons var unsafe)) ) ) ]
-
-	     [(if ##core#cond ##core#switch)
-	      (scan (first subs) e)
-	      (return #f) ]
-
-	     [(let)
-	      (scan (first subs) e)
-	      (scan (second subs) (append params e)) ]
-
-	     [(lambda ##core#callunit) #f]
-
-	     [(##core#call) (return #f)]
-
-	     [(set!)
-	      (let ([var (first params)])
-		(if (not (memq var e)) (mark var))
-		(scan (first subs) e) ) ]
-
-	     [else (scan-each subs e)] ) ) )
-
-       (scan node '()) ) )
-    (debugging 'o "safe globals" safe)
+    (scan node '())
+    (when (pair? safe)
+      (debugging 'o "safe globals" (delete-duplicates safe eq?)))
     (for-each (cut mark-variable <> '##compiler#always-bound) safe)))
 
 
@@ -133,12 +157,12 @@
 		 entry) )
 	  n) )
 
-    (define (walk n)
+    (define (walk n fids)
       (if (memq n broken-constant-nodes)
 	  n
 	  (simplify
 	   (let* ((odirty dirty)
-		  (n1 (walk1 n))
+		  (n1 (walk1 n fids))
 		  (subs (node-subexpressions n1)) )
 	     (case (node-class n1)
 
@@ -148,7 +172,8 @@
 		       (touch)
 		       (walk (if (node-value (car subs))
 				 (cadr subs)
-				 (caddr subs) ) ) )
+				 (caddr subs) )
+			     fids) )
 		      (else n1) ) )
 
 	       ((##core#call)
@@ -176,7 +201,7 @@
 
 	       (else n1) ) ) ) ) )
 
-    (define (walk1 n)
+    (define (walk1 n fids)
       (let ((subs (node-subexpressions n))
 	    (params (node-parameters n)) 
 	    (class (node-class n)) )
@@ -198,17 +223,17 @@
 
 	  ((let)
 	   (let ([var (first params)])
-	     (cond [(or ;(test var 'replacable)
-			(test var 'removable)
+	     (cond [(or (test var 'removable)
 			(and (test var 'contractable) (not (test var 'replacing))) )
 		    (touch)
 		    (set! removed-lets (add1 removed-lets))
-		    (walk (second subs)) ]
-		   [else (make-node 'let params (map walk subs))] ) ) )
+		    (walk (second subs) fids) ]
+		   [else (make-node 'let params (map (cut walk <> fids) subs))] ) ) )
 
 	  ((##core#lambda)
-	   (let ([llist (third params)])
-	     (cond [(test (first params) 'has-unused-parameters)
+	   (let ((llist (third params))
+		 (id (first params)))
+	     (cond [(test id 'has-unused-parameters)
 		    (decompose-lambda-list
 		     llist
 		     (lambda (vars argc rest)
@@ -218,13 +243,13 @@
 			 (make-node
 			  '##core#lambda
 			  (list (first params) (second params)
-				(cond [(and rest (test (first params) 'explicit-rest))
+				(cond [(and rest (test id 'explicit-rest))
 				       (debugging 'o "merged explicitly consed rest parameter" rest)
 				       (build-lambda-list used (add1 argc) #f) ]
 				      [else (build-lambda-list used argc rest)] )
 				(fourth params) )
-			  (list (walk (first subs))) ) ) ) ) ]
-		   [(test (first params) 'explicit-rest)
+			  (list (walk (first subs) (cons id fids))) ) ) ) ) ]
+		   [(test id 'explicit-rest)
 		    (decompose-lambda-list
 		     llist
 		     (lambda (vars argc rest)
@@ -236,8 +261,8 @@
 			      (second params)
 			      (build-lambda-list vars (add1 argc) #f)
 			      (fourth params) )
-			(list (walk (first subs))) ) ) ) ]
-		   [else (walk-generic n class params subs)] ) ) )
+			(list (walk (first subs) (cons id fids))) ) ) ) ]
+		   [else (walk-generic n class params subs (cons id fids))] ) ) )
 
 	  ((##core#call)
 	   (let* ([fun (car subs)]
@@ -256,7 +281,10 @@
 			   (check-signature var args llist)
 			   (debugging 'o "contracted procedure" var)
 			   (touch)
-			   (walk (inline-lambda-bindings llist args (first (node-subexpressions lval)) #f)) ) ]
+			   (for-each (cut put! db <> 'inline-target #t) fids)
+			   (walk
+			    (inline-lambda-bindings llist args (first (node-subexpressions lval)) #f db)
+			    fids) ) ]
 			[(memq var constant-declarations)
 			 (or (and-let* ((k (car args))
 					((eq? '##core#variable (node-class k)))
@@ -272,7 +300,7 @@
 			       (make-node
 				'##core#call '(#t)
 				(list k (make-node '##core#undefined '() '())) ) ) 
-			     (walk-generic n class params subs)) ]
+			     (walk-generic n class params subs fids)) ]
 			[(and lval
 			      (eq? '##core#lambda (node-class lval)))
 			 (let* ([lparams (node-parameters lval)]
@@ -280,10 +308,10 @@
 			   (decompose-lambda-list
 			    llist
 			    (lambda (vars argc rest)
-			      (let ([fid (first lparams)])
+			      (let ([ifid (first lparams)])
 				(cond [(and inline-locally 
-					    (test fid 'simple)
 					    (test var 'inlinable)
+					    (not (test ifid 'inline-target)) ; inlinable procedure has changed
 					    (case (variable-mark var '##compiler#inline) 
 					      ((yes) #t)
 					      ((no) #f)
@@ -292,23 +320,26 @@
 				       (debugging 
 					'i
 					(if (node? (variable-mark var '##compiler#inline-global))
-					    "procedure can be inlined (globally)" 
-					    "procedure can be inlined")
-					var fid (fourth lparams))
+					    "global inlining" 
+					    "inlining")
+					var ifid (fourth lparams))
+				       (for-each (cut put! db <> 'inline-target #t) fids)
 				       (check-signature var args llist)
 				       (debugging 'o "inlining procedure" var)
 				       (touch)
-				       (walk (inline-lambda-bindings llist args (first (node-subexpressions lval)) #t)) ]
-				      [(test fid 'has-unused-parameters)
+				       (walk
+					(inline-lambda-bindings llist args (first (node-subexpressions lval)) #t db)
+					fids) ]
+				      [(test ifid 'has-unused-parameters)
 				       (if (< (length args) argc) ; Expression was already optimized (should this happen?)
-					   (walk-generic n class params subs)
+					   (walk-generic n class params subs fids)
 					   (let loop ((vars vars) (argc argc) (args args) (used '()))
 					     (cond [(or (null? vars) (zero? argc))
 						    (touch)
 						    (make-node
 						     '##core#call
 						     params
-						     (map walk (cons fun (append-reverse used args))) ) ]
+						     (map (cut walk <> fids) (cons fun (append-reverse used args))) ) ]
 						   [(test (car vars) 'unused)
 						    (touch)
 						    (debugging
@@ -318,25 +349,25 @@
 							(make-node
 							 'let
 							 (list (gensym 't))
-							 (list (walk (car args))
+							 (list (walk (car args) fids)
 							       (loop (cdr vars) (sub1 argc) (cdr args) used) ) )
 							(loop (cdr vars) (sub1 argc) (cdr args) used) ) ]
 						   [else (loop (cdr vars)
 							       (sub1 argc)
 							       (cdr args)
 							       (cons (car args) used) ) ] ) ) ) ]
-				      [(and (test fid 'explicit-rest)
+				      [(and (test ifid 'explicit-rest)
 					    (not (memq n rest-consers)) ) ; make sure we haven't inlined rest-list already
 				       (let ([n (llist-length llist)])
 					 (if (< (length args) n)
-					     (walk-generic n class params subs)
+					     (walk-generic n class params subs fids)
 					     (begin
 					       (debugging 'o "consed rest parameter at call site" var n)
 					       (let-values ([(args rargs) (split-at args n)])
 						 (let ([n2 (make-node
 							    '##core#call
 							    params
-							    (map walk
+							    (map (cut walk <> fids)
 								 (cons fun
 								       (append 
 									args
@@ -349,13 +380,13 @@
 									      rargs) ) ) ) ) ) ) ] )
 						   (set! rest-consers (cons n2 rest-consers))
 						   n2) ) ) ) ) ]
-				      [else (walk-generic n class params subs)] ) ) ) ) ) ]
-			[else (walk-generic n class params subs)] ) ) ]
+				      [else (walk-generic n class params subs fids)] ) ) ) ) ) ]
+			[else (walk-generic n class params subs fids)] ) ) ]
 	       [(##core#lambda)
 		(if (first params)
-		    (walk-generic n class params subs)
-		    (make-node '##core#call (cons #t (cdr params)) (map walk subs)) ) ]
-	       [else (walk-generic n class params subs)] ) ) )
+		    (walk-generic n class params subs fids)
+		    (make-node '##core#call (cons #t (cdr params)) (map (cut walk <> fids) subs)) ) ]
+	       [else (walk-generic n class params subs fids)] ) ) )
 
 	  ((set!)
 	   (let ([var (first params)])
@@ -364,17 +395,18 @@
 		    (make-node '##core#undefined '() '()) ]
 		   [(and (or (not (test var 'global))
 			     (not (variable-visible? var)))
+			 (not (test var 'inline-transient))
 			 (not (test var 'references)) 
 			 (not (expression-has-side-effects? (first subs) db)) )
 		    (touch)
 		    (debugging 'o "removed side-effect free assignment to unused variable" var)
 		    (make-node '##core#undefined '() '()) ]
-		   [else (make-node 'set! params (list (walk (car subs))))] ) ) )
+		   [else (make-node 'set! params (list (walk (car subs) fids)))] ) ) )
 
-	  (else (walk-generic n class params subs)) ) ) )
+	  (else (walk-generic n class params subs fids)) ) ) )
     
-    (define (walk-generic n class params subs)
-      (let ((subs2 (map walk subs)))
+    (define (walk-generic n class params subs fids)
+      (let ((subs2 (map (cut walk <> fids) subs)))
 	(if (every eq? subs subs2)
 	    n
 	    (make-node class params subs2) ) ) )
@@ -384,7 +416,7 @@
 	(begin
 	  (debugging 'p "traversal phase...")
 	  (set! simplified-ops '())
-	  (let ((node2 (walk node)))
+	  (let ((node2 (walk node '())))
 	    (when (pair? simplified-classes) (debugging 'o "simplifications" simplified-classes))
 	    (when (and (pair? simplified-ops) (debugging 'o "  call simplifications:"))
 	      (for-each
@@ -423,35 +455,35 @@
 		  (subs (node-subexpressions n))
 		  (kont (first (node-parameters (second subs))))
 		  (lnode (and (not (test kont 'unknown)) (test kont 'value)))
-		  (krefs (test kont 'references)) )
+		  (krefs (get-list db kont 'references)) )
 	     ;; Call-site has one argument and a known continuation (which is a ##core#lambda)
 	     ;;  that has only one use:
-	     (if (and lnode krefs (= 1 (length krefs)) (= 3 (length subs))
-		      (eq? '##core#lambda (node-class lnode)) )
-		 (let* ((llist (third (node-parameters lnode)))
-			(body (first (node-subexpressions lnode))) 
-			(bodysubs (node-subexpressions body)) )
-		   ;; Continuation has one parameter?
-		   (if (and (proper-list? llist) (null? (cdr llist)))
-		       (let* ((var (car llist))
-			      (refs (test var 'references)) )
-			 ;; Parameter is only used once?
-			 (if (and refs (= 1 (length refs)) (eq? 'if (node-class body)))
-			     ;; Continuation contains an 'if' node?
-			     (let ((iftest (first (node-subexpressions body))))
-			       ;; Parameter is used only once and is the test-argument?
-			       (if (and (eq? '##core#variable (node-class iftest))
-					(eq? var (first (node-parameters iftest))) )
-				   ;; Modify call-site to call continuation directly and swap branches
-				   ;;  in the conditional:
-				   (begin
-				     (set! removed-nots (+ removed-nots 1))
-				     (node-parameters-set! n '(#t))
-				     (node-subexpressions-set! n (cdr subs))
-				     (node-subexpressions-set! 
-				      body
-				      (cons (car bodysubs) (reverse (cdr bodysubs))) )
-				     (touch) ) ) ) ) ) ) ) ) ) )
+	     (when (and lnode krefs (= 1 (length krefs)) (= 3 (length subs))
+			(eq? '##core#lambda (node-class lnode)) )
+	       (let* ((llist (third (node-parameters lnode)))
+		      (body (first (node-subexpressions lnode))) 
+		      (bodysubs (node-subexpressions body)) )
+		 ;; Continuation has one parameter?
+		 (if (and (proper-list? llist) (null? (cdr llist)))
+		     (let* ((var (car llist))
+			    (refs (get-list db var 'references)) )
+		       ;; Parameter is only used once?
+		       (if (and refs (= 1 (length refs)) (eq? 'if (node-class body)))
+			   ;; Continuation contains an 'if' node?
+			   (let ((iftest (first (node-subexpressions body))))
+			     ;; Parameter is used only once and is the test-argument?
+			     (if (and (eq? '##core#variable (node-class iftest))
+				      (eq? var (first (node-parameters iftest))) )
+				 ;; Modify call-site to call continuation directly and swap branches
+				 ;;  in the conditional:
+				 (begin
+				   (set! removed-nots (+ removed-nots 1))
+				   (node-parameters-set! n '(#t))
+				   (node-subexpressions-set! n (cdr subs))
+				   (node-subexpressions-set! 
+				    body
+				    (cons (car bodysubs) (reverse (cdr bodysubs))) )
+				   (touch) ) ) ) ) ) ) ) ) ) )
 	 (or (test 'not 'call-sites) '()) ) )
     
     (when (> removed-nots 0) (debugging 'o "Removed `not' forms" removed-nots))
@@ -504,8 +536,8 @@
       (and (equal? op eq-inline-operator)
 	   (immediate? const1)
 	   (immediate? const2)
-	   (= 1 (length (get db var1 'references)))
-	   (= 1 (length (get db var2 'references)))
+	   (= 1 (length (get-list db var1 'references)))
+	   (= 1 (length (get-list db var2 'references)))
 	   (make-node
 	    '##core#switch
 	    '(2)
@@ -530,7 +562,7 @@
    ,(lambda (db var op var0 const d body n clauses)
       (and (equal? op eq-inline-operator)
 	   (immediate? const)
-	   (= 1 (length (get db var 'references)))
+	   (= 1 (length (get-list db var 'references)))
 	   (make-node
 	    '##core#switch
 	    (list (add1 n))
@@ -559,6 +591,8 @@
 	      [subs (node-subexpressions body)] )
 	  (and (eq? c 'let)
 	       (null? (cdr params))
+               (not (get db (first params) 'inline-transient))
+               (not (get db (first params) 'references))
 	       (let* ([val (first subs)]
 		      [valparams (node-parameters val)]
 		      [valsubs (node-subexpressions val)] )
@@ -576,6 +610,7 @@
 				   [subs (node-subexpressions body)] )
 			       (cond [(and (eq? c 'let)
 					   (null? (cdr params))
+					   (not (get db (first params) 'inline-transient))
 					   (not (get db (first params) 'references))
 					   (pair? vars)
 					   (eq? 'set! (node-class (first subs)))
@@ -599,7 +634,7 @@
 	(##core#call p (##core#variable (var1)) . more) ) ; `p' was `#t', bombed also
    (var1 var2 p more)
    ,(lambda (db var1 var2 p more)
-      (and (= 1 (length (get db var1 'references)))
+      (and (= 1 (length (get-list db var1 'references)))
 	   (make-node
 	    '##core#call p
 	    (cons (varnode var2) more) ) ) ) )
@@ -617,7 +652,7 @@
    (var op args d x y)
    ,(lambda (db var op args d x y)
       (and (not (equal? op eq-inline-operator))
-	   (= 1 (length (get db var 'references)))
+	   (= 1 (length (get-list db var 'references)))
 	   (make-node
 	    'if d
 	    (list (make-node '##core#inline (list op) args)
@@ -760,7 +795,7 @@
 		  body
 		  sgraph) ] )
 	    (cond [(pair? optimized)
-		   (debugging 'o "eliminated assignments" optimized)
+		   (debugging 'o "converted assignments to bindings" optimized)
 		   (values n2 #t) ]
 		  [else (values n2 #f)] ) ) ) ) ) ) )
 
@@ -799,23 +834,17 @@
 		    (list cont (make-node '##core#inline (list (second classargs)) callargs)) ) ) ) ) )
 
     ;; (<op> ...) -> (##core#inline <iop> ...)
-    ;; (<op> <rest-vector>) -> (##core#inline <iopv> <rest-vector>)
-    ((2) ; classargs = (<argc> <iop> <safe> <iopv>)
+    ((2) ; classargs = (<argc> <iop> <safe>)
      (and inline-substitutions-enabled
 	  (= (length callargs) (first classargs))
 	  (intrinsic? name)
 	  (or (third classargs) unsafe)
-	  (let ([arg1 (first callargs)]
-		[iopv (fourth classargs)] )
+	  (let ((arg1 (first callargs)))
 	    (make-node
 	     '##core#call '(#t)
 	     (list 
 	      cont
-	      (cond [(and iopv
-			  (eq? '##core#variable (node-class arg1))
-			  (eq? 'vector (get db (first (node-parameters arg1)) 'rest-parameter)) )
-		     (make-node '##core#inline (list iopv) callargs) ]
-		    [else (make-node '##core#inline (list (second classargs)) callargs)] ) ) ) ) ) )
+	      (make-node '##core#inline (list (second classargs)) callargs) ) ) ) ) )
 
     ;; (<op>) -> <var>
     ((3) ; classargs = (<var>)
@@ -1194,8 +1223,8 @@
 		      (not (get db d 'unknown))
 		      (proper-list? llist)
 		      (and-let* ([val (get db d 'value)]
-				 [refs (get db d 'references)]
-				 [sites (get db d 'call-sites)] )
+				 [refs (get-list db d 'references)]
+				 [sites (get-list db d 'call-sites)] )
 			(and (eq? n val)
 			     (= (length refs) (length sites))
 			     (scan (first subs) (first llist) d dn (cons d llist)) ) ) )
@@ -1295,7 +1324,7 @@
       (let* ([params (node-parameters n)]
 	     [argc (length (third params))]
 	     [klambdas '()] 
-	     [sites (get db fnvar 'call-sites)] 
+	     [sites (or (get db fnvar 'call-sites) '())]
 	     [ksites '()] )
 	(if (and (list? params) (= (length params) 4) (list? (caddr params)))
 	    (let ((id (car params))
@@ -1419,20 +1448,20 @@
     (define (find-lifting-candidates)
       ;; Collect potentially liftable procedures and return as a list of (<name> . <value>) pairs:
       ;; - Also build a-list that maps lambda-nodes to names.
-      (let ([cs '()])
+      (let ((cs '()))
 	(##sys#hash-table-for-each
 	 (lambda (sym plist)
-	   (and-let* ([val (assq 'value plist)]
-		      [refs (assq 'references plist)]
-		      [css (assq 'call-sites plist)] 
-		      [nrefs (length (cdr refs))] )
-	     (when (and (not (assq 'unknown plist))
-			(eq? 'lambda (node-class (cdr val)))
-			(not (assq 'global plist)) 
-			#;(> nrefs 1)
-			(= nrefs (length (cdr css))) )
-	       (set! lambda-values (alist-cons (cdr val) sym lambda-values))
-	       (set! cs (alist-cons sym (cdr val) cs)) ) ) )
+	   (and-let* ((val (assq 'value plist))
+		      (refs (assq 'references plist))
+		      (css (assq 'call-sites plist)) )
+	     (let ((nrefs (length (cdr refs))))
+	       (when (and (not (assq 'unknown plist))
+			  (eq? 'lambda (node-class (cdr val)))
+			  (not (assq 'global plist)) 
+			  #;(> nrefs 1)
+			  (= nrefs (length (cdr css))) )
+		 (set! lambda-values (alist-cons (cdr val) sym lambda-values))
+		 (set! cs (alist-cons sym (cdr val) cs)) ) ) ) )
 	 db)
 	cs) )
 
@@ -1755,7 +1784,8 @@
 	    (when (debugging 'l "accessibles:") (pretty-print al))
 	    (debugging 'p "eliminating liftables by access-lists and non-liftable callees...")
 	    (let ([ls (eliminate3 (eliminate4 g2))]) ;(eliminate2 g2 al)))]) - why isn't this used?
-	      (debugging 'o "liftable local procedures" (delay (unzip1 ls)))
+	      (when (pair? ls)
+		(debugging 'o "liftable local procedures" (delay (unzip1 ls))))
 	      (debugging 'p "gathering extra parameters...")
 	      (let ([extra (compute-extra-variables ls)])
 		(when (debugging 'l "additional parameters:") (pretty-print extra))
@@ -1765,183 +1795,3 @@
 		(remove-local-bindings! ls)
 		(debugging 'p "moving liftables to toplevel...")
 		(reconstruct! ls extra) ) ) ) ) ) ) ) )
-
-
-;;; Compiler macros (that operate in the expansion phase)
-
-(define compiler-syntax-statistics '())
-
-(set! ##sys#compiler-syntax-hook
-  (lambda (name result)
-    (let ((a (alist-ref name compiler-syntax-statistics eq? 0)))
-      (set! compiler-syntax-statistics
-	(alist-update! name (add1 a) compiler-syntax-statistics)))))
-
-(define (r-c-s names transformer #!optional (se '()))
-  (let ((t (cons (##sys#er-transformer transformer) se)))
-    (for-each
-     (lambda (name)
-       (##sys#put! name '##compiler#compiler-syntax t) )
-     (if (symbol? names) (list names) names) ) ) )
-
-(r-c-s 
- '(for-each ##sys#for-each #%for-each)
- (lambda (x r c)
-   (let ((%let (r 'let))
-	 (%if (r 'if))
-	 (%loop (r 'loop))
-	 (%lst (r 'lst))
-	 (%begin (r 'begin))
-	 (%pair? (r 'pair?)))
-     (if (and (memq 'for-each standard-bindings) ; we have to check this because the db (and thus 
-	      (= 3 (length x)))			 ; intrinsic marks) isn't set up yet
-	 `(,%let ,%loop ((,%lst ,(caddr x)))
-		 (,%if (,%pair? ,%lst)
-		       (,%begin
-			(,(cadr x) (##sys#slot ,%lst 0))
-			(##core#app ,%loop (##sys#slot ,%lst 1))) ) )
-	 x)))
- `((pair? . ,(##sys#primitive-alias 'pair?))))
-
-(r-c-s
- '(o #%o)
- (lambda (x r c)
-   (if (and (fx> (length x) 1)
-	    (memq 'o extended-bindings) )
-       (let ((%tmp (r 'tmp)))
-	 `(,(r 'lambda) (,%tmp) ,(fold-right list %tmp (cdr x))))
-       x)))
-
-(let ((env `((display . ,(##sys#primitive-alias 'display)) ;XXX clean this up
-	     (write . ,(##sys#primitive-alias 'write))
-	     (fprintf . ,(##sys#primitive-alias 'fprintf))
-	     (number->string . ,(##sys#primitive-alias 'number->string))
-	     (write-char . ,(##sys#primitive-alias 'write-char))
-	     (open-output-string . ,(##sys#primitive-alias 'open-output-string))
-	     (get-output-string . ,(##sys#primitive-alias 'get-output-string)) ) ) )
-  (r-c-s
-   '(sprintf #%sprintf format #%format)
-   (lambda (x r c)
-     (let* ((out (gensym 'out))
-	    (code (compile-format-string 
-		   (if (memq (car x) '(sprintf #%sprintf))
-		       'sprintf
-		       'format)
-		   out 
-		   x
-		   (cdr x)
-		   r c)))
-       (if code
-	   `(,(r 'let) ((,out (,(r 'open-output-string))))
-	     ,code
-	     (,(r 'get-output-string) ,out))
-	   x)))
-   env)
-  (r-c-s
-   '(fprintf #%fprintf)
-   (lambda (x r c)
-     (if (>= (length x) 3)
-	 (let ((code (compile-format-string 
-		      'fprintf (cadr x) 
-		      x (cddr x)
-		      r c)))
-	   (if code
-	       code
-	       x))
-	 x))
-   env)
-  (r-c-s
-   '(printf #%printf)
-   (lambda (x r c)
-     (let ((code (compile-format-string 
-		  'printf '##sys#standard-output
-		  x (cdr x)
-		  r c)))
-       (if code
-	   code
-	   x)))
-   env))
-
-(define (compile-format-string func out x args r c)
-  (call/cc
-   (lambda (return)
-     (and (>= (length args) 1)
-	  (memq func extended-bindings)	; s.a.
-	  (or (string? (car args))
-	      (and (list? (car args))
-		   (c (r 'quote) (caar args))
-		   (string? (cadar args))))
-	  (let ((fstr (if (string? (car args)) (car args) (cadar args)))
-		(args (cdr args)))
-	    (define (fail ret? msg . args)
-	      (let ((ln (get-line x)))
-		(compiler-warning 
-		 'syntax
-		 "(~a) in format string ~s~a, ~?" 
-		 func fstr 
-		 (if ln (sprintf " in line ~a" ln) "")
-		 msg args) ) 
-	      (when ret? (return #f)))
-	    (let ((code '())
-		  (index 0)
-		  (len (string-length fstr)) 
-		  (%display (r 'display))
-		  (%write (r 'write))
-		  (%write-char (r 'write-char))
-		  (%out (r 'out))
-		  (%fprintf (r 'fprintf))
-		  (%let (r 'let))
-		  (%number->string (r 'number->string)))
-	      (define (fetch)
-		(let ((c (string-ref fstr index)))
-		  (set! index (fx+ index 1))
-		  c) )
-	      (define (next)
-		(if (null? args)
-		    (fail #t "too few arguments to formatted output procedure")
-		    (let ((x (car args)))
-		      (set! args (cdr args))
-		      x) ) )
-	      (define (endchunk chunk)
-		(when (pair? chunk)
-		  (push 
-		   (if (= 1 (length chunk))
-		       `(,%write-char ,(car chunk) ,%out)
-		       `(,%display ,(reverse-list->string chunk) ,%out)))))
-	      (define (push exp)
-		(set! code (cons exp code)))
-	      (let loop ((chunk '()))
-		(cond ((>= index len)
-		       (unless (null? args)
-			 (fail #f "too many arguments to formatted output procedure"))
-		       (endchunk chunk)
-		       `(,%let ((,%out ,out))
-			       ,@(reverse code)))
-		      (else
-		       (let ((c (fetch)))
-			 (if (eq? c #\~)
-			     (let ((dchar (fetch)))
-			       (endchunk chunk)
-			       (case (char-upcase dchar)
-				 ((#\S) (push `(,%write ,(next) ,%out)))
-				 ((#\A) (push `(,%display ,(next) ,%out)))
-				 ((#\C) (push `(,%write-char ,(next) ,%out)))
-				 ((#\B) (push `(,%display (,%number->string ,(next) 2) ,%out)))
-				 ((#\O) (push `(,%display (,%number->string ,(next) 8) ,%out)))
-				 ((#\X) (push `(,%display (,%number->string ,(next) 16) ,%out)))
-				 ((#\!) (push `(##sys#flush-output ,%out)))
-				 ((#\?)
-				  (let* ([fstr (next)]
-					 [lst (next)] )
-				    (push `(##sys#apply ,%fprintf ,%out ,fstr ,lst))))
-				 ((#\~) (push `(,write-char #\~ ,%out)))
-				 ((#\% #\N) (push `(,%write-char #\newline ,%out)))
-				 (else
-				  (if (char-whitespace? dchar)
-				      (let skip ((c (fetch)))
-					(if (char-whitespace? c)
-					    (skip (fetch))
-					    (set! index (sub1 index))))
-				      (fail #t "illegal format-string character `~c'" dchar) ) ) )
-			       (loop '()) )
-			     (loop (cons c chunk)))))))))))))
