@@ -1,7 +1,7 @@
 ;;; ports.scm - Optional non-standard ports
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
@@ -51,13 +51,12 @@
       ##sys#substring ##sys#for-each ##sys#map ##sys#setslot
       ##sys#allocate-vector ##sys#check-pair ##sys#error-not-a-proper-list
       ##sys#member ##sys#assoc ##sys#error ##sys#signal-hook ##sys#read-string!
-      ##sys#check-symbol ##sys#check-vector ##sys#floor ##sys#ceiling
-      ##sys#truncate ##sys#round ##sys#check-number ##sys#cons-flonum
+      ##sys#check-symbol ##sys#check-vector ##sys#check-number
       ##sys#flonum-fraction ##sys#make-port ##sys#fetch-and-check-port-arg
       ##sys#print ##sys#check-structure ##sys#make-structure make-parameter
       ##sys#flush-output ##sys#write-char-0 ##sys#number->string
       ##sys#fragments->string ##sys#symbol->qualified-string
-      ##extras#reverse-string-append ##sys#number? ##sys#procedure->string
+      ##sys#number? ##sys#procedure->string
       ##sys#pointer->string ##sys#user-print-hook ##sys#peek-char-0
       ##sys#read-char-0 ##sys#write-char ##sys#string-append ##sys#gcd ##sys#lcm
       ##sys#fudge ##sys#check-list ##sys#user-read-hook ##sys#check-closure ##sys#check-inexact
@@ -226,7 +225,7 @@
 	     read-line) )		; read-line
 	   (data (vector #f))
 	   (port (##sys#make-port #t class "(custom)" 'custom)) )
-      (##sys#setslot port 9 data) 
+      (##sys#set-port-data! port data) 
       port) ) )
 
 (define make-output-port
@@ -250,5 +249,5 @@
 	       #f) )			; read-line
 	     (data (vector #f))
 	     (port (##sys#make-port #f class "(custom)" 'custom)) )
-	(##sys#setslot port 9 data) 
+	(##sys#set-port-data! port data) 
 	port) ) ) )
