@@ -1,7 +1,7 @@
 ;;;; posixunix.scm - Miscellaneous file- and process-handling routines
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -31,8 +31,9 @@
   (disable-interrupts)
   (usual-integrations)
   (hide ##sys#stat group-member _get-groups _ensure-groups posix-error
-	      ##sys#terminal-check
-	      check-time-vector)
+        ##sys#terminal-check
+        check-time-vector)
+  (not inline ##sys#interrupt-hook ##sys#user-interrupt-hook)
   (foreign-declare #<<EOF
 #include <signal.h>
 #include <errno.h>
@@ -53,6 +54,7 @@ static C_TLS int C_wait_status;
 #include <fcntl.h>
 #include <dirent.h>
 #include <pwd.h>
+#include <utime.h>
 
 #if defined(__sun__) && defined(__svr4__)
 # include <sys/tty.h>
@@ -110,6 +112,10 @@ extern char **environ;
 # define ENV_MAX        1024
 #endif
 
+#ifndef FILENAME_MAX
+# define FILENAME_MAX          1024
+#endif
+
 static C_TLS char *C_exec_args[ ARG_MAX ];
 static C_TLS char *C_exec_env[ ENV_MAX ];
 static C_TLS struct utsname C_utsname;
@@ -137,10 +143,10 @@ static C_TLS struct stat C_statbuf;
 #define C_chdir(str)        C_fix(chdir(C_c_string(str)))
 #define C_rmdir(str)        C_fix(rmdir(C_c_string(str)))
 
-#define C_opendir(x,h)          C_set_block_item(h, 0, (C_word) opendir(C_c_string(x)))
-#define C_closedir(h)           (closedir((DIR *)C_block_item(h, 0)), C_SCHEME_UNDEFINED)
-#define C_readdir(h,e)          C_set_block_item(e, 0, (C_word) readdir((DIR *)C_block_item(h, 0)))
-#define C_foundfile(e,b)        (strcpy(C_c_string(b), ((struct dirent *) C_block_item(e, 0))->d_name), C_fix(strlen(((struct dirent *) C_block_item(e, 0))->d_name)))
+#define C_opendir(x,h)      C_set_block_item(h, 0, (C_word) opendir(C_c_string(x)))
+#define C_closedir(h)       (closedir((DIR *)C_block_item(h, 0)), C_SCHEME_UNDEFINED)
+#define C_readdir(h,e)      C_set_block_item(e, 0, (C_word) readdir((DIR *)C_block_item(h, 0)))
+#define C_foundfile(e,b)    (strcpy(C_c_string(b), ((struct dirent *) C_block_item(e, 0))->d_name), C_fix(strlen(((struct dirent *) C_block_item(e, 0))->d_name)))
 
 #define C_curdir(buf)       (getcwd(C_c_string(buf), 256) ? C_fix(strlen(C_c_string(buf))) : C_SCHEME_FALSE)
 
@@ -154,7 +160,6 @@ static C_TLS struct stat C_statbuf;
 
 #define C_fork              fork
 #define C_waitpid(id, o)    C_fix(waitpid(C_unfix(id), &C_wait_status, C_unfix(o)))
-#define C_getpid            getpid
 #define C_getppid           getppid
 #define C_kill(id, s)       C_fix(kill(C_unfix(id), C_unfix(s)))
 #define C_getuid            getuid
@@ -171,7 +176,7 @@ static C_TLS struct stat C_statbuf;
 #define C_setpgid(x, y)     C_fix(setpgid(C_unfix(x), C_unfix(y)))
 #define C_getpgid(x)        C_fix(getpgid(C_unfix(x)))
 #define C_symlink(o, n)     C_fix(symlink(C_data_pointer(o), C_data_pointer(n)))
-#define C_readlink(f, b)    C_fix(readlink(C_data_pointer(f), C_data_pointer(b), FILENAME_MAX))
+#define C_do_readlink(f, b)    C_fix(readlink(C_data_pointer(f), C_data_pointer(b), FILENAME_MAX))
 #define C_getpwnam(n)       C_mk_bool((C_user = getpwnam((char *)C_data_pointer(n))) != NULL)
 #define C_getpwuid(u)       C_mk_bool((C_user = getpwuid(C_unfix(u))) != NULL)
 #ifdef HAVE_GRP_H
@@ -191,7 +196,7 @@ static C_TLS struct stat C_statbuf;
 #define C_dup2(x, y)        C_fix(dup2(C_unfix(x), C_unfix(y)))
 #define C_alarm             alarm
 #define C_setvbuf(p, m, s)  C_fix(setvbuf(C_port_file(p), NULL, C_unfix(m), C_unfix(s)))
-#define C_access(fn, m)     C_fix(access((char *)C_data_pointer(fn), C_unfix(m)))
+#define C_test_access(fn, m)     C_fix(access((char *)C_data_pointer(fn), C_unfix(m)))
 #define C_close(fd)         C_fix(close(C_unfix(fd)))
 #define C_sleep             sleep
 
@@ -295,10 +300,6 @@ static C_TLS int C_uw;
 #define C_flock_lock(p)     C_fix(fcntl(fileno(C_port_file(p)), F_SETLK, &C_flock))
 #define C_flock_lockw(p)    C_fix(fcntl(fileno(C_port_file(p)), F_SETLKW, &C_flock))
 
-#ifndef FILENAME_MAX
-# define FILENAME_MAX          1024
-#endif
-
 static C_TLS sigset_t C_sigset;
 #define C_sigemptyset(d)    (sigemptyset(&C_sigset), C_SCHEME_UNDEFINED)
 #define C_sigaddset(s)      (sigaddset(&C_sigset, C_unfix(s)), C_SCHEME_UNDEFINED)
@@ -313,8 +314,9 @@ static C_TLS sigset_t C_sigset;
 #define C_write(fd, b, n)   C_fix(write(C_unfix(fd), C_data_pointer(b), C_unfix(n)))
 #define C_mkstemp(t)        C_fix(mkstemp(C_c_string(t)))
 
-#define C_ftell(p)            C_fix(ftell(C_port_file(p)))
-#define C_fseek(p, n, w)      C_mk_nbool(fseek(C_port_file(p), C_unfix(n), C_unfix(w)))
+/* It is assumed that 'int' is-a 'long' */
+#define C_ftell(p)          C_fix(ftell(C_port_file(p)))
+#define C_fseek(p, n, w)    C_mk_nbool(fseek(C_port_file(p), C_num_to_int(n), C_unfix(w)))
 #define C_lseek(fd, o, w)     C_fix(lseek(C_unfix(fd), C_unfix(o), C_unfix(w)))
 
 #define C_zero_fd_set(i)      FD_ZERO(&C_fd_sets[ i ])
@@ -430,13 +432,18 @@ C_tm_get( C_word v )
 #endif
 
 #define C_asctime(v)    (asctime(C_tm_set(v)))
-#define C_mktime(v)     ((C_temporary_flonum = mktime(C_tm_set(v))) != -1)
-#define C_timegm(v)     ((C_temporary_flonum = timegm(C_tm_set(v))) != -1)
+#define C_a_mktime(ptr, c, v)  C_flonum(ptr, mktime(C_tm_set(v)))
+#define C_a_timegm(ptr, c, v)  C_flonum(ptr, timegm(C_tm_set(v)))
 
 #define TIME_STRING_MAXLENGTH 255
 static char C_time_string [TIME_STRING_MAXLENGTH + 1];
 #undef TIME_STRING_MAXLENGTH
 
+#ifdef __linux__
+extern char *strptime(const char *s, const char *format, struct tm *tm);
+extern pid_t getpgid(pid_t pid);
+#endif
+
 #define C_strftime(v, f) \
         (strftime(C_time_string, sizeof(C_time_string), C_c_string(f), C_tm_set(v)) ? C_time_string : NULL)
 
@@ -472,6 +479,14 @@ static int get_tty_size(int p, int *rows, int *cols)
 }
 #endif
 
+static int set_file_mtime(char *filename, C_word tm)
+{
+  struct utimbuf tb;
+
+  tb.actime = tb.modtime = C_num_to_int(tm);
+  return utime(filename, &tb);
+}
+
 EOF
 ) )
 
@@ -488,8 +503,8 @@ EOF
      ##sys#error ##sys#signal-hook ##sys#peek-unsigned-integer make-pathname glob directory?
      pathname-file process-fork file-close duplicate-fileno process-execute get-environment-variable
      make-string make-input-port make-output-port ##sys#thread-block-for-i/o create-pipe
-     process-wait pathname-strip-directory ##sys#expand-home-path directory
-     decompose-pathname ##sys#cons-flonum ##sys#decode-seconds ##sys#null-pointer ##sys#pointer->address
+     process-wait pathname-strip-directory pathname-directory ##sys#expand-home-path directory
+     decompose-pathname ##sys#decode-seconds ##sys#null-pointer ##sys#pointer->address
      ##sys#substring ##sys#context-switch close-input-pipe close-output-pipe change-directory
      current-directory ##sys#make-pointer port? ##sys#schedule ##sys#process
      ##sys#peek-fixnum ##sys#make-structure ##sys#check-structure ##sys#enable-interrupts
@@ -773,7 +788,20 @@ EOF
           _stat_st_blksize _stat_st_blocks) )
 
 (define (file-size f) (##sys#stat f #f 'file-size) _stat_st_size)
-(define (file-modification-time f) (##sys#stat f #f 'file-modification-time) _stat_st_mtime)
+
+(define file-modification-time
+  (getter-with-setter 
+   (lambda (f)
+     (##sys#stat f #f 'file-modification-time) _stat_st_mtime)
+   (lambda (f t)
+     (##sys#check-number t 'set-file-modification-time)
+     (let ((r ((foreign-lambda int "set_file_mtime" c-string scheme-object)
+	       (##sys#expand-home-path file) t)))
+       (when (fx< r 0)
+	 (posix-error 
+	  #:file-error 'set-file-modification-time
+	  "cannot set file modification-time" f t))))))
+
 (define (file-access-time f) (##sys#stat f #f 'file-access-time) _stat_st_atime)
 (define (file-change-time f) (##sys#stat f #f 'file-change-time) _stat_st_ctime)
 (define (file-owner f) (##sys#stat f #f 'file-owner) _stat_st_uid)
@@ -789,68 +817,54 @@ EOF
   (##sys#stat fname #t 'symbolic-link?)
   (foreign-value "C_islink" bool) )
 
-(define (stat-regular? fname)		; DEPRECATED
-    (##sys#check-string fname 'stat-regular?)
-    (##sys#stat fname #f 'stat-regular?)
-    (foreign-value "C_isreg" bool))
-
-(define (stat-directory? fname)		; DEPRECATED
-    (##sys#check-string fname 'stat-directory?)
-    (##sys#stat fname #f 'stat-directory?)
-    (foreign-value "C_isdir" bool))
-
 (define (character-device? fname)
     (##sys#check-string fname 'character-device?)
     (##sys#stat fname #f 'character-device?)
     (foreign-value "C_ischr" bool))
 
-(define stat-char-device? character-device?) ; DEPRECATED
-
 (define (block-device? fname)
     (##sys#check-string fname 'block-device?)
     (##sys#stat fname #f 'block-device?)
     (foreign-value "C_isblk" bool))
 
-(define stat-block-device? block-device?) ; DEPRECATED
-
 (define (fifo? fname)
     (##sys#check-string fname 'stat-fifo?)
     (##sys#stat fname #f 'stat-fifo?)
     (foreign-value "C_isfifo" bool))
 
-(define stat-fifo? fifo?)		; DEPRECATED
-(define stat-symlink? symbolic-link?)	; DEPRECATED
-
 (define (socket? fname)
   (##sys#check-string fname 'socket?)
   (##sys#stat fname #f 'socket?)
   (foreign-value "C_issock" bool))
 
-(define stat-socket? socket?)		; DEPRECATED
-
 (define set-file-position!
    (lambda (port pos . whence)
-     (let ([whence (if (pair? whence) (car whence) _seek_set)])
+     (let ((whence (if (pair? whence) (car whence) _seek_set)))
        (##sys#check-exact pos 'set-file-position!)
        (##sys#check-exact whence 'set-file-position!)
-       (when (fx< pos 0) (##sys#signal-hook #:bounds-error 'set-file-position! "invalid negative port position" pos port))
-       (unless (cond [(port? port)
+       (when (negative? pos)
+         (##sys#signal-hook #:bounds-error 'set-file-position! "invalid negative port position" pos port))
+       (unless (cond ((port? port)
 		      (and (eq? (##sys#slot port 7) 'stream)
-			   (##core#inline "C_fseek" port pos whence) ) ]
-		     [(fixnum? port) (##core#inline "C_lseek" port pos whence)]
-		     [else (##sys#signal-hook #:type-error 'set-file-position! "invalid file" port)] )
+			   (##core#inline "C_fseek" port pos whence) ) )
+		     ((fixnum? port)
+		      (##core#inline "C_lseek" port pos whence))
+		     (else
+		      (##sys#signal-hook #:type-error 'set-file-position! "invalid file" port)) )
 	 (posix-error #:file-error 'set-file-position! "cannot set file position" port pos) ) ) ) )
 
 (define file-position
   (getter-with-setter
    (lambda (port)
-     (let ([pos (cond [(port? port)
+     (let ((pos (cond ((port? port)
 		       (if (eq? (##sys#slot port 7) 'stream)
 			   (##core#inline "C_ftell" port)
-			   -1) ]
-		      [(fixnum? port) (##core#inline "C_lseek" port 0 _seek_cur)]
-		      [else (##sys#signal-hook #:type-error 'file-position "invalid file" port)] ) ] )
-       (when (fx< pos 0)
+			   -1) )
+		      ((fixnum? port)
+		       (##core#inline "C_lseek" port 0 _seek_cur) )
+		      (else
+		       (##sys#signal-hook #:type-error 'file-position "invalid file" port)) ) ) )
+       (when (< pos 0)
 	 (posix-error #:file-error 'file-position "cannot retrieve file position of port" port) )
        pos) )
    set-file-position!) )		; doesn't accept WHENCE
@@ -858,51 +872,45 @@ EOF
 
 ;;; Directory stuff:
 
-(define-inline (create-directory-helper name)
-    (unless (zero? (##core#inline "C_mkdir" (##sys#make-c-string name)))
-            (posix-error #:file-error 'create-directory
-                         "cannot create directory" name)))
+(define-inline (*directory? loc name)
+  (and (fx= 0 (##core#inline "C_stat" (##sys#make-c-string name)))
+       (foreign-value "C_isdir" bool) ) )
 
-(define-inline (create-directory-check name)
-    (if (file-exists? name)
-        (if (fx< (##core#inline "C_stat" (##sys#make-c-string name)) 0)
-            (posix-error #:file-error 'create-directory
-                         "cannot stat file" name)
-            (or (foreign-value "C_isdir" bool)
-                (posix-error #:file-error 'create-directory
-                             "path segment is a file" name)))
-        #f))
-
-(define-inline (create-directory-helper-silent name)
-    (unless (create-directory-check name)
-            (create-directory-helper name)))
-
-(define-inline (create-directory-helper-parents name)
-    (let ((c   ""))
-        (for-each
-             (lambda (x)
-                 (set! c (string-append c "/" x))
-                 (create-directory-helper-silent c))
-             (string-split name "/"))))
+(define-inline (*create-directory loc name)
+  (unless (fx= 0 (##core#inline "C_mkdir" (##sys#make-c-string name)))
+    (posix-error #:file-error loc "cannot create directory" name)) )
 
 (define create-directory
-  (lambda (name #!optional parents?)
-    (##sys#check-string name 'create-directory)
-    (if parents?
-        (create-directory-helper-parents name)
-        (create-directory-helper name))))
+  (let ((decompose-pathname decompose-pathname)
+        (pathname-directory pathname-directory) )
+    (lambda (name #!optional parents?)
+      (##sys#check-string name 'create-directory)
+      (let ((name (##sys#expand-home-path name)))
+        (unless (or (fx= 0 (##sys#size name)) (*directory? 'create-directory name))
+          (if parents?
+              (let loop ((dir (let-values (((dir file ext) (decompose-pathname name)))
+                                (if file (make-pathname dir file ext) dir))))
+                (when (and dir (not (*directory? 'create-directory dir)))
+                  (loop (pathname-directory dir))
+                  (*create-directory 'create-directory dir)) )
+              (*create-directory 'create-directory name) ) )
+	name))))
 
 (define change-directory
   (lambda (name)
     (##sys#check-string name 'change-directory)
-    (unless (zero? (##core#inline "C_chdir" (##sys#make-c-string (##sys#expand-home-path name))))
-      (posix-error #:file-error 'change-directory "cannot change current directory" name) ) ) )
+    (let ((sname (##sys#make-c-string (##sys#expand-home-path name))))
+      (unless (fx= 0 (##core#inline "C_chdir" sname))
+	(posix-error #:file-error 'change-directory "cannot change current directory" name) )
+      name)))
 
 (define delete-directory
   (lambda (name)
     (##sys#check-string name 'delete-directory)
-    (unless (zero? (##core#inline "C_rmdir" (##sys#make-c-string (##sys#expand-home-path name))))
-      (posix-error #:file-error 'delete-directory "cannot delete directory" name) ) ) )
+    (let ((sname (##sys#make-c-string (##sys#expand-home-path name))))
+      (unless (fx= 0 (##core#inline "C_rmdir" sname))
+	(posix-error #:file-error 'delete-directory "cannot delete directory" name) )
+      name)))
 
 (define directory
   (let ([string-ref string-ref]
@@ -935,8 +943,7 @@ EOF
 
 (define (directory? fname)
   (##sys#check-string fname 'directory?)
-  (let ((info (##sys#file-info (##sys#expand-home-path fname))))
-    (and info (fx= 1 (##sys#slot info 4))) ) )
+  (*directory? 'directory? (##sys#expand-home-path fname)) )
 
 (define current-directory
   (let ([make-string make-string])
@@ -1312,10 +1319,10 @@ EOF
           _user-shell) ) ) )
 
 (define (current-user-name)
-  (list-ref (user-information (current-user-id)) 0) )
+  (car (user-information (current-user-id))) )
 
 (define (current-effective-user-name)
-  (list-ref (user-information (current-effective-user-id)) 0) )
+  (car (user-information (current-effective-user-id))) )
 
 (define-foreign-variable _group-name nonnull-c-string "C_group->gr_name")
 (define-foreign-variable _group-passwd nonnull-c-string "C_group->gr_passwd")
@@ -1484,7 +1491,7 @@ EOF
 (let ()
   (define (check filename acc loc)
     (##sys#check-string filename loc)
-    (let ([r (fx= 0 (##core#inline "C_access" (##sys#make-c-string (##sys#expand-home-path filename)) acc))])
+    (let ([r (fx= 0 (##core#inline "C_test_access" (##sys#make-c-string (##sys#expand-home-path filename)) acc))])
       (unless r (##sys#update-errno))
       r) )
   (set! file-read-access? (lambda (filename) (check filename _r_ok 'file-read-access?)))
@@ -1535,7 +1542,7 @@ EOF
         [buf (make-string (fx+ _filename_max 1))] )
     (lambda (fname #!optional canonicalize)
       (##sys#check-string fname 'read-symbolic-link)
-      (let ([len (##core#inline "C_readlink" (##sys#make-c-string (##sys#expand-home-path fname)) buf)])
+      (let ([len (##core#inline "C_do_readlink" (##sys#make-c-string (##sys#expand-home-path fname)) buf)])
       (when (fx< len 0)
         (posix-error #:file-error 'read-symbolic-link "cannot read symbolic link" fname) )
       (let ((pathname (substring buf 0 len)))
@@ -1930,10 +1937,10 @@ EOF
       (let ([addr (if (not addr) (##sys#null-pointer) addr)]
             [off (if (pair? off) (car off) 0)] )
         (unless (and (##core#inline "C_blockp" addr) (##core#inline "C_specialp" addr))
-                (##sys#signal-hook #:type-error 'map-file-to-memory "bad argument type - not a foreign pointer" addr) )
+	  (##sys#signal-hook #:type-error 'map-file-to-memory "bad argument type - not a foreign pointer" addr) )
         (let ([addr2 (mmap addr len prot flag fd off)])
           (when (eq? -1 (##sys#pointer->address addr2))
-                (posix-error #:file-error 'map-file-to-memory "cannot map file to memory" addr len prot flag fd off) )
+	    (posix-error #:file-error 'map-file-to-memory "cannot map file to memory" addr len prot flag fd off) )
           (##sys#make-structure 'mmap addr2 len) ) ) ) ) )
 
 (define unmap-file-from-memory
@@ -1942,7 +1949,7 @@ EOF
       (##sys#check-structure mmap 'mmap 'unmap-file-from-memory)
       (let ([len (if (pair? len) (car len) (##sys#slot mmap 2))])
         (unless (eq? 0 (munmap (##sys#slot mmap 1) len))
-                (posix-error #:file-error 'unmap-file-from-memory "cannot unmap file from memory" mmap len) ) ) ) ) )
+	  (posix-error #:file-error 'unmap-file-from-memory "cannot unmap file from memory" mmap len) ) ) ) ) )
 
 (define (memory-mapped-file-pointer mmap)
   (##sys#check-structure mmap 'mmap 'memory-mapped-file-pointer)
@@ -1958,17 +1965,17 @@ EOF
   (when (fx< (##sys#size tm) 10)
     (##sys#error loc "time vector too short" tm) ) )
 
-(define (seconds->local-time secs)
+(define (seconds->local-time #!optional (secs (current-seconds)))
   (##sys#check-number secs 'seconds->local-time)
   (##sys#decode-seconds secs #f) )
 
-(define (seconds->utc-time secs)
+(define (seconds->utc-time #!optional (secs (current-seconds)))
   (##sys#check-number secs 'seconds->utc-time)
   (##sys#decode-seconds secs #t) )
 
 (define seconds->string
   (let ([ctime (foreign-lambda c-string "C_ctime" integer)])
-    (lambda (secs)
+    (lambda (#!optional (secs (current-seconds)))
       (##sys#check-number secs 'seconds->string)
       (let ([str (ctime secs)])
         (if str
@@ -1999,20 +2006,22 @@ EOF
 
 (define (local-time->seconds tm)
   (check-time-vector 'local-time->seconds tm)
-  (if (##core#inline "C_mktime" tm)
-      (##sys#cons-flonum)
-      (##sys#error 'local-time->seconds "cannot convert time vector to seconds" tm) ) )
+  (let ((t (##core#inline_allocate ("C_a_mktime" 4) tm)))
+    (if (fp= -1.0 t)
+	(##sys#error 'local-time->seconds "cannot convert time vector to seconds" tm)
+	t)))
 
 (define (utc-time->seconds tm)
   (check-time-vector 'utc-time->seconds tm)
-  (if (##core#inline "C_timegm" tm)
-      (##sys#cons-flonum)
-      (##sys#error 'utc-time->seconds "cannot convert time vector to seconds" tm) ) )
+  (let ((t (##core#inline_allocate ("C_a_timegm" 4) tm)))
+    (if (fp= -1.0 t)
+	(##sys#error 'utc-time->seconds "cannot convert time vector to seconds" tm) 
+	t)))
 
 (define local-timezone-abbreviation
   (foreign-lambda* c-string ()
    "\n#if !defined(__CYGWIN__) && !defined(__SVR4) && !defined(__uClinux__) && !defined(__hpux__)\n"
-   "time_t clock = (time_t)0;"
+   "time_t clock = time(NULL);"
    "struct tm *ltm = C_localtime(&clock);"
    "char *z = ltm ? (char *)ltm->tm_zone : 0;"
    "\n#else\n"
@@ -2075,13 +2084,13 @@ EOF
       (##sys#terminal-check 'terminal-size port)
       (let-location ((columns int)
 		     (rows int))
-		    (if (fx= 0
-			     (ttysize (##core#inline "C_C_fileno" port)
-				      (location columns)
-				      (location rows)))
-			(values columns rows)
-			(posix-error #:error 'terminal-size
-				     "Unable to get size of terminal" port))))))
+	(if (fx= 0
+		 (ttysize (##core#inline "C_C_fileno" port)
+			  (location columns)
+			  (location rows)))
+	    (values columns rows)
+	    (posix-error #:error 'terminal-size
+			 "Unable to get size of terminal" port))))))
   
 (define get-host-name
   (let ([getit
@@ -2354,6 +2363,7 @@ EOF
   (let ([glob glob]
 	[string-match string-match]
 	[make-pathname make-pathname]
+	[pathname-file pathname-file]
 	[directory? directory?] )
     (lambda (dir pred . action-id-limit)
       (let-optionals
