@@ -1,9 +1,7 @@
 ;;;; posixwin.scm - Miscellaneous file- and process-handling routines, available on Windows
 ;
-; By Sergey Khorev
-;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -71,6 +69,7 @@
   (hide ##sys#stat posix-error
 	$quote-args-list $exec-setup $exec-teardown
 	check-time-vector)
+  (not inline ##sys#interrupt-hook ##sys#user-interrupt-hook)
   (foreign-declare #<<EOF
 #ifndef WIN32_LEAN_AND_MEAN
 # define WIN32_LEAN_AND_MEAN
@@ -102,6 +101,7 @@ int C_not_implemented() { return -1; }
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <direct.h>
+#include <utime.h>
 
 #include <time.h>
 
@@ -250,7 +250,7 @@ readdir(DIR * dir)
 #define C_dup(x)	    C_fix(dup(C_unfix(x)))
 #define C_dup2(x, y)	    C_fix(dup2(C_unfix(x), C_unfix(y)))
 #define C_setvbuf(p, m, s)  C_fix(setvbuf(C_port_file(p), NULL, C_unfix(m), C_unfix(s)))
-#define C_access(fn, m)	    C_fix(access((char *)C_data_pointer(fn), C_unfix(m)))
+#define C_test_access(fn, m)	    C_fix(access((char *)C_data_pointer(fn), C_unfix(m)))
 #define C_pipe(d, m)	    C_fix(_pipe(C_pipefds, PIPE_BUF, C_unfix(m)))
 #define C_close(fd)	    C_fix(close(C_unfix(fd)))
 
@@ -317,8 +317,9 @@ C_free_arg_string(char **where) {
 #define C_write(fd, b, n)   C_fix(write(C_unfix(fd), C_data_pointer(b), C_unfix(n)))
 #define C_mkstemp(t)	    C_fix(mktemp(C_c_string(t)))
 
-#define C_ftell(p)	    C_fix(ftell(C_port_file(p)))
-#define C_fseek(p, n, w)    C_mk_nbool(fseek(C_port_file(p), C_unfix(n), C_unfix(w)))
+/* It is assumed that 'int' is-a 'long' */
+#define C_ftell(p)          C_fix(ftell(C_port_file(p)))
+#define C_fseek(p, n, w)    C_mk_nbool(fseek(C_port_file(p), C_num_to_int(n), C_unfix(w)))
 #define C_lseek(fd, o, w)   C_fix(lseek(C_unfix(fd), C_unfix(o), C_unfix(w)))
 
 #define C_flushall()	    C_fix(_flushall())
@@ -340,7 +341,7 @@ C_free_arg_string(char **where) {
 #define C_tm_set(v) (C_tm_set_08(v), &C_tm)
 
 #define C_asctime(v)    (asctime(C_tm_set(v)))
-#define C_mktime(v)     ((C_temporary_flonum = mktime(C_tm_set(v))) != -1)
+#define C_a_mktime(ptr, c, v)  C_flonum(ptr, mktime(C_tm_set(v)))
 
 #define TIME_STRING_MAXLENGTH 255
 static char C_time_string [TIME_STRING_MAXLENGTH + 1];
@@ -911,6 +912,14 @@ C_process(const char * app, const char * cmdlin, const char ** env,
 
     return success;
 }
+
+static int set_file_mtime(char *filename, C_word tm)
+{
+  struct _utimbuf tb;
+
+  tb.actime = tb.modtime = C_num_to_int(tm);
+  return _utime(filename, &tb);
+}
 EOF
 ) )
 
@@ -1091,7 +1100,22 @@ EOF
 	  0 0 0 0) )
 
 (define (file-size f) (##sys#stat f) _stat_st_size)
-(define (file-modification-time f) (##sys#stat f) _stat_st_mtime)
+
+(define file-modification-time
+  (getter-with-setter 
+   (lambda (f)
+     (##sys#stat f) _stat_st_mtime)
+   (lambda (f t)
+     (##sys#check-string f 'set-file-modification-time)
+     (##sys#check-number t 'set-file-modification-time)
+     (let ((r ((foreign-lambda int "set_file_mtime" c-string scheme-object)
+	       (##sys#expand-home-path f) 
+	       t)))
+       (when (fx< r 0)
+	 (posix-error 
+	  #:file-error 'set-file-modification-time
+	  "cannot set file modification-time" f t))))))
+
 (define (file-access-time f) (##sys#stat f) _stat_st_atime)
 (define (file-change-time f) (##sys#stat f) _stat_st_ctime)
 (define (file-owner f) (##sys#stat f) _stat_st_uid)
@@ -1111,52 +1135,48 @@ EOF
              (lambda (fname)
                  (##sys#check-string fname name)
                  #f))))
-    (set! stat-regular? regular-file?)	; DEPRECATED
-    (set! stat-directory? (stat-type 'stat-directory?)) ; DEPRECATED
-    (set! stat-device? (stat-type 'stat-char-device?))	; DEPRECATED
     (set! character-device? (stat-type 'character-device?))
     (set! block-device? (stat-type 'block-device?))
-    (set! stat-block-device? (stat-type 'stat-block-device?)) ; DEPRECATED
-    (set! stat-fifo? (stat-type 'stat-fifo?))		      ; DEPRECATED
     (set! fifo? (stat-type 'fifo?))
-    (set! stat-symlink? (stat-type 'stat-symlink?)) ; DEPRECATED
-    (set! socket? (stat-type 'socket?))
-    (set! stat-socket? (stat-type 'stat-socket?))) ; DEPRECATED
+    (set! socket? (stat-type 'socket?)))
 
 (define set-file-position!
-  (lambda (port pos . whence)
-    (let ([whence (if (pair? whence) (car whence) _seek_set)])
-      (##sys#check-exact pos 'set-file-position!)
-      (##sys#check-exact whence 'set-file-position!)
-      (when (fx< pos 0) (##sys#signal-hook #:bounds-error 'set-file-position! "invalid negative port position" pos port))
-      (unless (cond [(port? port)
-		     (and (eq? (##sys#slot port 7) 'stream)
-			  (##core#inline "C_fseek" port pos whence) ) ]
-		    [(fixnum? port) (##core#inline "C_lseek" port pos whence)]
-		    [else (##sys#signal-hook #:type-error 'set-file-position! "invalid file" port)] )
-	(##sys#update-errno)
-	(##sys#signal-hook #:file-error 'set-file-position! "cannot set file position" port pos) ) ) ) )
+   (lambda (port pos . whence)
+     (let ((whence (if (pair? whence) (car whence) _seek_set)))
+       (##sys#check-exact pos 'set-file-position!)
+       (##sys#check-exact whence 'set-file-position!)
+       (when (negative? pos)
+         (##sys#signal-hook #:bounds-error 'set-file-position! "invalid negative port position" pos port))
+       (unless (cond ((port? port)
+		      (and (eq? (##sys#slot port 7) 'stream)
+			   (##core#inline "C_fseek" port pos whence) ) )
+		     ((fixnum? port)
+		      (##core#inline "C_lseek" port pos whence))
+		     (else
+		      (##sys#signal-hook #:type-error 'set-file-position! "invalid file" port)) )
+	 (posix-error #:file-error 'set-file-position! "cannot set file position" port pos) ) ) ) )
 
 (define file-position
   (getter-with-setter
    (lambda (port)
-     (let ([pos (cond [(port? port)
+     (let ((pos (cond ((port? port)
 		       (if (eq? (##sys#slot port 7) 'stream)
-			   (##core#inline "C_ftell" port)
-			   -1) ]
-		      [(fixnum? port) (##core#inline "C_lseek" port 0 _seek_cur)]
-		      [else (##sys#signal-hook #:type-error 'file-position "invalid file" port)] ) ] )
-       (when (fx< pos 0)
-	 (##sys#update-errno)
-	 (##sys#signal-hook #:file-error 'file-position "cannot retrieve file position of port" port) )
+			   (##core#inline "C_ftell" port) 
+			   -1) )
+		      ((fixnum? port)
+		       (##core#inline "C_lseek" port 0 _seek_cur))
+		      (else
+		       (##sys#signal-hook #:type-error 'file-position "invalid file" port)) ) ) )
+       (when (< pos 0)
+	 (posix-error #:file-error 'file-position "cannot retrieve file position of port" port) )
        pos) )
-   set-file-position!) )		; doesn't accept WHENCE argument
+   set-file-position!) )		; doesn't accept WHENCE
 
 
 ;;; Directory stuff:
 
 (define-inline (create-directory-helper name)
-    (unless (zero? (##core#inline "C_mkdir" (##sys#make-c-string name)))
+    (unless (fx= 0 (##core#inline "C_mkdir" (##sys#make-c-string name)))
             (##sys#update-errno)
             (##sys#signal-hook #:file-error 'create-directory
                                "cannot create directory" name)))
@@ -1170,43 +1190,44 @@ EOF
 
 (define-inline (create-directory-helper-silent name)
     (unless (create-directory-check name)
-            (create-directory-helper name)))
+      (create-directory-helper name)))
 
 (define-inline (create-directory-helper-parents name)
-    (let* ((l   (string-split name "\\"))
-           (c   (car l)))
-        (for-each
-             (lambda (x)
-                 (set! c (string-append c "\\" x))
-                 (create-directory-helper-silent c))
-             (cdr l))))
+  (let* ((l   (string-split name "/\\"))
+	 (c   (car l)))
+    (for-each
+     (lambda (x)
+       (set! c (string-append c "/" x))
+       (create-directory-helper-silent c))
+     (cdr l))))
 
 (define create-directory
   (lambda (name #!optional parents?)
     (##sys#check-string name 'create-directory)
-    (if parents?
-        (create-directory-helper-parents (canonical-path name))
-        (create-directory-helper (canonical-path name)))))
-;(define create-directory
-;  (lambda (name)
-;    (##sys#check-string name 'create-directory)
-;    (unless (zero? (##core#inline "C_mkdir" (##sys#make-c-string (##sys#expand-home-path name))))
-;      (##sys#update-errno)
-;      (##sys#signal-hook #:file-error 'create-directory "cannot create directory" name) ) ) )
+    (let ((name (##sys#expand-home-path name)))
+      (if parents?
+          (create-directory-helper-parents name)
+          (create-directory-helper name))
+      name)))
 
 (define change-directory
   (lambda (name)
     (##sys#check-string name 'change-directory)
-    (unless (zero? (##core#inline "C_chdir" (##sys#make-c-string (##sys#expand-home-path name))))
-      (##sys#update-errno)
-      (##sys#signal-hook #:file-error 'change-directory "cannot change current directory" name) ) ) )
+    (let ((sname (##sys#make-c-string (##sys#expand-home-path name))))
+      (unless (fx= 0 (##core#inline "C_chdir" sname))
+	(##sys#update-errno)
+	(##sys#signal-hook
+	 #:file-error 'change-directory "cannot change current directory" name) )
+      name)))
 
 (define delete-directory
   (lambda (name)
     (##sys#check-string name 'delete-directory)
-    (unless (zero? (##core#inline "C_rmdir" (##sys#make-c-string (##sys#expand-home-path name))))
-      (##sys#update-errno)
-      (##sys#signal-hook #:file-error 'delete-directory "cannot delete directory" name) ) ) )
+    (let ((sname (##sys#make-c-string (##sys#expand-home-path name))))
+      (unless (fx= 0 (##core#inline "C_rmdir" sname))
+	(##sys#update-errno)
+	(##sys#signal-hook #:file-error 'delete-directory "cannot delete directory" name) )
+      name)))
 
 (define directory
   (let ([string-append string-append]
@@ -1258,7 +1279,7 @@ EOF
 		(##sys#signal-hook #:file-error 'current-directory "cannot retrieve current directory") ) ) ) ) ) )
 
 
-(define canonical-path
+(define canonical-path                                  ;;DEPRECATED
     (let ((null?      null?)
           (char=?     char=?)
           (string=?   string=?)
@@ -1589,7 +1610,7 @@ EOF
 (let ()
   (define (check filename acc loc)
     (##sys#check-string filename loc)
-    (let ([r (fx= 0 (##core#inline "C_access" (##sys#make-c-string (##sys#expand-home-path filename)) acc))])
+    (let ([r (fx= 0 (##core#inline "C_test_access" (##sys#make-c-string (##sys#expand-home-path filename)) acc))])
       (unless r (##sys#update-errno))
       r) )
   (set! file-read-access? (lambda (filename) (check filename _r_ok 'file-read-access?)))
@@ -1695,17 +1716,17 @@ EOF
   (when (fx< (##sys#size tm) 10)
     (##sys#error loc "time vector too short" tm) ) )
 
-(define (seconds->local-time secs)
+(define (seconds->local-time #!optional (secs (current-seconds)))
   (##sys#check-number secs 'seconds->local-time)
   (##sys#decode-seconds secs #f) )
 
-(define (seconds->utc-time secs)
+(define (seconds->utc-time #!optional (secs (current-seconds)))
   (##sys#check-number secs 'seconds->utc-time)
   (##sys#decode-seconds secs #t) )
 
 (define seconds->string
   (let ([ctime (foreign-lambda c-string "C_ctime" integer)])
-    (lambda (secs)
+    (lambda (#!optional (secs (current-seconds)))
       (let ([str (ctime secs)])
         (if str
             (##sys#substring str 0 (fx- (##sys#size str) 1))
@@ -1728,9 +1749,10 @@ EOF
 
 (define (local-time->seconds tm)
   (check-time-vector 'local-time->seconds tm)
-  (if (##core#inline "C_mktime" tm)
-      (##sys#cons-flonum)
-      (##sys#error 'local-time->seconds "cannot convert time vector to seconds" tm) ) )
+  (let ((t (##core#inline_allocate ("C_a_mktime" 4) tm)))
+    (if (fp= t -1.0)
+	(##sys#error 'local-time->seconds "cannot convert time vector to seconds" tm) 
+	t)))
 
 (define local-timezone-abbreviation
   (foreign-lambda* c-string ()
@@ -1746,7 +1768,13 @@ EOF
 
 (define (terminal-port? port)
   (##sys#check-port port 'terminal-port?)
-  #f)
+  (let ([fp (##sys#peek-unsigned-integer port 0)])
+    (and (not (eq? 0 fp)) (##core#inline "C_tty_portp" port) ) ) )
+
+(define (terminal-size port)
+  (if (terminal-port? port)
+      (values 0 0)
+      (##sys#error 'terminal-size "port is not connected to a terminal" port)))
 
 (define-foreign-variable _iofbf int "_IOFBF")
 (define-foreign-variable _iolbf int "_IOLBF")
@@ -1858,7 +1886,7 @@ EOF
       (##sys#check-string filename loc)
       (let ([s (pathname-strip-directory filename)])
 	(setarg 0 s (##sys#size s)) )
-      (build-exec-argvec loc ($quote-args-list arglst exactf) setarg 1)
+      (build-exec-argvec loc (and arglst ($quote-args-list arglst exactf)) setarg 1)
       (build-exec-argvec loc envlst setenv 0)
       (##core#inline "C_flushall")
       (##sys#make-c-string (##sys#expand-home-path filename)) ) ) )
@@ -1949,8 +1977,8 @@ EOF
 		(##sys#update-errno)
 		(##sys#signal-hook #:process-error loc "cannot execute process" cmdlin))) ) ) ) ) ) )
 
-#;(define process (void))
-#;(define process* (void))
+(define process)
+(define process*)
 (let ([%process
 	(lambda (loc err? cmd args env exactf)
 	  (let ([chkstrlst
@@ -2037,10 +2065,12 @@ EOF
   (let ([glob glob]
 	[string-match string-match]
 	[make-pathname make-pathname]
+	[pathname-file pathname-file]
 	[directory? directory?] )
     (lambda (dir pred . action-id-limit)
-      (let-optionals action-id-limit
-	  ([action (lambda (x y) (cons x y))] ; no eta reduction here - we want cons inlined.
+      (let-optionals
+	  action-id-limit
+	  ([action (lambda (x y) (cons x y))] ; we want cons inlined
 	   [id '()]
 	   [limit #f] )
 	(##sys#check-string dir 'find-files)
@@ -2050,7 +2080,7 @@ EOF
 		      [(fixnum? limit) (lambda _ (fx< depth limit))]
 		      [else limit] ) ]
 	       [pproc
-		(if (string? pred)
+		(if (or (string? pred) (regexp? pred))
 		    (lambda (x) (string-match pred x))
 		    pred) ] )
 	  (let loop ([fs (glob (make-pathname dir "*"))]
@@ -2064,8 +2094,9 @@ EOF
 			       [(lproc f)
 				(loop rest
 				      (fluid-let ([depth (fx+ depth 1)])
-					(loop (glob (make-pathname f "*")) r) ) ) ]
-			       [else (loop rest r)] ) ]
+					(loop (glob (make-pathname f "*"))
+					      (if (pproc f) (action f r) r)) ) ) ]
+			       [else (loop rest (if (pproc f) (action f r) r))] ) ]
 			[(pproc f) (loop rest (action f r))]
 			[else (loop rest r)] ) ) ) ) ) ) ) ) )
 
@@ -2115,7 +2146,6 @@ EOF
 (define-unimplemented signal-masked?)
 (define-unimplemented signal-unmask!)
 (define-unimplemented terminal-name)
-(define-unimplemented terminal-size)
 (define-unimplemented unmap-file-from-memory)
 (define-unimplemented user-information)
 (define-unimplemented utc-time->seconds)
