@@ -1,6 +1,6 @@
 ;;;; setup-api.scm - build + installation API for eggs
 ;
-; Copyright (c) 2008-2009, The Chicken Team
+; Copyright (c) 2008-2010, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -29,19 +29,24 @@
 ; This code is partially quite messy and the API is not overly consistent,
 ; mainly because it has grown "organically" while the old chicken-setup program
 ; evolved. The code was extracted and put into this module, without much
-; cleaning up. Nevertheless, it should work.
+; cleaning up.
+;
+; *windows-shell* and, to a lesser extent, 'sudo' processing knowledge is
+; scattered in the code.
 
 (module setup-api
 
     (move-file 
      (run execute)
      compile
+     standard-extension
      make make/proc
      host-extension
      install-extension install-program install-script
-     setup-verbose-mode setup-install-mode
-     setup-verbose-flag setup-install-flag			; DEPRECATED
-     installation-prefix chicken-prefix 
+     setup-verbose-mode setup-install-mode deployment-mode
+     installation-prefix
+     destination-prefix
+     chicken-prefix 			;XXX remove at some stage from exports
      find-library find-header 
      program-path remove-file* 
      patch yes-or-no? abort-setup
@@ -57,6 +62,7 @@
      remove-directory
      remove-extension
      read-info
+     register-program find-program
      shellpath)
   
   (import scheme chicken foreign
@@ -67,12 +73,6 @@
 
 (define-constant setup-file-extension "setup-info")
 
-(define *installed-executables* 
-  `(("chicken" . ,(foreign-value "C_CHICKEN_PROGRAM" c-string))
-    ("csc" . ,(foreign-value "C_CSC_PROGRAM" c-string))
-    ("csi" . ,(foreign-value "C_CSI_PROGRAM" c-string))
-    ("chicken-bug" . ,(foreign-value "C_CHICKEN_BUG_PROGRAM" c-string))))
-
 (define *cc* (foreign-value "C_TARGET_CC" c-string))
 (define *cxx* (foreign-value "C_TARGET_CXX" c-string))
 (define *target-cflags* (foreign-value "C_TARGET_CFLAGS" c-string))
@@ -80,6 +80,7 @@
 (define *target-lib-home* (foreign-value "C_TARGET_LIB_HOME" c-string))
 (define *sudo* #f)
 (define *windows-shell* (foreign-value "C_WINDOWS_SHELL" bool))
+(define *registered-programs* '())
 
 (define *windows*
   (and (eq? (software-type) 'windows) 
@@ -94,24 +95,14 @@
 	(make-pathname p "bin") )
       (foreign-value "C_INSTALL_BIN_HOME" c-string) ) )
 
-(define *doc-path*
-  (or (and-let* ((p (get-environment-variable "CHICKEN_PREFIX")))
-	(make-pathname p "share/chicken/doc") )
-      (make-pathname
-       (foreign-value "C_INSTALL_SHARE_HOME" c-string) 
-       "doc")))
-
-(define chicken-prefix
+(define chicken-prefix			
   (or (get-environment-variable "CHICKEN_PREFIX")
-      (let ((m (string-match "(.*)/bin/?" *chicken-bin-path*)))
-	(if m
-	    (cadr m)
-	    "/usr/local") ) ) )
+      (foreign-value "C_INSTALL_PREFIX" c-string)))
 
 (define (shellpath str)
   (qs (normalize-pathname str)))
 
-(define (cross-chicken) (##sys#fudge 39))
+(define (cross-chicken) (##sys#fudge 39)) ; DEPRECATED
 
 (define *csc-options* '())
 (define *base-directory* (current-directory))
@@ -119,8 +110,7 @@
 (define setup-root-directory      (make-parameter *base-directory*))
 (define setup-verbose-mode        (make-parameter #f))
 (define setup-install-mode        (make-parameter #t))
-(define setup-verbose-flag setup-verbose-mode) ; DEPRECATED
-(define setup-install-flag setup-install-mode) ; DEPRECATED
+(define deployment-mode           (make-parameter #f))
 (define program-path              (make-parameter *chicken-bin-path*))
 (define keep-intermediates (make-parameter #f))
 
@@ -134,18 +124,18 @@
 (define *mkdir-command*)
 
 (define (windows-user-install-setup)
-  (set! *copy-command*        'copy)
+  (set! *copy-command*        "copy")
   (set! *remove-command*      "del /Q /S")
-  (set! *move-command*        'move)
+  (set! *move-command*        "move")
   (set! *chmod-command*       "chmod")
   (set! *ranlib-command*      "ranlib") )
 
 (define (unix-user-install-setup)
   (set! *copy-command*        "cp -r")
   (set! *remove-command*      "rm -fr")
-  (set! *move-command*        'mv)
+  (set! *move-command*        "mv")
   (set! *chmod-command*       "chmod")
-  (set! *ranlib-command*      "ranlib") 
+  (set! *ranlib-command*      "ranlib")
   (set! *mkdir-command*       "mkdir") )
 
 (define (windows-sudo-install-setup)
@@ -157,7 +147,7 @@
   (set! *remove-command*      "sudo rm -fr")
   (set! *move-command*        "sudo mv")
   (set! *chmod-command*       "sudo chmod")
-  (set! *ranlib-command*      "sudo ranlib") 
+  (set! *ranlib-command*      "sudo ranlib")
   (set! *mkdir-command*       "sudo mkdir") )
 
 (define (user-install-setup)
@@ -177,30 +167,7 @@
         ((car args)     (sudo-install-setup))
         (else           (user-install-setup)) ) )
 
-; User setup by default
-(user-install-setup)
-
-
-(define create-directory/parents
-  (let ()
-    (define create-directory-0
-      (let ([create-directory create-directory])
-	(lambda (dir)
-	  (let loop ([dir dir])
-	    (when (and dir (not (directory? dir)))
-	      (loop (pathname-directory dir))
-	      (create-directory dir))) ) ) )
-    (define (verb dir)
-      (when (setup-verbose-mode) (printf "  creating directory `~a'~%~!" dir)) )
-    (if *windows-shell*
-	(lambda (dir)
-	  (verb dir)
-	  (create-directory-0 dir) ) 
-	(lambda (dir)
-	  (verb dir)
-	  ($system (sprintf "mkdir -p ~a" (shellpath dir) ) ) ) ) ) )
-
-(define abort-setup (make-parameter exit))
+(define abort-setup (make-parameter (cut exit 1)))
 
 (define (yes-or-no? str #!key default (abort (abort-setup)))
   (let loop ()
@@ -237,21 +204,44 @@
 
 (define run-verbose (make-parameter #t))
 
+(define (register-program name #!optional
+			  (path (make-pathname *chicken-bin-path* (->string name))))
+  (set! *registered-programs* 
+    (alist-cons (->string name) path *registered-programs*)))
+
+(define (find-program name)
+  (let* ((name (->string name))
+	 (a (assoc name *registered-programs*)))
+    (if a
+	(shellpath (cdr a))
+	name)))
+
+(let ()
+  (define (reg name rname) 
+    (register-program name (make-pathname *chicken-bin-path* rname)))
+  (reg "chicken" (foreign-value "C_CHICKEN_PROGRAM" c-string))
+  (reg "csi" (foreign-value "C_CSI_PROGRAM" c-string))
+  (reg "csc" (foreign-value "C_CSC_PROGRAM" c-string))
+  (reg "chicken-install" (foreign-value "C_CHICKEN_INSTALL_PROGRAM" c-string))
+  (reg "chicken-uninstall" (foreign-value "C_CHICKEN_UNINSTALL_PROGRAM" c-string))
+  (reg "chicken-status" (foreign-value "C_CHICKEN_STATUS_PROGRAM" c-string))
+  (reg "chicken-bug" (foreign-value "C_CHICKEN_BUG_PROGRAM" c-string)))
+
 (define (fixpath prg)
-  (cond ((string=? prg "csc")
-	 (string-intersperse 
-	  (cons* (shellpath
-		  (make-pathname 
-		   *chicken-bin-path*
-		   (cdr (assoc prg *installed-executables*))))
-		 "-feature" "compiling-extension" "-setup-mode"
-		 (if (keep-intermediates) "-k" "")
-		 (if (host-extension) "-host" "")
-		 *csc-options*) 
-	  " ") )
-	((assoc prg *installed-executables*) =>
-	 (lambda (a) (shellpath (make-pathname *chicken-bin-path* (cdr a)))))
-	(else prg) ) )
+  (if (string=? prg "csc")
+      (string-intersperse 
+       (cons*
+	(shellpath (find-program "csc"))
+	"-feature" "compiling-extension" 
+	(if (and (feature? #:cross-chicken)
+		 (not (host-extension)))
+	    "" "-setup-mode")
+	(if (keep-intermediates) "-k" "")
+	(if (host-extension) "-host" "")
+	(if (deployment-mode) "-deployed" "")
+	*csc-options*) 
+       " ")
+      (find-program prg)))
 
 (define (fixmaketarget file)
   (if (and (equal? "so" (pathname-extension file))
@@ -437,13 +427,20 @@
 (define (make-setup-info-pathname fn #!optional (rpath (repository-path)))
   (make-pathname rpath fn setup-file-extension) )
 
+(define destination-prefix (make-parameter #f))
+
 (define installation-prefix
-  (make-parameter (or (get-environment-variable "CHICKEN_INSTALL_PREFIX") #f)))
+  (let ((prefix (get-environment-variable "CHICKEN_INSTALL_PREFIX")))
+    (lambda ()
+      (or (destination-prefix)
+	  prefix
+	  chicken-prefix))))
 
 (define create-directory/parents
   (let ()
     (define (verb dir)
-      (when (setup-verbose-mode) (printf "  creating directory `~a'~%~!" dir)) )
+      (when (setup-verbose-mode)
+	(printf "  mkdir ~a~%~!" dir)) )
     (if *windows*
         (lambda (dir)
           (verb dir)
@@ -469,16 +466,19 @@
   ;;XXX the prefix handling is completely bogus
   (let ((from (if (pair? from) (car from) from))
 	(to (let ((to-path (if (pair? from) (make-pathname to (cadr from)) to)))
-	      (if (and prefix (not (string-prefix? prefix to-path)))
-		  (make-pathname prefix to-path) 
+	      (if (not (path-prefix? prefix to-path))
+		  (if (absolute-pathname? to-path)
+		      to-path
+		      (make-pathname prefix to-path) )
 		  to-path))))
     (ensure-directory to)
-    (cond ((or (glob? from) (file-exists? from))
-	   (begin
-	     (run (,*copy-command* ,(shellpath from) ,(shellpath to))) 
-	     to))
-	  (err (error "file does not exist" from))
-	  (else (warning "file does not exist" from)))))
+    (run (,*copy-command* ,(shellpath from) ,(shellpath to)))
+    to))
+
+(define (path-prefix? pref path)
+  (string-prefix?
+   (normalize-pathname pref)
+   (normalize-pathname path)))
 
 (define (move-file from to)
   (let ((from  (if (pair? from) (car from) from))
@@ -513,6 +513,26 @@
 	   (else ext)))))
 
 
+;;; Convenience function
+
+(define (standard-extension name version #!key (static #t) (info '()))
+  (let* ((sname (->string name))
+	 (fname (make-pathname #f sname "scm"))
+	 (iname (make-pathname #f sname "import.scm")))
+    (compile -s -O3 -d1 ,fname -j ,name)
+    (when static
+      (compile -c -O3 -d1 ,fname -j ,name -unit ,name))
+    (compile -s -O3 -d0 ,iname)
+    (install-extension
+     name
+     (list (pathname-replace-extension fname "so")
+	   (pathname-replace-extension iname "so")
+	   (make-pathname #f sname "setup"))
+     `((version ,version)
+       ,@info
+       ,@(if static `((static ,(make-pathname #f fname "o"))) '())))))
+
+
 ;;; Installation
 
 (define (install-extension id files #!optional (info '()))
@@ -535,28 +555,7 @@
 					(equal? (pathname-extension to) "a"))
 			       (run (,*ranlib-command* ,(shellpath to)) ) ))
 			   (make-dest-pathname rpath f)))
-		       files) ) 
-	   (pre (installation-prefix))
-	   (docpath (if pre
-			(ensure-directory (make-pathname pre "share/chicken/doc"))
-			*doc-path*)))
-      (and-let* ((docs (assq 'documentation info)))
-	(print "\n* Installing documentation files in " docpath ":")
-	(for-each
-	 (lambda (f)
-	   (copy-file f (make-pathname docpath f) #f) )
-	 (cdr docs))
-	(newline))
-      (and-let* ((exs (assq 'examples info)))
-	(print "\n* Installing example files in " docpath ":")
-	(for-each 
-	 (lambda (f)
-	   (let ((destf (make-pathname docpath f)))
-	     (copy-file f destf #f)
-	     (unless *windows-shell*
-	       (run (,*chmod-command* a+rx ,destf)) ) ))
-	 (cdr exs))
-	(newline))
+		       files) ) )
       (write-info id dests info) ) ) )
 
 (define (install-program id files #!optional (info '()))
@@ -566,11 +565,8 @@
      (if *windows-shell* "exe" #f) ) )
   (when (setup-install-mode)
     (let* ((files (check-filelist (if (list? files) files (list files))))
-	   (ppath ((lambda (pre)
-		     (if pre 
-			 (ensure-directory (make-pathname pre "bin"))
-			 (program-path)))
-		   (installation-prefix)))
+	   (pre (installation-prefix))
+	   (ppath (ensure-directory (make-pathname pre "bin")))
 	   (files (if *windows*
                       (map (lambda (f)
                              (if (list? f) 
@@ -591,17 +587,14 @@
 (define (install-script id files #!optional (info '()))
   (when (setup-install-mode)
     (let* ((files (check-filelist (if (list? files) files (list files))))
-	   (ppath ((lambda (pre) 
-		     (if pre
-			 (ensure-directory (make-pathname pre "bin"))
-			 (program-path)))
-		   (installation-prefix)))
+	   (pre (installation-prefix))
+	   (ppath (ensure-directory (make-pathname pre "bin")))
 	   (pfiles (map (lambda (f)
 			  (let ((from (if (pair? f) (car f) f))
 				(to (make-dest-pathname ppath f)) )
 			    (copy-file from to) 
 			    (unless *windows-shell*
-				    (run (,*chmod-command* a+r ,(shellpath to))))
+			      (run (,*chmod-command* a+r ,(shellpath to))))
 			    to) )
 			files) ) )
       (unless *windows-shell*
@@ -612,10 +605,15 @@
 ;;; More helper stuff
 
 (define (repo-path #!optional ddir?)
-  (let ((p (if (and ddir? (installation-prefix))
-	       (make-pathname 
-		(installation-prefix) 
-		(sprintf "lib/chicken/~a" (##sys#fudge 42)))
+  (let ((p (if ddir?
+	       (if (deployment-mode)
+		   (installation-prefix) ; deploy: copy directly into destdir
+		   (let ((p (destination-prefix)))
+		     (if p		; installation-prefix changed: use it
+			 (make-pathname 
+			  p
+			  (sprintf "lib/chicken/~a" (##sys#fudge 42)))
+			 (repository-path)))) ; otherwise use repo-path
 	       (repository-path))) )
     (ensure-directory p)
     p) )
@@ -628,7 +626,7 @@
 	(begin
 	  (create-directory/parents dir)
 	  (unless *windows-shell*
-		  (run (,*chmod-command* a+x ,(shellpath dir)))))))
+	    (run (,*chmod-command* a+x ,(shellpath dir)))))))
   path)
 
 (define (try-compile code #!key c++ (cc (if c++ *cxx* *cc*)) (cflags "") (ldflags "") 
@@ -658,11 +656,11 @@
   (when (version>=? v (chicken-version) ) 
     (error (sprintf "CHICKEN version ~a or higher is required" v)) ) )
 
-(define (upgrade-message ext msg)
+(define (upgrade-message ext msg #!optional version)
   (error
    (sprintf
-    "the required extension `~s' ~a - please run~%~%  chicken-install ~a~%~%and repeat the current installation operation."
-    ext msg ext) ) )
+    "the required extension `~s' ~a - please run~%~%  chicken-install ~a~a~%~%and repeat the current installation operation."
+    ext msg ext (if version (conc ":" version) "")) ) )
 
 (define (required-extension-version . args)
   (let loop ((args args))
@@ -679,7 +677,8 @@
 			  (upgrade-message 
 			   ext
 			   (sprintf "is older than ~a, which is what this extension requires"
-				    version) ) )
+				    version) 
+                           version) )
 			 (else (loop more)) ) ) 
 		 (upgrade-message ext "is not installed") ) ) )
 	  (else 
@@ -786,6 +785,16 @@
 		(string-append "\"" str "\"")	; double quotes, yes - thanks to Matthew Flatt
 		str))))
     (unless (zero? r)
-      (error "shell command failed with nonzero exit status" r str))))
+      (quit "shell command failed with nonzero exit status ~a:~%~%  ~a" r str))))
+
+(define (quit fstr . args)
+  (flush-output)
+  (fprintf (current-error-port) "~%~?~%" fstr args)
+  (reset))
+
+;;; Module Setup
+
+; User setup by default
+(user-install-setup)
 
 )
