@@ -10,7 +10,7 @@
 	%string-map %string-copy! %string-compare %substring/shared %string-suffix? %multispan-repcopy!
 	%string-prefix-length-ci %string-suffix-length-ci %string-prefix-ci? %string-suffix-ci?
 	##srfi13#traverse
-	%string-titlecase! %string-map! %kmp-search %string-compare-ci ##srfi13#string-fill!)
+	%string-titlecase! %string-map! %string-compare-ci ##srfi13#string-fill!)
   (standard-bindings not boolean? apply call-with-current-continuation eq? eqv? equal? pair? cons car cdr caar cadr
 		     cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr caaaar caaadr caadar caaddr cadaar
 		     cadadr caddar cadddr cdaaar cdaadr cdadar cdaddr cddaar cddadr cdddar cddddr set-car! set-cdr!
@@ -44,22 +44,7 @@
      string-parse-start+end substring-spec-ok?)
     (no-bound-checks) ) ] )
 
-(cond-expand
- [unsafe
-  (eval-when (compile)
-    (define-macro (##sys#check-structure . _) '(##core#undefined))
-    (define-macro (##sys#check-range . _) '(##core#undefined))
-    (define-macro (##sys#check-pair . _) '(##core#undefined))
-    (define-macro (##sys#check-list . _) '(##core#undefined))
-    (define-macro (##sys#check-symbol . _) '(##core#undefined))
-    (define-macro (##sys#check-string . _) '(##core#undefined))
-    (define-macro (##sys#check-char . _) '(##core#undefined))
-    (define-macro (##sys#check-exact . _) '(##core#undefined))
-    (define-macro (##sys#check-port . _) '(##core#undefined))
-    (define-macro (##sys#check-number . _) '(##core#undefined))
-    (define-macro (##sys#check-bytevector . _) '(##core#undefined)) ) ]
- [else
-  (declare (emit-exports "srfi-13.exports"))] )
+(include "unsafe-declarations.scm")
 
 (register-feature! 'srfi-13)
 
@@ -185,16 +170,34 @@
 ;;; Support for START/END substring specs
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(eval-when (compile eval)
-  (define-macro (let-string-start+end2 s-e proc s1 s2 args . body)
-    (let ([procv (gensym)]
-	  [rest (gensym)] )
-      `(let ((,procv ,proc))
-	 (let-string-start+end 
-	  (,(car s-e) ,(cadr s-e) ,rest) ,procv ,s1 ,args
-	  (let-string-start+end 
-	   ,(cddr s-e) ,procv ,s2 ,rest
-	   ,@body) ) ) ) ) )
+(define-syntax let-string-start+end2
+  (syntax-rules ()
+    ((_ (s-e1 s-e2 s-e3 s-e4) proc s1 s2 args . body)
+     (let ((procv proc))
+       (let-string-start+end 
+	(s-e1 s-e2 rest) procv s1 args
+	(let-string-start+end 
+	 (s-e3 s-e4) procv s2 rest
+	 . body) ) ) ) ) )
+
+(define-syntax let-string-start+end
+  (lambda (form r c)
+    (##sys#check-syntax 'let-string-start+end form '(_ _ _ _ _ . _))
+    (let ((s-e-r (cadr form))
+	  (proc (caddr form))
+	  (s-exp (cadddr form))
+	  (args-exp (car (cddddr form)))
+	  (body (cdr (cddddr form)))
+	  (%receive (r 'receive))
+	  (%string-parse-start+end (r 'string-parse-start+end))
+	  (%string-parse-final-start+end (r 'string-parse-final-start+end)))
+      (if (pair? (cddr s-e-r))
+	  `(,%receive (,(caddr s-e-r) ,(car s-e-r) ,(cadr s-e-r))
+		      (,%string-parse-start+end ,proc ,s-exp ,args-exp)
+		      ,@body)
+	  `(,%receive ,s-e-r
+		      (,%string-parse-final-start+end ,proc ,s-exp ,args-exp)
+		      ,@body) ) )))
 
 
 ;;; Returns three values: rest start end
@@ -272,7 +275,7 @@
   (let ((slen (string-length s)))
 ;    (check-arg (lambda (start) (and (integer? start) (exact? start) (<= 0 start)))
 ;	       start substring/shared)
-    (let ([n (:optional maybe-end slen)])
+    (let ([n (optional maybe-end slen)])
       (##sys#check-exact n 'substring/shared)
       (check-substring-spec 'substring/shared s start n)
       (%substring/shared s start n) ) ) )
@@ -322,8 +325,8 @@
 ;;; You want compiler support for high-level transforms on fold and unfold ops.
 ;;; You'd at least like a lot of inlining for clients of these procedures.
 ;;; Don't hold your breath.
-;;;
-;;; Shut up, Olin.
+
+;;; Shut up, Olin (flw)
 
 (define (string-map proc s . maybe-start+end)
 ;  (check-arg procedure? proc string-map)
@@ -1380,7 +1383,7 @@
 ;;; Searching for an occurrence of a substring
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-; this is broken. Thanks, Olin! ;-)
+; this is completely broken and was probably never tested. Thanks, Olin! (flw)
 
 
 ;;; Knuth-Morris-Pratt string searching
@@ -1397,30 +1400,6 @@
 ;;; This algorithm is O(m + n) where m and n are the 
 ;;; lengths of the pattern and string respectively
 
-;;; KMP search source[start,end) for PATTERN. Return starting index of
-;;; leftmost match or #f.
-
-(define (%kmp-search pattern text c= p-start p-end t-start t-end)
-  (let ((plen (- p-end p-start))
-	(rv (make-kmp-restart-vector pattern c= p-start p-end)))
-
-    ;; The search loop. TJ & PJ are redundant state.
-    (let lp ((ti t-start) (pi 0)
-	     (tj (- t-end t-start))	; (- tlen ti) -- how many chars left.
-	     (pj plen))			; (- plen pi) -- how many chars left.
-
-      (if (= pi plen) (- ti plen)			; Win.
-	  
-	  (and (<= pj tj)				; Lose.
-		 
-	       (if (c= (string-ref text ti)		; Search.
-		       (string-ref pattern (+ p-start pi)))
-		   (lp (+ 1 ti) (+ 1 pi) (- tj 1) (- pj 1))	; Advance.
-		   
-		   (let ((pi (vector-ref rv pi)))		; Retreat.
-		     (if (= pi -1)
-			 (lp (+ ti 1)  0   (- tj 1)  plen)	; Punt.
-			 (lp ti        pi  tj        (- plen pi))))))))))
 
 ;;; (make-kmp-restart-vector pattern [c= start end]) -> integer-vector
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -2036,6 +2015,9 @@
 ;;; The KMP string-search code was influenced by implementations written
 ;;; by Stephen Bevan, Brian Dehneyer and Will Fitzgerald. However, this
 ;;; version was written from scratch by myself.
+
+;;; I guessed that much. (flw)
+
 ;;;
 ;;; The remainder of this code was written from scratch by myself for scsh.
 ;;; The scsh copyright is a BSD-style open source copyright. See below for
