@@ -23,22 +23,7 @@
      char-set? char-set-size char-set:s)
     (no-bound-checks) ) ] )
 
-(cond-expand
- [unsafe
-  (eval-when (compile)
-    (define-macro (##sys#check-structure . _) '(##core#undefined))
-    (define-macro (##sys#check-range . _) '(##core#undefined))
-    (define-macro (##sys#check-pair . _) '(##core#undefined))
-    (define-macro (##sys#check-list . _) '(##core#undefined))
-    (define-macro (##sys#check-symbol . _) '(##core#undefined))
-    (define-macro (##sys#check-string . _) '(##core#undefined))
-    (define-macro (##sys#check-char . _) '(##core#undefined))
-    (define-macro (##sys#check-exact . _) '(##core#undefined))
-    (define-macro (##sys#check-port . _) '(##core#undefined))
-    (define-macro (##sys#check-number . _) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
- [else
-  (declare (emit-exports "srfi-14.exports"))] )
+(include "unsafe-declarations.scm")
 
 (register-feature! 'srfi-14)
 
@@ -222,7 +207,7 @@
 ;;; and everything will be copacetic.
 
 (define (char-set-hash cs . maybe-bound)
-  (let ((bound (:optional maybe-bound 4194304)))
+  (let ((bound (optional maybe-bound 4194304)))
     (if (zero? bound) (set! bound 4194304))
     (##sys#check-exact bound 'char-set-hash)
     (let* ((s (%char-set:s/check cs 'char-set-hash))
