@@ -1,7 +1,7 @@
 ;;;; srfi-4.scm - Homogeneous numeric vectors
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -34,6 +34,7 @@
        ##sys#u32vector-set! ##sys#s32vector-set! ##sys#f32vector-set! ##sys#f64vector-set!
        ##sys#u8vector-ref ##sys#s8vector-ref ##sys#u16vector-ref ##sys#s16vector-ref subvector
        ##sys#u32vector-ref ##sys#s32vector-ref ##sys#f32vector-ref ##sys#f64vector-ref)
+ (not inline ##sys#user-print-hook ##sys#number-hash-hook)
  (foreign-declare #<<EOF
 #define C_u8peek(b, i)         C_fix(((unsigned char *)C_data_pointer(b))[ C_unfix(i) ])
 #define C_s8peek(b, i)         C_fix(((char *)C_data_pointer(b))[ C_unfix(i) ])
@@ -46,8 +47,6 @@
 # define C_a_u32peek(ptr, d, b, i) C_unsigned_int_to_num(ptr, ((C_u32 *)C_data_pointer(b))[ C_unfix(i) ])
 # define C_a_s32peek(ptr, d, b, i) C_int_to_num(ptr, ((C_s32 *)C_data_pointer(b))[ C_unfix(i) ])
 #endif
-#define C_f32peek(b, i)        (C_temporary_flonum = ((float *)C_data_pointer(b))[ C_unfix(i) ], C_SCHEME_UNDEFINED)
-#define C_f64peek(b, i)        (C_temporary_flonum = ((double *)C_data_pointer(b))[ C_unfix(i) ], C_SCHEME_UNDEFINED)
 #define C_u8poke(b, i, x)      ((((unsigned char *)C_data_pointer(b))[ C_unfix(i) ] = C_unfix(x)), C_SCHEME_UNDEFINED)
 #define C_s8poke(b, i, x)      ((((char *)C_data_pointer(b))[ C_unfix(i) ] = C_unfix(x)), C_SCHEME_UNDEFINED)
 #define C_u16poke(b, i, x)     ((((unsigned short *)C_data_pointer(b))[ C_unfix(i) ] = C_unfix(x)), C_SCHEME_UNDEFINED)
@@ -75,7 +74,7 @@ EOF
      ##sys#s32vector-set! read list->f64vector list->s32vector list->u32vector list->u16vector list-s8vector
      list->u8vector set-finalizer!
      ##sys#f32vector-ref ##sys#f32vector-set! ##sys#f64vector-ref ##sys#f64vector-set! ##sys#check-exact-interval
-     ##sys#check-inexact-interval ##sys#check-number ##sys#check-structure ##sys#cons-flonum ##sys#check-list
+     ##sys#check-inexact-interval ##sys#check-number ##sys#check-structure ##sys#check-list
      ##sys#check-range ##sys#error ##sys#signal-hook
      ##sys#error-not-a-proper-list ##sys#print ##sys#allocate-vector) ) ] )
 
@@ -108,12 +107,10 @@ EOF
 (define (##sys#s32vector-ref v i) (##core#inline_allocate ("C_a_s32peek" 4) (##core#inline "C_slot" v 1) i))
 
 (define (##sys#f32vector-ref v i)
-  (##core#inline "C_f32peek" (##core#inline "C_slot" v 1) i)
-  (##sys#cons-flonum) )
+  (##core#inline_allocate ("C_a_i_f32vector_ref" 4) v i))
 
 (define (##sys#f64vector-ref v i)
-  (##core#inline "C_f64peek" (##core#inline "C_slot" v 1) i)
-  (##sys#cons-flonum) )
+  (##core#inline_allocate ("C_a_i_f64vector_ref" 4) v i))
 
 (define (##sys#u8vector-set! v i x) (##core#inline "C_u8poke" (##core#inline "C_slot" v 1) i x))
 (define (##sys#s8vector-set! v i x) (##core#inline "C_s8poke" (##core#inline "C_slot" v 1) i x))
@@ -121,8 +118,12 @@ EOF
 (define (##sys#s16vector-set! v i x) (##core#inline "C_s16poke" (##core#inline "C_slot" v 1) i x))
 (define (##sys#u32vector-set! v i x) (##core#inline "C_u32poke" (##core#inline "C_slot" v 1) i x))
 (define (##sys#s32vector-set! v i x) (##core#inline "C_s32poke" (##core#inline "C_slot" v 1) i x))
-(define (##sys#f32vector-set! v i x) (##core#inline "C_f32poke" (##core#inline "C_slot" v 1) i x))
-(define (##sys#f64vector-set! v i x) (##core#inline "C_f64poke" (##core#inline "C_slot" v 1) i x))
+
+(define (##sys#f32vector-set! v i x)
+  (##core#inline "C_u_i_f32vector_set" v i x))
+
+(define (##sys#f64vector-set! v i x)
+  (##core#inline "C_u_i_f64vector_set" v i x))
 
 
 ;;; Get vector length:
