@@ -1,7 +1,7 @@
 ;;; srfi-69.scm - Optional non-standard extensions
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -51,14 +51,11 @@
     *hash-table-for-each *hash-table-fold
     hash-table-canonical-length hash-table-rehash! hash-table-check-resize! ) )
 
-(cond-expand
- [unsafe]
- [else
-   (declare
-     (bound-to-procedure
-       ##sys#check-string ##sys#check-symbol
-       ##sys#check-exact ##sy#check-inexact
-       ##sys#check-closure ##sys#check-structure ) ) ] )
+(declare
+  (bound-to-procedure
+   ##sys#check-string ##sys#check-symbol
+   ##sys#check-exact ##sy#check-inexact
+   ##sys#check-closure ##sys#check-structure ) )
 
 (include "unsafe-declarations.scm")
 
@@ -134,7 +131,8 @@
   (if (fx< fxn 0) (fxneg fxn) fxn ) )
 
 (define-inline (%hash/limit hsh lim)
-  (fxmod (fxand (foreign-value "C_MOST_POSITIVE_FIXNUM" int)
+  ;; use 32-bit mask to have identical hashes on 64-bit platforms
+  (fxmod (fxand (foreign-value "C_MOST_POSITIVE_32_BIT_FIXNUM" int)
 		(%fxabs hsh))
 	 lim) )
 
@@ -676,7 +674,7 @@
 (define hash-table-update!
   (let ([core-eq? eq?] )
     (lambda (ht key
-	     #!optional (func identity)
+	     #!optional (func (lambda (x) x))
 		        (thunk
 		         (let ([thunk (##sys#slot ht 9)])
 		           (or thunk
@@ -965,7 +963,7 @@
       (do ([lst (##sys#slot vec i) (##sys#slot lst 1)])
 	  [(null? lst)]
 	(let ([b (##sys#slot lst 0)])
-	  (*hash-table-update!/default ht1 (##sys#slot b 0) identity (##sys#slot b 1)) ) ) ) ) )
+	  (*hash-table-update!/default ht1 (##sys#slot b 0) (lambda (x) x) (##sys#slot b 1)) ) ) ) ) )
 
 (define (hash-table-merge! ht1 ht2)
   (##sys#check-structure ht1 'hash-table 'hash-table-merge!)
@@ -1002,7 +1000,7 @@
 	(for-each
 	 (lambda (x)
 	   (##sys#check-pair x 'alist->hash-table)
-	   (*hash-table-update!/default  ht (##sys#slot x 0) identity (##sys#slot x 1)) )
+	   (*hash-table-update!/default  ht (##sys#slot x 0) (lambda (x) x) (##sys#slot x 1)) )
 	 alist)
 	ht ) ) ) )
 
@@ -1087,3 +1085,13 @@
   (##sys#check-structure ht 'hash-table 'hash-table-map)
   (##sys#check-closure func 'hash-table-map)
   (*hash-table-fold ht (lambda (k v a) (cons (func k v) a)) '()) )
+
+
+;;; printing
+
+(##sys#register-record-printer
+ 'hash-table
+ (lambda (ht port)
+   (##sys#print "#<hash-table (" #f port)
+   (##sys#print (##sys#slot ht 2) #f port)
+   (##sys#print ")>" #f port) ) )
