@@ -1,7 +1,7 @@
 ;;;; support.scm - Miscellaneous support code for the CHICKEN compiler
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -25,13 +25,17 @@
 ; POSSIBILITY OF SUCH DAMAGE.
 
 
-(declare (unit support))
+(declare
+  (unit support))
 
 
 (include "compiler-namespace")
 (include "tweaks")
 (include "banner")
 
+(declare
+  (not inline compiler-cleanup-hook ##sys#user-read-hook) )
+
 
 ;;; Debugging and error-handling stuff:
 
@@ -173,7 +177,7 @@
   (cond [(string=? fname "-") (current-input-port)]
 	[(file-exists? fname) (open-input-file fname)]
 	[(or (null? line) (not (car line))) (quit "Can not open file ~s" fname)]
-	[else (quit "Can not open file ~s in line ~s" fname (car line))] ) )
+	[else (quit "(~a) can not open file ~s" (car line) fname)] ) )
 
 (define (close-checked-input-file port fname)
   (unless (string=? fname "-") (close-input-port port)) )
@@ -257,16 +261,15 @@
 		(if (exn? ex) 
 		    (exn-msg ex)
 		    (->string ex) ) ) 
-	(let ([xs (with-input-from-string str (lambda () (unfold eof-object? values (lambda (x) (read)) (read))))])
+	(let ([xs (with-input-from-string
+		      str
+		    (lambda () (unfold eof-object? values (lambda (x) (read)) (read))))])
 	  (cond [(null? xs) '(##core#undefined)]
 		[(null? (cdr xs)) (car xs)]
 		[else `(begin ,@xs)] ) ) ) ) ) )
 
 (define decompose-lambda-list ##sys#decompose-lambda-list)
 
-(define (process-lambda-documentation id doc proc)
-  proc)					; Hook this
-
 (define (llist-length llist)
   (##core#inline "C_u_i_length" llist))
 
@@ -351,6 +354,10 @@
 		[else (##sys#setslot plist 1 (alist-cons prop n (##sys#slot plist 1)))] ) )
 	(##sys#hash-table-set! db key (list (cons prop val)))) ) )
 
+(define (get-list db key prop)		; returns '() if not set
+  (let ((x (get db key prop)))
+    (or x '())))
+
 
 ;;; Line-number database management:
 
@@ -383,6 +390,7 @@
   (let ((names '((captured . cpt) (assigned . set) (boxed . box) (global . glo) (assigned-locally . stl)
 		 (contractable . con) (standard-binding . stb) (simple . sim) (inlinable . inl)
 		 (collapsable . col) (removable . rem) (constant . con)
+		 (inline-target . ilt) (inline-transient . itr)
 		 (undefined . und) (replacing . rpg) (unused . uud) (extended-binding . xtb) (inline-export . ilx)
 		 (customizable . cst) (has-unused-parameters . hup) (boxed-rest . bxr) ) ) 
 	(omit #f))
@@ -418,7 +426,7 @@
 		       ((potential-value)
 			(set! pval (cdar es)) )
 		       ((replacable home contains contained-in use-expr closure-size rest-parameter
-				    o-r/access-count captured-variables explicit-rest)
+				    captured-variables explicit-rest)
 			(printf "\t~a=~s" (caar es) (cdar es)) )
 		       ((references)
 			(set! refs (cdar es)) )
@@ -520,7 +528,8 @@
 			       [else #f] )
 			 (if ln
 			     (let ([rn (real-name name)])
-			       (list source-filename ln (or rn (##sys#symbol->qualified-string name))) )
+			       (list ln
+				     (or rn (##sys#symbol->qualified-string name))) )
 			     (##sys#symbol->qualified-string name) ) )
 		   (map walk x) ) ) ) ) )
 	    (else (make-node '##core#call '(#f) (map walk x))) ) )
@@ -558,6 +567,8 @@
 	       (loop (- n 1) (cdr vals) (cons (walk (car vals)) bindings)) ) ) )
 	((##core#unbox ##core#ref ##core#update ##core#update_i)
 	 (cons* class (walk (car subs)) params (map walk (cdr subs))) ) 
+	((##core#inline_allocate ##core#let_unboxed)
+	 (cons* class params (map walk subs)))
 	(else (cons class (append params (map walk subs)))) ) ) ) )
 
 (define (fold-boolean proc lst)
@@ -569,14 +580,14 @@
 	 (list (proc (first vars) (second vars))
 	       (fold (cdr vars)) ) ) ) ) )
 
-(define (inline-lambda-bindings llist args body copy?)
+(define (inline-lambda-bindings llist args body copy? db)
   (decompose-lambda-list
    llist
    (lambda (vars argc rest)
      (receive (largs rargs) (split-at args argc)
        (let* ([rlist (if copy? (map gensym vars) vars)]
 	      [body (if copy? 
-			(copy-node-tree-and-rename body vars rlist)
+			(copy-node-tree-and-rename body vars rlist db)
 			body) ] )
 	 (fold-right
 	  (lambda (var val body) (make-node 'let (list var) (list val body)) )
@@ -591,7 +602,7 @@
 	  (take rlist argc)
 	  largs) ) ) ) ) )
 
-(define (copy-node-tree-and-rename node vars aliases)
+(define (copy-node-tree-and-rename node vars aliases db)
   (let ([rlist (map cons vars aliases)])
     (define (rename v rl) (alist-ref v rl eq? v))
     (define (walk n rl)
@@ -599,19 +610,35 @@
 	    [params (node-parameters n)]
 	    [class (node-class n)] )
 	(case class
-	  [(##core#variable) (varnode (rename (first params) rl))]
-	  [(set!) (make-node 'set! (list (rename (first params) rl)) (map (cut walk <> rl) subs))]
+	  ((quote)
+	   (make-node class params '()))
+	  [(##core#variable) 
+	   (let ((var (first params)))
+	     (when (get db var 'contractable)
+	       (put! db var 'contractable #f) )
+	     (varnode (rename var rl))) ]
+	  [(set!) 
+	   (make-node
+	    'set! (list (rename (first params) rl))
+	    (list (walk (first subs) rl)) ) ]
 	  [(let) 
-	   (let* ([v (first params)]
-		  [a (gensym v)]
-		  [rl2 (alist-cons v a rl)] )
-	     (make-node 'let (list a) (map (cut walk <> rl2) subs)) ) ]
+	   (let* ((v (first params))
+		  (val1 (walk (first subs) rl))
+		  (a (gensym v))
+		  (rl2 (alist-cons v a rl)) )
+	     (make-node 
+	      'let (list a)
+	      (list val1 (walk (second subs) rl2)))) ]
 	  [(##core#lambda)
 	   (decompose-lambda-list
 	    (third params)
 	    (lambda (vars argc rest)
-	      (let* ([as (map gensym vars)]
-		     [rl2 (append as rl)] )
+	      (let* ((as (map (lambda (v)
+				(let ((a (gensym v)))
+				  (put! db v 'inline-transient #t)
+				  a))
+			      vars) )
+		     (rl2 (append (map cons vars as) rl)) )
 		(make-node 
 		 '##core#lambda
 		 (list (gensym 'f) (second params) ; new function-id
@@ -659,7 +686,7 @@
 			       (not (eq? 'unknown (cdr val))))))
 			((assq 'inlinable plist))
 			(lparams (node-parameters (cdr val)))
-			((get db (first lparams) 'simple)) 
+			;;((get db (first lparams) 'simple)) 
 			((not (get db sym 'hidden-refs)))
 			((case (variable-mark sym '##compiler#inline)
 			   ((yes) #t)
@@ -762,7 +789,8 @@
 (define (dump-undefined-globals db)
   (##sys#hash-table-for-each
    (lambda (sym plist)
-     (when (and (assq 'global plist)
+     (when (and (not (keyword? sym))
+		(assq 'global plist)
 		(not (assq 'assigned plist)) )
        (write sym)
        (newline) ) )
@@ -771,7 +799,8 @@
 (define (dump-defined-globals db)
   (##sys#hash-table-for-each
    (lambda (sym plist)
-     (when (and (assq 'global plist)
+     (when (and (not (keyword? sym))
+		(assq 'global plist)
 		(assq 'assigned plist))
        (write sym)
        (newline) ) )
@@ -780,7 +809,7 @@
 (define (dump-global-refs db)
   (##sys#hash-table-for-each
    (lambda (sym plist)
-     (when (assq 'global plist)
+     (when (and (not (keyword? sym)) (assq 'global plist))
        (let ((a (assq 'references plist)))
 	 (write (list sym (if a (length (cdr a)) 0)))
 	 (newline) ) ) )
@@ -1065,7 +1094,11 @@
      (cond 
        [(and (list? type) (= 3 (length type)) 
 	     (memq (car type) '(instance instance-ref)))
-	`(##tinyclos#make-instance-from-pointer ,body ,(caddr type)) ] ;XXX eggified, needs better treatment...
+	(let ((tmp (gensym)))
+	  `(let ((,tmp ,body))
+	     (and ,tmp
+		  (not (##sys#null-pointer? ,tmp))
+		  (make ,(caddr type) 'this ,tmp) ) ) ) ]
        [(and (list? type) (= 3 (length type)) (eq? 'nonnull-instance (car type)))
 	`(make ,(caddr type) 'this ,body) ]
        [else body] ) ] ) )
@@ -1232,6 +1265,7 @@ Usage: chicken FILENAME OPTION ...
     -inline                      enable inlining
     -inline-limit                set inlining threshold
     -inline-global               enable cross-module inlining
+    -unboxing                    use unboxed temporaries if possible
     -emit-inline-file FILENAME   generate file with globally inlinable
                                   procedures (implies -inline -local)
     -consult-inline-file FILENAME  explicitly load inline file
@@ -1265,8 +1299,6 @@ Usage: chicken FILENAME OPTION ...
   Obscure options:
 
     -debug MODES                 display debugging output for the given modes
-    -unsafe-libraries            marks the generated file as being linked with
-                                  the unsafe runtime system
     -raw                         do not generate implicit init- and exit code                           
     -emit-external-prototypes-first
                                  emit prototypes for callbacks before foreign
@@ -1340,22 +1372,21 @@ EOF
 
 (define (source-info->string info)
   (if (list? info)
-      (let ((file (car info))
-	    (ln (cadr info))
-	    (name (caddr info)))
-	(let ((lns (->string ln)))
-	  (conc file ": " lns (make-string (max 0 (- 4 (string-length lns))) #\space) " " name) ) )
-      (and info (->string info))) )
+      (let ((ln (car info))
+	    (name (cadr info)))
+	(conc ln ":" (make-string (max 0 (- 4 (string-length ln))) #\space) " " name) )
+      info))
 
 (define (source-info->line info)
   (if (list? info)
-      (cadr info)
-      (and info (->string info))) )
+      (car info)
+      (and info (->string info))))
 
 
 ;;; We need this for constant folding:
 
-(define (string-null? x) (string-null? x))
+(define (string-null? x) 
+  (##core#inline "C_i_string_null_p" s))
 
 
 ;;; Dump node structure:
@@ -1380,6 +1411,23 @@ EOF
   (newline) )
 
 
+;;; Hook for source information
+
+(define (read-info-hook class data val)
+  (when (and (eq? 'list-info class) (symbol? (car data)))
+    (##sys#hash-table-set!
+     ##sys#line-number-database
+     (car data)
+     (alist-cons 
+      data (conc ##sys#current-source-filename ":" val)
+      (or (##sys#hash-table-ref ##sys#line-number-database (car data))
+	  '() ) ) ) )
+  data)
+
+(define (read/source-info in)
+  (##sys#read in read-info-hook) )
+
+
 ;;; "#> ... <#" syntax:
 
 (set! ##sys#user-read-hook
