@@ -1,7 +1,7 @@
 ;;;; tcp.scm - Networking stuff
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -38,7 +38,7 @@
   (bound-to-procedure
    ##net#socket ##net#bind ##net#connect ##net#listen ##net#accept make-parameter ##sys#string-append ##sys#tcp-port->fileno
    ##sys#check-port ##sys#port-data ##sys#thread-block-for-i/o! make-string make-input-port make-output-port ##sys#substring
-   substring ##sys#make-c-string ##sys#schedule
+   substring ##sys#make-c-string ##sys#schedule ##sys#set-port-data!
    ##net#close ##net#recv ##net#send ##net#select ##net#select-write ##net#gethostaddr ##net#io-ports ##sys#update-errno
    ##sys#error ##sys#signal-hook ##net#getservbyname ##net#parse-host ##net#fresh-addr
    ##net#bind-socket ##net#shutdown)
@@ -239,7 +239,8 @@ EOF
 		       (when (eq? 0 p)
 			 (##sys#update-errno)
 			 (##sys#signal-hook
-			  #:network-error 'tcp-connect (##sys#string-append "cannot compute port from service - " strerror)
+			  #:network-error 'tcp-connect
+			  (##sys#string-append "cannot compute port from service - " strerror)
 			  s) )
 		       p) )
 		    (loop (fx+ i 1)) ) ) ) ) ) ) ) )
@@ -254,11 +255,8 @@ EOF
 
 (define (##net#bind-socket port style host)
   (##sys#check-exact port)
-  (cond-expand
-   (unsafe)
-   (else
-    (when (or (fx< port 0) (fx>= port 65535))
-      (##sys#signal-hook #:domain-error 'tcp-listen "invalid port number" port) ) ) )
+  (when (or (fx< port 0) (fx>= port 65535))
+    (##sys#signal-hook #:domain-error 'tcp-listen "invalid port number" port) )
   (let ((s (##net#socket _af_inet style 0)))
     (when (eq? _invalid_socket s)
       (##sys#update-errno)
@@ -269,16 +267,22 @@ EOF
                       return(setsockopt(socket, SOL_SOCKET, SO_REUSEADDR, (const char *)&yes, sizeof(int)));") 
 		   s) )
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "error while setting up socket - " strerror) s) )
+      (##sys#signal-hook 
+       #:network-error 'tcp-listen
+       (##sys#string-append "error while setting up socket - " strerror) s) )
     (let ((addr (make-string _sockaddr_in_size)))
       (if host
 	  (unless (##net#gethostaddr addr host port)
-	    (##sys#signal-hook #:network-error 'tcp-listen "getting listener host IP failed - " host port) )
+	    (##sys#signal-hook 
+	     #:network-error 'tcp-listen 
+	     "getting listener host IP failed - " host port) )
 	  (##net#fresh-addr addr port) )
       (let ((b (##net#bind s addr _sockaddr_in_size)))
 	(when (eq? -1 b)
 	  (##sys#update-errno)
-	  (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "cannot bind to socket - " strerror) s port) )
+	  (##sys#signal-hook
+	   #:network-error 'tcp-listen
+	   (##sys#string-append "cannot bind to socket - " strerror) s port) )
 	(values s addr) ) ) ) )
 
 (define-constant default-backlog 10)
@@ -290,7 +294,9 @@ EOF
       (let ((l (##net#listen s w)))
 	(when (eq? -1 l)
 	  (##sys#update-errno)
-	  (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "cannot listen on socket - " strerror) s port) )
+	  (##sys#signal-hook 
+	   #:network-error 'tcp-listen
+	   (##sys#string-append "cannot listen on socket - " strerror) s port) )
 	(##sys#make-structure 'tcp-listener s) ) ) ) )
 
 (define (tcp-listener? x) 
@@ -302,7 +308,9 @@ EOF
   (let ((s (##sys#slot tcpl 1)))
     (when (fx= -1 (##net#close s))
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error 'tcp-close (##sys#string-append "cannot close TCP socket - " strerror) tcpl) ) ) )
+      (##sys#signal-hook 
+       #:network-error 'tcp-close
+       (##sys#string-append "cannot close TCP socket - " strerror) tcpl) ) ) )
 
 (define-constant +input-buffer-size+ 1024)
 (define-constant +output-chunk-size+ 8192)
@@ -331,9 +339,10 @@ EOF
     (lambda (fd)
       (unless (##net#make-nonblocking fd)
 	(##sys#update-errno)
-	(##sys#signal-hook #:network-error (##sys#string-append "cannot create TCP ports - " strerror)) )
+	(##sys#signal-hook 
+	 #:network-error (##sys#string-append "cannot create TCP ports - " strerror)) )
       (let* ((buf (make-string +input-buffer-size+))
-	     (data (vector fd #f #f))
+	     (data (vector fd #f #f buf 0))
 	     (buflen 0)
 	     (bufindex 0)
 	     (iclosed #f) 
@@ -356,8 +365,8 @@ EOF
 				  (yield)
 				  (when (##sys#slot ##sys#current-thread 13)
 				    (##sys#signal-hook
-				     #:network-error
-				     "read operation timed out" fd) )
+				     #:network-timeout-error
+				     "read operation timed out" tmr fd) )
 				  (loop) )
 				 (else
 				  (##sys#update-errno)
@@ -367,6 +376,7 @@ EOF
 				   fd) ) ) )
 			  (else
 			   (set! buflen n)
+			   (##sys#setislot data 4 n)
 			   (set! bufindex 0) ) ) ) ) ) )
 	     (in
 	      (make-input-port
@@ -398,7 +408,12 @@ EOF
 		      #:network-error
 		      (##sys#string-append "cannot close socket input port - " strerror)
 		      fd) ) ) )
-	       #f
+	       (lambda ()
+		 (when (fx>= bufindex buflen)
+		   (read-input))
+		 (if (fx< bufindex buflen)
+		     (##core#inline "C_subchar" buf bufindex)
+		     #!eof))
 	       (lambda (p n dest start)	; read-string!
 		 (let loop ((n n) (m 0) (start start))
 		   (cond ((eq? n 0) m)
@@ -458,8 +473,8 @@ EOF
 				  (yield) 
 				  (when (##sys#slot ##sys#current-thread 13)
 				    (##sys#signal-hook
-				     #:network-error
-				     "write operation timed out" fd) )
+				     #:network-timeout-error
+				     "write operation timed out" tmw fd) )
 				  (loop len offset) )
 				 (else
 				  (##sys#update-errno)
@@ -500,8 +515,8 @@ EOF
 	(##sys#setslot out 3 "(tcp)")
 	(##sys#setslot in 7 'socket)
 	(##sys#setslot out 7 'socket)
-	(##sys#setslot in 9 data)
-	(##sys#setslot out 9 data)
+	(##sys#set-port-data! in data)
+	(##sys#set-port-data! out data)
 	(values in out) ) ) ) )
 
 (define (tcp-accept tcpl)
@@ -526,9 +541,9 @@ EOF
 	    (yield)
 	    (when (##sys#slot ##sys#current-thread 13)
 	      (##sys#signal-hook
-	       #:network-error
+	       #:network-timeout-error
 	       'tcp-accept
-	       "accept operation timed out" fd) )
+	       "accept operation timed out" tma fd) )
 	    (loop) ) ) ) ) )
 
 (define (tcp-accept-ready? tcpl)
@@ -569,7 +584,9 @@ EOF
 	 host port) )
       (when (eq? -1 s)
 	(##sys#update-errno)
-	(##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "cannot create socket - " strerror) host port) )
+	(##sys#signal-hook 
+	 #:network-error 'tcp-connect
+	 (##sys#string-append "cannot create socket - " strerror) host port) )
       (unless (##net#gethostaddr addr host port)
 	(##sys#signal-hook #:network-error 'tcp-connect "cannot find host address" host) )
       (unless (##net#make-nonblocking s)
@@ -589,18 +606,22 @@ EOF
 		  (yield)
 		  (when (##sys#slot ##sys#current-thread 13)
 		    (##sys#signal-hook
-		     #:network-error
+		     #:network-timeout-error
 		     'tcp-connect
-		     "connect operation timed out" s) )
+		     "connect operation timed out" tmc s) )
 		  (loop) ) ) )
 	    (fail) ) )
       (let ((err (get-socket-error s)))
 	(cond ((= err -1)
 	       (##net#close s)
-	       (##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "getsockopt() failed - " strerror)))
+	       (##sys#signal-hook 
+		#:network-error 'tcp-connect
+		(##sys#string-append "getsockopt() failed - " strerror)))
 	      ((> err 0)
 	       (##net#close s)
-	       (##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "cannot create socket - " (general-strerror err))))))
+	       (##sys#signal-hook 
+		#:network-error 'tcp-connect
+		(##sys#string-append "cannot create socket - " (general-strerror err))))))
       (##net#io-ports s) ) ) )
 
 (define (##sys#tcp-port->fileno p)
@@ -614,18 +635,26 @@ EOF
   (let ((fd (##sys#tcp-port->fileno p)))
     (values 
      (or (##net#getsockname fd)
-	 (##sys#signal-hook #:network-error 'tcp-addresses (##sys#string-append "cannot compute local address - " strerror) p) )
+	 (##sys#signal-hook 
+	  #:network-error 'tcp-addresses
+	  (##sys#string-append "cannot compute local address - " strerror) p) )
      (or (##net#getpeername fd)
-	 (##sys#signal-hook #:network-error 'tcp-addresses (##sys#string-append "cannot compute remote address - " strerror) p) ) ) ) )
+	 (##sys#signal-hook
+	  #:network-error 'tcp-addresses
+	  (##sys#string-append "cannot compute remote address - " strerror) p) ) ) ) )
 
 (define (tcp-port-numbers p)
   (##sys#check-port p 'tcp-port-numbers)
   (let ((fd (##sys#tcp-port->fileno p)))
     (values
      (or (##net#getsockport fd)
-	 (##sys#signal-hook #:network-error 'tcp-port-numbers (##sys#string-append "cannot compute local port - " strerror) p) )
+	 (##sys#signal-hook 
+	  #:network-error 'tcp-port-numbers
+	  (##sys#string-append "cannot compute local port - " strerror) p) )
      (or (##net#getpeerport fd)
-	 (##sys#signal-hook #:network-error 'tcp-port-numbers (##sys#string-append "cannot compute remote port - " strerror) p) ) ) ) )
+	 (##sys#signal-hook
+	  #:network-error 'tcp-port-numbers
+	  (##sys#string-append "cannot compute remote port - " strerror) p) ) ) ) )
 
 (define (tcp-listener-port tcpl)
   (##sys#check-structure tcpl 'tcp-listener 'tcp-listener-port)
