@@ -1,4 +1,4 @@
-;;; compiler-tests-2.scm - tests for compiler with -lambda-lift
+;;; compiler-tests-2.scm - tests for lambda-lifting
 
 
 ;;; rev. 12113 - lambda-lifting breakage, because lambda-bound variables
@@ -12,7 +12,7 @@
    (+ a b))
  (plus 1))
 
-(print (plus1 1))
+(assert (= 2 (plus1 1)))
 
 (define (len lst)
  (define (len n)
@@ -23,4 +23,4 @@
        n))
  (len 0))
 
-(print (len '(1 2 3)))
+(assert (= 3 (len '(1 2 3))))
