@@ -1,5 +1,7 @@
 ;;;; hash-table-tests.scm
 
+(require-extension srfi-69)
+
 (print "SRFI 69 procedures")
 (assert (eq? hash equal?-hash))
 (assert (eq? hash-by-identity eq?-hash))
@@ -57,7 +59,7 @@
 (assert (eq? (hash-table-ref ht 'baz) 'foo))
 (assert (= (hash-table-size ht) 2))
 
-(print "HT - AList")
+(print "HT - A-List")
 (let ([alist (hash-table->alist ht)])
   (assert (list? alist))
   (assert (= (length alist) 2))
@@ -78,7 +80,7 @@
 (assert (not (hash-table-exists? ht 'baz)))
 (assert (= (hash-table-size ht) 0))
 
-(print "HT - Make from AList")
+(print "HT - Make from A-List")
 (set! ht (alist->hash-table '(("abc" . #t) ("cbs" . #t) ("cnn" . #f))))
 (assert (hash-table? ht))
 (assert (= (hash-table-size ht) 3))
