@@ -28,7 +28,11 @@
 
 (test "./" (normalize-pathname "" 'unix))
 (test ".\\" (normalize-pathname "" 'windows))
-(test "./" (normalize-pathname "/" 'unix))
+(test "\\..\\" (normalize-pathname "/../" 'windows))
+(test "\\." (normalize-pathname "/abc/../." 'windows))
+(test "/." (normalize-pathname "/" 'unix))
+(test "/." (normalize-pathname "/./" 'unix))
+(test "/." (normalize-pathname "/." 'unix))
 (test "./" (normalize-pathname "./" 'unix))
 (test "a" (normalize-pathname "a"))
 (test "a/" (normalize-pathname "a/" 'unix))
@@ -55,3 +59,15 @@
 (test (string-append home "/foo") (normalize-pathname "~/foo" 'unix))
 (test "c:~/foo" (normalize-pathname "c:~/foo" 'unix))
 (test (string-append home "\\foo") (normalize-pathname "c:~\\foo" 'windows))
+
+(assert (directory-null? "/.//"))
+(assert (directory-null? ""))
+(assert (not (directory-null? "//foo//")))
+
+(test '(#f "/" (".")) (receive (decompose-directory "/.//")))
+(test '(#f "\\" (".")) (receive (decompose-directory (normalize-pathname "/.//" 'windows))))
+(test '(#f "/" #f) (receive (decompose-directory "///\\///")))
+(test '(#f "/" ("foo")) (receive (decompose-directory "//foo//")))
+(test '(#f "/" ("foo" "bar")) (receive (decompose-directory "//foo//bar")))
+(test '(#f #f (" " "foo" "bar")) (receive (decompose-directory " //foo//bar")))
+(test '(#f #f ("foo" "bar")) (receive (decompose-directory "foo//bar/")))
