@@ -1,7 +1,7 @@
 ;;;; utils.scm - Utilities for scripting and file stuff
 ;
+; Copyright (c) 2008-2010, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -27,7 +27,7 @@
 
 (declare
   (unit utils)
-  (uses extras srfi-13 posix files)
+  (uses extras srfi-13 posix files regex)
   (usual-integrations)
   (fixnum)
   (hide chop-pds)
@@ -40,7 +40,7 @@
     (always-bound
       ##sys#windows-platform)
     (bound-to-procedure
-      ##sys#check-port port? read-string for-each-line read-line with-input-from-file
+      ##sys#check-port port? read-string read-line with-input-from-file
       command-line-arguments
       string-append
       system)
@@ -64,35 +64,6 @@
 	  (##sys#error "shell invocation failed with non-zero return status" str n) ) ) ) ) )
 
 
-;;; Handy I/O procedures:
-
-(define for-each-line
-  (let ([read-line read-line])
-    (lambda (proc . port)
-      (let ([port (if (pair? port) (car port) ##sys#standard-input)])
-	(##sys#check-port port 'for-each-line)
-	(let loop ()
-	  (let ([ln (read-line port)])
-	    (unless (eof-object? ln)
-	      (proc ln)
-	      (loop) ) ) ) ) ) ) )
-
-
-;; This one is from William Annis:
-
-(define (for-each-argv-line thunk)
-  (define (file-iterator file thunk)
-    (if (string=? file "-")
-        (for-each-line thunk)
-        (with-input-from-file file (cut for-each-line thunk) ) ) )
-  (let ((args (command-line-arguments)))
-    (if (null? args)
-        ;; If no arguments, take from stdin,
-        (for-each-line thunk)
-        ;; otherwise, hit each file named in argv.
-        (for-each (lambda (arg) (file-iterator arg thunk)) args))))
-
-
 ;;; Read file as string from given filename or port:
 
 (define (read-all . file)
@@ -132,7 +103,7 @@
     (lambda (filename #!key (options '()) output-file (load #t))
       (let ((cscpath (or (file-exists? (make-pathname path csc)) "csc"))
 	    (tmpfile (and (not output-file) (create-temporary-file "so")))
-	    (crapshell (cond-expand ((or mingw32 msvc) #t) (else #f))))
+	    (crapshell (memq (build-platform) '(mingw32 msvc))))
 	(print "; compiling " filename " ...")
 	(system* 
 	 "~a~a -s ~a ~a -o ~a~a" 
@@ -154,3 +125,20 @@
 		  (abort ex))
 	      (load-file f)
 	      f)))))))
+
+
+;;; Scan lines until regex or predicate matches
+
+(define scan-input-lines
+  (let ((regexp regexp)
+	(read-line read-line)
+	(string-search string-search))
+    (lambda (rx #!optional (port ##sys#standard-input))
+      (let ((rx (if (procedure? rx)
+		    rx
+		    (cut string-search (regexp rx) <>))))
+	(let loop ()
+	  (let ((ln (read-line port)))
+	    (and (not (eof-object? ln))
+		 (or (rx ln)
+		     (loop)))))))))
